/*
 * DTMF decoding routines
 * using a 'goertzel' algorithm
 * modified to give square power as described in
 * 'Digital Signal Processing Applications using the
 *  ADSP-2100 family' by Analog Devices
 *       gcc file.c -lm -o output
 * we could choose the same N value for all the
 * frequencies and do the goertzel algorithm in
 * parallel for all the tones but instead I chose
 * simplicity over speed and call the goertzel routine
 * for each frequency.  This also gives us freedom
 * in choosing the best N for each frequency.
 * 
 *                        Tim N.
 */

#include <stdio.h>
#include <math.h>
#ifdef PI
#ifndef M_PI
#define M_PI PI
#endif
#endif

#define FSAMPLE 8000
#define THRESH  100.0
#define FLEN    205    /* length of frames */

/* modified goertzel algorithm (returns square power)
 * detecting Ftone sampled at Fsample
 * using length N
 * N should be chosen so that 
 *   k = N * Ftone/Fsample
 * is close to an integer
 */
float goertzel(data,Ftone,Fsample,N)
char *data;
int Ftone,Fsample,N;
{
  float coef;   /* determines freq we look for */
  float u1,u2;  /* delay storage */
  float input,temp,power;
  int i,k;

  k = (int) (0.5 +  (float)N*(float)Ftone/(float)Fsample);
  coef = 2.0 * cos(2.0 * M_PI * (float)k/(float)N );

  u1=0.0; u2=0.0;
  for(i=0;i<N;i++) {
     input = data[i]/128.0;
     temp = u1;
     u1 = input + coef * u1 - u2;
     u2 = temp;
  }
  power = u1*u1 + u2*u2 - coef*u1*u2;
  return(power);
}

/* take one frame of data
 * and decode as dtmf data
 * 
 * prints out one result per frame as
 * 123456789*0# for valid tone and - for 
 * invalid or no data. 
 */
decode(data)
char *data;
{
  float r[4],c[4];
  int i,row,col,x,rcount,ccount;
  char tab[] = { '1', '2', '3', 'A', 
	         '4', '5', '6', 'B',
		 '7', '8', '9', 'C',
		 '*', '0', '#', 'D',  };
  
  /* rows */
  r[0]= goertzel(data, 697, FSAMPLE, FLEN);
  r[1]= goertzel(data, 770, FSAMPLE, FLEN);
  r[2]= goertzel(data, 852, FSAMPLE, FLEN);
  r[3]= goertzel(data, 941, FSAMPLE, FLEN);
  
  /* columns */
  c[0]= goertzel(data, 1209, FSAMPLE, FLEN);
  c[1]= goertzel(data, 1336, FSAMPLE, FLEN);
  c[2]= goertzel(data, 1477, FSAMPLE, FLEN);
  c[3]= goertzel(data, 1633, FSAMPLE, FLEN);
  
  for(i=0,rcount=0, ccount=0; i<4; i++) {
#ifdef NOISY
     printf("%f\t %f\n",r[i], c[i]);
#endif
     if(r[i] > THRESH) {
	row = i;
	rcount++;
      } 
      if(c[i] > THRESH) {
	ccount++;
	col = i;
      } 
  }
  
  if (rcount==1 && ccount==1) 
     putchar(tab[col + row*4]); 
  else if (rcount || ccount )
     putchar('-');
  else
     putchar(' ');

#ifdef NOISY
  printf("\n");
#endif
  fflush(stdout);
}



#define BUFLEN FLEN+1   /* must be greater than any N values used */

main(argc,argv)   /* read in frames, call decode */
int argc;
char **argv;
{
  int fd;
  char buf[BUFLEN];
  
  if(argc!=2) {
    printf("usage:  %s input-file\n",argv[0]);
    exit(1);
  }
  printf("File: %s\n",argv[1]);
  fd=open(argv[1],0);
  if(fd<0) {
    perror("open");
    exit(1);
  }

  /* assumes a single read of length BUFLEN will work fine
  * and that the input file has samples in
  * 8 bit signed format
  * should be sampled at 8KHz (or we will detect wrong freqs)
  */
  while(read(fd,buf,BUFLEN) == BUFLEN) {  /* read in a frame and decode */
    decode(buf); 
  }
  printf("\nDone.\n");
  return(0);
}
