
/* *************************************************************************
 *  File : netmedia.c
 * -------------------------------------------------------------------------
 *  All rights are reserved by Bo-kyun Na, IntellActive Research Inc.
 * -------------------------------------------------------------------------
 * This example reads an image from /dev/video, transmits compressed video 
 * over the local network, and optionally displays the image in Xwindow.
 * ************************************************************************* */

#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>
#include <semaphore.h>

#include "params.h"
#include "network.h"
#include "transceive.h"
#include "ext-pthread.h"
// #include "vdo.h"		/* video related header */
#include "h261.h"
#include "odo.h"		/* audio related headers */

/* Thread related variables */
pthread_mutex_t utex_lock = PTHREAD_MUTEX_INITIALIZER;
int count = 0;			/* used in thrd_stop () */
sem_t VS, VR, AS, AR;


unsigned char *clp;
void error ();

unsigned long wpixel[3];
int expand;

int verbose = 0;

void *Thread_TxVideo (void *);
void *Thread_RxVideo (void *);
void *Thread_TxAudio (void *);
void *Thread_RxAudio (void *);



/* *************************************************************************
 * thrd_stop (sem_t *sem)
 *
 * ************************************************************************* */
void thrd_stop (sem_t *sem)
{
  pthread_mutex_lock (&utex_lock);
  (count) ? sem_post (sem), count = 0 : count++;
  pthread_mutex_unlock (&utex_lock);
}


/* *************************************************************************
 * usage ()
 *
 *
 * ************************************************************************* */
void usage (char *pgm)
{
  fprintf (stderr, "usage: %s [options] \n \
Legal options:                                       Defaults\n \
<-i file_name> : audio input file name               0\n \
<-o file_name> : audio output file name              0\n \
", pgm);

  exit (1);
}




/* *************************************************************************
 * Function    : void *Thread_TxAudio (void *arg)
 * ------------:------------------------------------------------------------
 * Inputs      : 
 * Outputs     : 
 * Description : 
 * ************************************************************************* */
void *Thread_TxAudio (void *arg)
{
  int fd;			/* sound device file descriptor */
  //  int status;			/* return status of system calls */
  int rd;			/* size in byte unit */
  int odo_frame;

  unsigned char *buf;		/* this buffer holds the digitized audio */

  /* Network parameters */
  Parameters *param = (Parameters *) arg;
  TransceiveState *tstate;


  odo_frame = ODO_LENGTH * ODO_RATE * ODO_SIZE * ODO_CHANNELS / 8;
  buf = (unsigned char *) malloc(odo_frame);
  if (buf == NULL) {
    printf ("malloc fail: sizeof (buf) = %d\n", odo_frame);
    exit (1);
  }

  /* Initialization of transmission parameters */
  if (param->transmit)
    tstate = (TransceiveState *) initialize_transmit(param, AUDIO);

  for ( ; ; ) {			/* loop until Control-C */
    sem_wait (&AS);		/* Rx first */
    //    if (verbose) fprintf (stderr, "SendAudThread\n");

    /* Initialization of audio */
    fd = BV_ODO_initialize (param->dev_odo);
    rd = BV_ODO_record (fd, buf, odo_frame);
    if (verbose) printf ("Recorded audio size is %d\n", rd);
    close (fd);

    BV_odo_packetize (buf, odo_frame, tstate, param);

    /* echo recorded audio */
    //    BV_ODO_play (fd, buf, odo_frame);
    /* wait for playback to complete before recording again */
    //    status = ioctl(fd, SOUND_PCM_SYNC, 0); 
    //    if (status == -1)
    //      perror("SOUND_PCM_SYNC ioctl failed");

    //    if (verbose) printf ("AudioSend Thread done\n");
    //    thrd_stop (&VR);		/* Rx first */
    sem_post (&AR);		/* only for audio */
  }
}



/* *************************************************************************
 * Function    : Thread : void *Thread_RxAudio(void *arg)
 * ------------:------------------------------------------------------------
 * Inputs      : 
 * Outputs     : 
 * Description : 
 * ************************************************************************* */
void *Thread_RxAudio (void *arg)
{
  int fd;			/* sound device file descriptor */
  int wr;			/* played audio size in byte unit */
  int status;			/* return status of system calls */
  int odo_frame;		/* size of audio buffer */

  unsigned char *buf;		/* this buffer holds the digitized audio */

  Parameters *param = (Parameters *) arg;
  TransceiveState *RxState;
  MsgType *message;


  odo_frame = ODO_LENGTH * ODO_RATE * ODO_SIZE * ODO_CHANNELS / 8;
  buf = (unsigned char *) malloc(odo_frame);
  if (buf == NULL) {
    printf ("malloc fail: sizeof (buf) = %d\n", odo_frame);
    exit (1);
  }

  message = (MsgType *) malloc (sizeof (MsgType));
  if (message == NULL) {
    printf ("malloc fail: message = %d\n", sizeof (MsgType));
    exit (1);
  }

  /* Init. of network */
  RxState = (TransceiveState *) initialize_receive (param, AUDIO);

  /* loop for audio play until Control-C */
  for ( ; ; ) {		
    sem_wait (&AR);
    //    sem_post (&AS);
    //    if (verbose)  fprintf (stderr, "RecvAudThread ");

    message = (MsgType *) BV_ODO_receive (RxState, message);

    /* Initialization of audio */
    fd = BV_ODO_initialize (param->dev_odo);

    /* play audio data */
    wr = BV_ODO_play (fd, message->data, odo_frame);
    if (verbose) printf ("Played audio size is %d bytes\n", wr);

    /* wait for playback to complete before recording again */
    status = ioctl(fd, SNDCTL_DSP_SYNC, 0); 
    if (status == -1)
      perror("SOUND_PCM_SYNC ioctl failed");
    close (fd);

    //    if (verbose) printf ("AudioReceive Thread done\n");
    sem_post (&AS);
   //    thrd_stop (&VR);
  }
}


/* *************************************************************************
 * Function    : void main (int argc, char **argv) 
 * ------------:------------------------------------------------------------
 * Inputs      : 
 * Outputs     : 
 * Description : 
 * ************************************************************************* */
int main (int argc, char **argv)
{
  int   i;

  Parameters *param;

  pthread_t tid_VidTx,
            tid_VidRx,
            tid_AudRx, 
            tid_AudTx;

  pthread_attr_t attr;

  SEM_INIT (&VS, 0, 0);
  SEM_INIT (&VR, 0, 0);
  SEM_INIT (&AS, 0, 0);
  SEM_INIT (&AR, 0, 0);

  PTHREAD_ATTR_INIT (&attr);
  pthread_attr_setscope (&attr, PTHREAD_SCOPE_SYSTEM);


  /* Initialization of video and audio parameters  */
  param = (Parameters *) init_Param(argc, argv);
  param->argc = argc;
  param->argv = argv;

  /* audio device node name */
  strcpy (param->dev_odo, ODO_DEVICE);		

  /* video device node name */
  //  strcpy(param->dev_vdo, VDO_DEVICE);

  /* Process command line arguments */
  for (i = 1; i < argc; i++) {
    if (!strcmp (argv[i], "-C")) {
      if (!strcmp(argv[++i], "H261")) {
      	param->codec = H261;
      	param->cis_type = "H261";
      }
      else if (!strcmp(argv[i], "H263")) {
      	param->codec = H263;
      	param->cis_type = "H263";
      }
      else if (!strcmp(argv[i], "JPEG")) {
      	param->codec = JPEG;
      	param->cis_type = "Jpeg";
      }
      else if (!strcmp(argv[i], "MPEG")) {
      	param->codec = MPEG;
      	param->cis_type = "Mpeg1";
      }
      else if (!strcmp(argv[i], "RAW")) {
	param->codec = RAW;
      	param->cis_type = "RAW";
      }
      else {
	printf ("Specifiy your codec mode\n");
        usage(argv[0]);
      }
      if (verbose)
	printf ("codec is %s\n", param->cis_type);
    }
    if(!strcmp(argv[i], "-D")) {
      strcpy (param->dev_vdo, argv[++i]);
    }
    if(!strcmp(argv[i], "-d")) {
      param->depth = atoi(argv[++i]);
      if (param->depth != (15 || 16 || 24 || 32)) {
	printf("Unrecognized display depth. Type correct depth\n");
        usage(argv[0]);
      }
    }
    if (!strcmp(argv[i], "-i")) {
      strcpy (param->rdfile, argv[++i]);
      param->fp = fopen(param->rdfile, "r");
      if (param->fp == NULL) 
	printf ("Can't open image file: %s\n", param->rdfile);
      else  printf ("Image file, \"%s\", is opened.\n", param->rdfile);
    }
    if (!strcmp(argv[i], "-o")) {
      strcpy (param->wrfile, argv[++i]);
      param->fp = fopen(param->wrfile, "w");
      if (param->fp == NULL) 
	printf ("Can't open image file: %s\n", param->wrfile);
      else  printf ("Image file, \"%s\", is opened\n", param->wrfile);
    }
    //    if(!strcmp(argv[i], "-r")) {
    //    rate = atoi(argv[++i]);
    //    }
    if(!strcmp(argv[i], "-s")) {
      param->width = atoi(argv[++i]);
      param->height = atoi(argv[++i]);
    }
    if(!strcmp(argv[i], "-v")) {
      verbose = 1;
    }
  }

  //  sem_post (&VR);		/* Rx first */
  sem_post (&AS);		/* If AR, everybody wait to recv! thus AS! */

  /* Creation of send/receive video threads */
  sleep (1);
  //  PTHREAD_CREATE (&tid_VidTx, &attr, Thread_TxVideo, (void *) param);

  sleep (1);
  //  PTHREAD_CREATE (&tid_VidRx, &attr, Thread_RxVideo, (void *) param);

  sleep (1);
  PTHREAD_CREATE (&tid_AudTx, &attr, Thread_TxAudio, (void *) param);      

  sleep (1);
  PTHREAD_CREATE (&tid_AudRx, &attr, Thread_RxAudio, (void *) param);  

  pthread_exit ((void *) NULL);
}

/* *************************************************************************
 * Function    : 
 * ------------:------------------------------------------------------------
 * Inputs      : 
 * Outputs     : 
 * Description : 
 * ************************************************************************* */
