#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <sys/ioctl.h>

#include "i2c.h"

char *dev="/dev/i2c0";
int  wr=0, x=0, adr=0xa0;

/* ---------------------------------------------------------------------- */

void usage(char *name)
{
    fprintf(stderr,"This is a i2c EEPROM tool\n");
    fprintf(stderr,"  read  data: %s [ options ] > file\n",name);
    fprintf(stderr,"  write data: %s [ options ] -w < file\n",name);
    fprintf(stderr,"\n");
    fprintf(stderr,"File format is a hex dump.  Without \"-w\" switch\n");
    fprintf(stderr,"only parsing (and printing) of the data is done,\n");
    fprintf(stderr,"writing is skipped.\n");
    fprintf(stderr,"\n");
    fprintf(stderr,"options:\n");
    fprintf(stderr,"  -d device  device to use         [%s]\n",dev);
    fprintf(stderr,"  -a addr    set slave address     [0x%02x]\n",adr);
    fprintf(stderr,"  -x         print some hauppauge data decoded\n");
    fprintf(stderr,"\n");
    exit(1);
}

/* ---------------------------------------------------------------------- */
/* Hauppauge boards */

#define STR(array,i) (i < sizeof(array)/sizeof(char*) ? array[i] : "unknown")

static char *h_tuner [] = {
    "",
    "External",
    "Unspecified",
    "Phillips FI1216",
    "Phillips FI1216MF",
    "Phillips FI1236",
    "Phillips FI1246",
    "Phillips FI1256",
    "Phillips FI1216 MK2",
    "Phillips FI1216MF MK2",
    "Phillips FI1236 MK2",
    "Phillips FI1246 MK2",
    "Phillips FI1256 MK2",
    "Temic 4032FY5",
    "Temic 4002FH5",
    "Temic 4062FY5",
    "Phillips FR1216 MK2",
    "Phillips FR1216MF MK2",
    "Phillips FR1236 MK2",
    "Phillips FR1246 MK2",
    "Phillips FR1256 MK2",
    "Phillips FM1216",
    "Phillips FM1216MF",
    "Phillips FM1236",
};

void dump_hauppauge(unsigned char *buf)
{
    /* block 1 */
    printf("\nHauppauge eeprom decoded:\n");
    printf("  Model: %d\n",(int)(buf[12]<<8 | buf[11]));
    printf("  Tuner: %s\n",STR(h_tuner,buf[9]));
}

/* ---------------------------------------------------------------------- */

void dump_buf(unsigned char *buf)
{
    int i,j;
    
    for (i = 0; i < 256; i += 16) {
	printf("%04x  ",i);
	for (j = i; j < i+16; j++) {
	    if (!(j%4))
		printf(" ");
	    printf("%02x ",buf[j]);
	}
	printf("  ");
	for (j = i; j < i+16; j++)
	    printf("%c",isalnum(buf[j]) ? buf[j] : '.');
	printf("\n");
    }
}

int parse_buf(unsigned char *buf)
{
    int  i,j,n,pos,count;
    char line[100];
    
    for (i = 0; i < 256; i += 16) {
	if (NULL == fgets(line,99,stdin)) {
	    fprintf(stderr,"unexpected EOF\n");
	    return -1;
	}
	if ('#' == line[0] || '\n' == line[0]) {
            i -= 16;
	    continue;
	}
	if (1 != sscanf(line,"%x%n",&n,&pos)) {
	    fprintf(stderr,"addr parse error (%d)\n",i>>4);
	    return -1;
	}
	if (n != i) {
	    fprintf(stderr,"addr mismatch\n");
	    return -1;
	}
	for (j = i; j < i+16; j++) {
	    if (1 != sscanf(line+pos,"%x%n",&n,&count)) {
		fprintf(stderr,"value parse error\n");
		return -1;
	    }
	    buf[j] = n;
	    pos += count;
	}
    }
    return 0;
}

/* ---------------------------------------------------------------------- */

void
read_buf(int fd, unsigned char *buf)
{
    int i,n=8; 
    unsigned char addr;
    
    for (i = 0; i < 256; i += n) {
	addr = i;
	if (-1 == (write(fd,&addr,1))) {
	    fprintf(stderr,"write addr %s: %s\n",dev,strerror(errno));
	    exit(1);
	}
	if (-1 == (read(fd,buf+i,n))) {
	    fprintf(stderr,"read data %s: %s\n",dev,strerror(errno));
	    exit(1);
	}
    }
}

void
write_buf(int fd, unsigned char *buf)
{
    int i,j,n = 8; 
    unsigned char tmp[17];
    
    for (i = 0; i < 256; i += n) {
	tmp[0] = i;
	for (j = 0; j < n; j++)
	    tmp[j+1] = buf[i+j];
	if (-1 == (write(fd,tmp,n+1))) {
	    fprintf(stderr," write data %s: %s\n",dev,strerror(errno));
	    exit(1);
	}
	fprintf(stderr,"*");
	usleep(100000); /* 0.1 sec */
    }
}

/* ---------------------------------------------------------------------- */

int main(int argc, char *argv[])
{
    int            f,c;
    unsigned char  buf[256];
    
    /* parse options */
    opterr=1;
    while ( (c=getopt(argc,argv,"hwxa:d:")) != -1) {
	switch (c){
	case 'w':
	    wr=1;
	    break;
	case 'x':
	    x=1;
	    break;
	case 'd':
	    dev = optarg;
	    break;
	case  'a':
	    adr = strtol(optarg,NULL,0);
	    break;
	case 'h':
	default:
	    usage(argv[0]);
	}
    }

    if (-1 == (f = open(dev,O_RDWR))) {
	fprintf(stderr,"open %s: %s\n",dev,strerror(errno));
	exit(1);
    }
    ioctl(f,I2C_SLAVE,adr>>1);
    memset(buf,0,256);

    if (isatty(fileno(stdin))) {
	/* read */
	read_buf(f,buf);
	dump_buf(buf);
	if (x)
	    dump_hauppauge(buf);
    } else {
	/* write */
	if (-1 == parse_buf(buf))
	    exit(1);
	dump_buf(buf);
	if (wr) {
	    fprintf(stderr,"writing to eeprom now... ");
	    write_buf(f,buf);
	    fprintf(stderr," ok\n");
	}
    }
    
    close(f);
    exit(0);
}
