
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <math.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/param.h>
#include <sys/file.h>
#include <sys/filio.h>
#include <sys/stat.h>
#include <sys/signal.h>
#include <sys/ioctl.h>
#include <stropts.h>

#include <sys/audioio.h>

#include <multimedia/libaudio.h>
#include <multimedia/audio_device.h>
#include <multimedia/audio_encode.h>

#include "oti_libaudio.h"
#include "oti_audio_device.h"

#define MAX_BUFFER_SIZE (1024*4)

typedef struct {
  int output_fd; 
  int   byte_count;
  int   recording;
  int   hash_amt;
  int   hash_cnt;

} REC;

typedef struct {
  int   Audio_fd;
  int   Audioctl_fd;
  int   poll_enabled;

  char  filename[80];
  char  aud_data[MAX_BUFFER_SIZE];

  REC rec;
} ODO;


void usage(char* pgm)
{
  fprintf(stderr,"usage: %s [options] -o filename\n",pgm);
  fprintf(stderr,"Legal Options:                                  Defaults\n");
  fprintf(stderr, " <-D device>   : device name                   /dev/o1k0\n");  
  fprintf(stderr, " <-f format>   : encoding format               0\n");
  fprintf(stderr, "                 (0=ULAW,  1=ALAW, 2=PCM8,\n");
  fprintf(stderr, "                  3=PCM16, 4=G722, 5=G728)\n");
  fprintf(stderr, " <-r rate>     : sample rate                   8000\n");
  fprintf(stderr, "                 (8000, 11025, 16000, 22050)\n");
  fprintf(stderr, " <-c channels> : channels (1,2)                1\n");
  fprintf(stderr, " <-p port>     : input port (0=LINE,1=MIC)     0\n");
  fprintf(stderr, " <-g gain>     : input gain (0..1.0)           .125\n");
  fprintf(stderr, " <-m mgain>    : monitor gain (0..1.0)         0\n");
  fprintf(stderr, " <-i info>     : audio file string info        \"\"\n");
  fprintf(stderr, " <-b size>     : buffer size                   1000\n");
  fprintf(stderr, " <-h amount>   : print hash marks after amount 0\n");  
}

/*------------------------------------------------------------------------
 *
 *  Function:	record_service
 *
 *  Description:
 *	This routine is called async from the POLL signal handler.
 *      Care is taken not to enter the routine recursively. Typically
 *      the POLL signal was issued because there is data avail in
 *      the audio device. We read some data from the audio device and
 *      write this audio data to the file. 
 *      If an error occurs (overflow), the error is cleared and we
 *      continue playing.
 *
 *  Parameters: none
 *	
 *  Returns: void
 *	
 *------------------------------------------------------------------------*/

void record_service(ODO *odo)
{
  int	read_size;
  int	rtn, wrtn;
  unsigned int error;

  static int in_record_service = 0;

  /*
   * avoid recursive calls
   */
  if (in_record_service != 0){
    return;
  }
  in_record_service = 1;

  if (!odo->rec.recording){
    return;
  }
  
  do {
    /* Determine how much we can read */
    if (oti_ioctl(odo->Audio_fd, FIONREAD, &read_size) < 0) {
      perror("record_service error: oti_ioctl FIONREAD");
      break;
    }

    /*
     * Check for underflow: Sun's /dev/audio might stop
     * recording on overflow. This is detected here and
     * the error is cleared. The Osprey doesn't stop
     * recording on overflow.
     */
    oti_audio_get_record_error(odo->Audio_fd, &error);
    if (error){
      error = 0;
      oti_audio_set_record_error(odo->Audio_fd, &error);
    }
    
    if ((read_size == 0))
      break;

    read_size = (read_size < MAX_BUFFER_SIZE) ? read_size : MAX_BUFFER_SIZE;
    
    if (read_size == 0)
      break;

    rtn = oti_read(odo->Audio_fd, odo->aud_data, read_size);
    
    if (rtn < 0) {
      break;
    }

    /* write audio data to file */
    wrtn = write(odo->rec.output_fd, (void*)odo->aud_data, rtn);
    if (wrtn != rtn){
      fprintf(stderr,"record_service error: only wrote %d bytes out of %d bytes\n",wrtn,rtn);
    }
    odo->rec.byte_count += wrtn;

    /* print out hash marks if desired */
    if (odo->rec.hash_amt){
      if (odo->rec.byte_count > odo->rec.hash_cnt){
        printf("#"); fflush(stdout);
        odo->rec.hash_cnt += odo->rec.hash_amt;
      }
    }
    
  } while (0);

  in_record_service = 0;
}

/*------------------------------------------------------------------------
 *
 *  Function:	stop_record
 *
 *  Description:
 *	Stops recording by pausing the audio device and then flushing it.
 *
 *  Parameters: none
 *	
 *  Returns: void
 *	
 *------------------------------------------------------------------------*/

void stop_record(ODO *odo)
{
  /* Pause the device */
  oti_audio_pause_record(odo->Audio_fd);
  oti_audio_flush_record(odo->Audio_fd);
}

/*------------------------------------------------------------------------
 *
 *  Function:	g_exit
 *
 *  Description:
 *	Hanldes signals which would cause process to exit (like Ctl-C).
 *      Stops recording and rewrites the audio file header with the
 *      recorded size. The various file descriptors are closed. This
 *      routine can be called directly.
 *
 *  Parameters:
 *	int signo: signal number
 *	
 *  Returns: void
 *	
 *------------------------------------------------------------------------*/

void g_exit(ODO *odo, int signo)
{
  int rc;

  odo->poll_enabled = 0;
  
  stop_record(odo);
  odo->rec.recording = 0;
  
  if (odo->Audio_fd != -1){
    oti_close(odo->Audio_fd);
    odo->Audio_fd = -1;
  }
  
  if (odo->Audioctl_fd != -1){
    oti_close(odo->Audioctl_fd);
    odo->Audioctl_fd = -1;
  }

  if (odo->rec.output_fd != -1){
    if ((rc=audio_rewrite_filesize(odo->rec.output_fd, odo->rec.byte_count)) \
	!= AUDIO_SUCCESS){
      fprintf(stderr,"g_exit error: not able to write filesize to output file: %d\n",rc);
    }
    
    close(odo->rec.output_fd);
    odo->rec.output_fd = -1;
  }

  exit(0);
}



/*------------------------------------------------------------------------
 *
 *  Function:	poll_handler
 *
 *  Description:
 *	Hanldes SIGPOLL signal. A SIGPOLL is generated by the audio device
 *	when input and output ready conditions change. This is used to play
 *      or record audio data in an async, efficient manner.
 *	
 *  Parameters:
 *	int signo: signal number
 *	
 *  Returns: void
 *	
 *------------------------------------------------------------------------*/

void poll_handler(ODO *odo, int signo)
{
  /* must re-signal immediately in case another signal occurs */
  signal(signo, poll_handler);
  
  /*
   * you can only make oti_xxx calls in signal handler when you
   * know that you are not making such calls anywhere else.
   */
  if (odo->poll_enabled)
    record_service(odo);

}

/*------------------------------------------------------------------------
 *
 *  Function:	get_audioctl_name
 *
 *  Description:
 *	Given an audio device name, determine the control device name
 *	
 *  Parameters:
 *	char* adevice_ctl_name_ptr: ptr to place control device name into
 *      char* adevice_name_ptr:     ptr to audio device name
 *	
 *  Returns: void
 *	
 *  Side Effects:
 *	Modifies buffer pointed to by adevice_ctl_name_ptr. 
 *	
 *  Notes:
 *	adevice_ctl_name_ptr must point to a reasonable sized buffer
 *	
 *------------------------------------------------------------------------*/

void get_audioctl_name(char* adevice_ctl_name_ptr, char* adevice_name_ptr)
{
  int i;
  char  c = '\0';
  char* s;

  /*
   *    Given         Produce      Example:
   *   xxxxaudio -> xxxxxaudioctl  /dev/audio -> /dev/audioctl
   *   xxxxo1kN  -> xxxxo1kctlN    /dev/o1k0  -> /dev/o1kctl0
   */

  strcpy(adevice_ctl_name_ptr, adevice_name_ptr);
  i = strlen(adevice_name_ptr);
  
  if (strstr(adevice_name_ptr,"o1k") != NULL){
    s = &adevice_ctl_name_ptr[i-1];
    c = *s;      
  } else {
    s = &adevice_ctl_name_ptr[i];
  }

  *s++ = 'c';
  *s++ = 't';
  *s++ = 'l';
  *s++ = c;
  *s++ = '\0';

}

 
void main(int argc, char* argv[])
{
  ODO odo;

  int   i;
  int   rc;
  int format   = 0;
  int rate     = 8000;
  int channels = 1;
  
  unsigned int port = AUDIO_LINE_IN;
  unsigned int buffer_size = 1000;
  
  double  igain = 0.125;
  double  mgain = 0.0;

  char  audctl_device[80];
  char info[120];
  char* aud_device = "/dev/o1k0";
  
  Audio_hdr hdr;
  Audio_info tmpinfo;

  odo.Audio_fd    = -1;
  odo.Audioctl_fd = -1;
  odo.poll_enabled = 0;
  odo.rec.output_fd   = -1;
  odo.rec.byte_count = 0;
  odo.rec.recording = 0;
  odo.rec.hash_amt = 0;
  odo.rec.hash_cnt = 0;

  strcpy(info, "");
  
  for (i=1; i<argc; i++){
    switch (argv[i][1])
    {
      case 'D':
      {
        aud_device = argv[++i];
        break;
      }
      case 'f':
      {
        format = atoi(argv[++i]);
        break;
      }
      case 'r':
      {
        rate = atoi(argv[++i]);
        break;
      }
      case 'c':
      {
        channels = atoi(argv[++i]);
        break;
      }
      case 'p':
      {
        port = atoi(argv[++i]);
        break;
      }
      case 'g':
      {
        igain = atof(argv[++i]);
        printf("gain %f\n", igain);
        break;
      }
      case 'm':
      {
        mgain = atof(argv[++i]);
        break;
      }
      case 'i':
      {
        strncpy(info, argv[++i],120);
        break;
      }
      case 'b':
      {
        buffer_size = atoi(argv[++i]);
        break;
      }
      case 'h':
      {
        odo.rec.hash_amt = atoi(argv[++i]);
        odo.rec.hash_cnt = odo.rec.hash_amt;
        break;
      }
      case 'o':
      {
        strcpy(odo.filename, argv[++i]);
        odo.rec.output_fd = open(odo.filename, 
				 O_WRONLY | O_CREAT | O_TRUNC, 0666);
        if (odo.rec.output_fd < 0)
          fprintf(stderr,"could not create %s\n", odo.filename);
        break;
      }
      case '?':
      default:
      {
        usage(argv[0]);
        exit(0);
      }
    }
  }

  if (odo.rec.output_fd < 0){
    usage(argv[0]);
    exit(0);
  }

  /*
   * setup signal handlers
   */
  signal(SIGINT,  g_exit);
  signal(SIGPOLL, poll_handler);

  printf("initializing audio device...\n");

  /*
   * This call must be made before any other
   */
  if ((rc = oti_audio_init(NULL, NULL)) < 0){
    fprintf(stderr,"oti_audio_init failed %d\n",rc);
    g_exit(&odo, 0);
  }
fprintf (stderr, "1:AUD_LINE_IN %d\n", AUDIO_LINE_IN);
  
  /*
   * Open audio device and setup POLL signal
   */
  if ((odo.Audio_fd = oti_open(aud_device, O_RDONLY | O_NDELAY))<0){
    fprintf(stderr,"could not open osprey audio device\n");
    exit(0);
  }
  if (oti_ioctl(odo.Audio_fd, I_SETSIG, (void*)(S_INPUT | S_MSG)) < 0){
    perror("I_SETSIG ioctl");
    g_exit(&odo, 0);
  }
  
  /*
   * Open audio control device and setup POLL signal
   */
  get_audioctl_name(audctl_device, aud_device);
  if ((odo.Audioctl_fd = oti_open(audctl_device, O_RDWR))<0){
    fprintf(stderr,"could not open audio ctl device\n");
    g_exit(&odo, 0);
  }   
  if (oti_ioctl(odo.Audioctl_fd, I_SETSIG, (void*)(S_MSG)) < 0){
    oti_close(odo.Audioctl_fd);
    perror("I_SETSIG ioctl");
    g_exit(&odo, 0);
  }

  /*
   * Set the desired input port and gains
   */
  oti_audio_set_record_port(odo.Audio_fd, &port);
  oti_audio_set_record_gain(odo.Audio_fd, &igain);
  oti_audio_set_monitor_gain(odo.Audio_fd, &mgain);  

  /*
   * Set up the desired format
   */
  switch (format)
  {
    case 0: hdr.encoding = AUDIO_ENCODING_ULAW;   break;
    case 1: hdr.encoding = AUDIO_ENCODING_ALAW;   break;
    case 2: hdr.encoding = AUDIO_ENCODING_LINEAR; break;
    case 3: hdr.encoding = AUDIO_ENCODING_LINEAR; break;
    case 4: hdr.encoding = AUDIO_ENCODING_G722;   break;
    case 5: hdr.encoding = AUDIO_ENCODING_G728;   break;
    default: hdr.encoding = AUDIO_ENCODING_ULAW;
  }

  switch (rate)
  {
    case 8000:  hdr.sample_rate =  8000; break;
    case 11025: hdr.sample_rate = 11025; break;
    case 16000: hdr.sample_rate = 16000; break;
    case 22050: hdr.sample_rate = 22050; break;
    default:    hdr.sample_rate =  8000; 
  }

  switch (channels)
  {
    case 1:  hdr.channels = 1; break;
    case 2:  hdr.channels = 2; break;
    default: hdr.channels = 1; break;
  }

  switch(hdr.encoding)
  {
    case AUDIO_ENCODING_ULAW:
    case AUDIO_ENCODING_ALAW:    
    {
      hdr.samples_per_unit = 1;
      hdr.bytes_per_unit = 1;
      break;
    }
    case AUDIO_ENCODING_LINEAR:
    {
      hdr.samples_per_unit = 1;
      if (format == 2)
        hdr.bytes_per_unit = 1; /* 8 bit PCM */
      else
        hdr.bytes_per_unit = 2; /* 16 bit PCM */
      break;
    }
    case AUDIO_ENCODING_G722:
    {
      /* G722 fixed rate and channels */
      hdr.sample_rate = 8000;
      hdr.channels    = 1;
      hdr.samples_per_unit = 1;
      hdr.bytes_per_unit = 1;
      break;
    }
    case AUDIO_ENCODING_G728:
    {
      /* G728 fixed rate and channels */
      hdr.sample_rate = 1600;
      hdr.channels    = 1;
      hdr.samples_per_unit = 8;
      hdr.bytes_per_unit = 10;
      break;
    }
  }
  
  hdr.data_size = AUDIO_UNKNOWN_SIZE;
  hdr.endian = AUDIO_ENDIAN_BIG;

  /*
   * Set the desired buffer size
   */
  AUDIO_INITINFO(&tmpinfo);
  tmpinfo.record.buffer_size = buffer_size;
  tmpinfo.play.buffer_size = buffer_size;
  oti_audio_setinfo(odo.Audio_fd, &tmpinfo);

  /*
   * Write audio format to file and then set this format
   */
 fprintf(stderr, "Recording: encoding=%d sample_rate=%d samples_per_unit=%d\n",
	 hdr.encoding, hdr.sample_rate, hdr.samples_per_unit);
  printf("           bytes_per_unit=%d channels=%d endian=%d\n",
         hdr.bytes_per_unit, hdr.channels, hdr.endian);
    
  if ((rc = audio_write_filehdr(odo.rec.output_fd, &hdr, info, 
                                ((unsigned)strlen(info)))) != AUDIO_SUCCESS){
    fprintf(stderr,"unable to write audio header %d\n",rc);
    g_exit(&odo, 0);
  }

  oti_audio_set_record_config(odo.Audio_fd, &hdr);
       
  odo.rec.recording  = 1;

  /*
   * Pause the input and flush the read queue
   */
  (void) oti_audio_pause_record(odo.Audio_fd);
  AUDIO_INITINFO(&tmpinfo);
  tmpinfo.record.pause = 0;
  tmpinfo.record.samples = 0;
  tmpinfo.record.error = 0;
  oti_audio_setinfo(odo.Audio_fd, &tmpinfo);
  (void) oti_audio_flush_record(odo.Audio_fd);

  /*
   * Recording happens in the background via async POLL signals
   */
  odo.poll_enabled = 1;  
  while (1){
    poll(NULL,0,5);
  }
}
  
  

