#include "aud.h"


void usage(char* pgm)
{
  fprintf(stderr,"usage: %s [options] -i/o filename\n",pgm);
  fprintf(stderr,"Common Options:                                Defaults\n");
  fprintf(stderr, " <-D device>   : device name                  /dev/o1k0\n");
  fprintf(stderr, " <-b size>     : buffer size                   1000\n");
  fprintf(stderr, "Recording Options:                             \n");  
  fprintf(stderr, " <-c channels> : channels (1,2)                1\n");
  fprintf(stderr, " <-f format>   : encoding formt                0\n");
  fprintf(stderr, "                 (0=ULAW,  1=ALAW, 2=PCM8,\n");
  fprintf(stderr, "                  3=PCM16, 4=G722, 5=G728)\n");
  fprintf(stderr, " <-g gain>     : input gain (0..1.0)           .125\n");
  fprintf(stderr, " <-h amount>   : print hash marks after amount 0\n");  
  fprintf(stderr, " <-m mgain>    : monitor gain (0..1)           0\n");
  fprintf(stderr, " <-p port>     : input port (0=LINE,1=MIC)     0\n");
  fprintf(stderr, " <-r rate>     : sample rate                   8000\n");
  fprintf(stderr, "                 (8000, 11025, 16000, 22050)   \n");
  fprintf(stderr, " <-s info>     : audio file string info        \"\"\n");
  fprintf(stderr, "Playing Options:                               \n");
  fprintf(stderr, " <-v gain>     : output gain (0...1.0)         1.0\n");
}


/* ---------------------------------------------------------------
 *  Function: select_param ()
 *
 *  Definition: selecting optional parameters
 * ---------------------------------------------------------------  */
int select_param (int argc, char **argv)
{
  int i;

  strcpy(info, "");

  for (i=1; i<argc; i++){
    switch (argv[i][1])
      {
      case 'D':
	{
	  aud_device = argv[++i];
	  break;
	}
      case 'f':
	{
	  format = atoi(argv[++i]);
	  break;
	}
      case 'r':
	{
	  rate = atoi(argv[++i]);
	  break;
	}
      case 'c':
	{
	  channels = atoi(argv[++i]);
	  break;
	}
      case 'p':
	{
	  port = atoi(argv[++i]);
	  break;
	}
      case 'g':
	{
	  igain = atof(argv[++i]);
	  fprintf(stderr, "Mic gain %f\n", igain);
	  break;
	}
      case 'm':
	{
	  mgain = atof(argv[++i]);
	  break;
	}
      case 's':
	{
	  strncpy(info, argv[++i],120);
	  break;
	}
      case 'b':
	{
	  buffer_size = atoi(argv[++i]);
	  if (buffer_size > MAX_BUFFER_SIZE){
	    fprintf(stderr,"Desired buffer size %d too large. Using %d\n",
		    buffer_size, MAX_BUFFER_SIZE);  
	    buffer_size = MAX_BUFFER_SIZE;
	  }
	  break;
	}
      case 'h':
	{
	  hash_amt = atoi(argv[++i]);
	  hash_cnt = hash_amt;
	  break;
	}
      case 'i':
	{
	  aud_mode = PLAY;
	  strcpy(filename, argv[++i]);
	  if (audio_isaudiofile(filename)) {
	    input_fd = open(filename, O_RDONLY, 0666);
	    if (input_fd < 0)
	      fprintf(stderr,"could not open %s\n", filename);
	  } 
	  else {
	    fprintf(stderr,"file %s is not an audio file\n", filename);
	  }
	  break;
	}
      case 'o':
	{
	  aud_mode = REC;
	  strcpy(filename, argv[++i]);
	  output_fd = open(filename, O_WRONLY | O_CREAT | O_TRUNC, 0666);
	  if (output_fd < 0)
	    fprintf(stderr,"could not create %s\n", filename);
	  break;
	}
      case 'v':
	{
	  ogain = atof(argv[++i]);
	  break;
	}
      case '?':
      default:
	{
	  usage(argv[0]);
	  exit(0);
	}
      }
  }

  if ((output_fd || input_fd) < 0){
    usage(argv[0]);
    exit(0);
  }

  return 0;
}



/*------------------------------------------------------------------------
 *
 *  Function:	record_service
 *
 *  Description:
 *	This routine is called async from the POLL signal handler.
 *      Care is taken not to enter the routine recursively. Typically
 *      the POLL signal was issued because there is data avail in
 *      the audio device. We read some data from the audio device and
 *      write this audio data to the file. 
 *      If an error occurs (overflow), the error is cleared and we
 *      continue playing.
 *
 *  Parameters: none
 *	
 *  Returns: void
 *	
 *------------------------------------------------------------------------*/

void record_service()
{
  int	read_size;
  int	rtn, wrtn;
  unsigned int error;

  static int in_record_service = 0;

  /*
   * avoid recursive calls
   */
  if (in_record_service != 0){
    return;
  }
  in_record_service = 1;

  if (!recording){
    return;
  }
  
  do {
    /* Determine how much we can read */
    if (oti_ioctl(Audio_fd, FIONREAD, &read_size) < 0) {
      perror("record_service error: oti_ioctl FIONREAD");
      break;
    }

    /*
     * Check for underflow: Sun's /dev/audio might stop
     * recording on overflow. This is detected here and
     * the error is cleared. The Osprey doesn't stop
     * recording on overflow.
     */
    oti_audio_get_record_error(Audio_fd, &error);
    if (error){
      error = 0;
      oti_audio_set_record_error(Audio_fd, &error);
    }
    
    if ((read_size == 0))
      break;

    read_size = (read_size < MAX_BUFFER_SIZE) ? read_size : MAX_BUFFER_SIZE;
    
    if (read_size == 0)
      break;

    rtn = oti_read(Audio_fd, aud_data, read_size);
    
    if (rtn < 0) {
      break;
    }

    /* write audio data to file */
    wrtn = write(output_fd, (void*)aud_data, rtn);
    if (wrtn != rtn){
      fprintf(stderr,"record_service error: only wrote %d bytes out of %d bytes\n",wrtn,rtn);
    }
    byte_count += wrtn;

    /* print out hash marks if desired */
    if (hash_amt){
      if (byte_count > hash_cnt){
        printf("#"); fflush(stdout);
        hash_cnt += hash_amt;
      }
    }
    
  } while (0);

  in_record_service = 0;
}

/*------------------------------------------------------------------------
 *
 *  Function:	stop_record
 *
 *  Description:
 *	Stops recording by pausing the audio device and then flushing it.
 *
 *  Parameters: none
 *	
 *  Returns: void
 *	
 *------------------------------------------------------------------------*/

void stop_record()
{
  /* Pause the device */
  oti_audio_pause_record(Audio_fd);
  oti_audio_flush_record(Audio_fd);
}


/*------------------------------------------------------------------------
 *
 *  Function:	stop_play
 *
 *  Description:
 *	Stops playing by pausing the audio device and then flushing it.
 *
 *  Parameters: none
 *	
 *  Returns: void
 *	
 *------------------------------------------------------------------------*/
void stop_play()
{
  /* Pause the device */
  oti_audio_pause_play(Audio_fd);
  oti_audio_flush_play(Audio_fd);
}


/*------------------------------------------------------------------------
 *
 *  Function:	g_exit
 *
 *  Description:
 *	Hanldes signals which would cause process to exit (like Ctl-C).
 *      Stops recording and rewrites the audio file header with the
 *      recorded size. The various file descriptors are closed. This
 *      routine can be called directly.
 *
 *  Parameters:
 *	int signo: signal number
 *	
 *  Returns: void
 *	
 *------------------------------------------------------------------------*/
void g_exit(int signo)
{
  int rc;

  poll_enabled = 0;
  
  if (aud_mode == PLAY) {
    stop_play();
    playing = 0;
  } else if (aud_mode == REC) {
    stop_record ();
    recording = 0;
  }

  if (Audio_fd != -1){
    oti_close(Audio_fd);
    Audio_fd = -1;
  }
  
  if (Audioctl_fd != -1){
    oti_close(Audioctl_fd);
    Audioctl_fd = -1;
  }
  
  if (aud_mode == PLAY) {
    if (input_fd != -1){
      close(input_fd);
      input_fd = -1;
    }
  }else if (aud_mode == REC) {
    if (output_fd != -1) {
      if((rc = audio_rewrite_filesize(output_fd, byte_count)) != AUDIO_SUCCESS) {
	fprintf (stderr, 
	       "g_exit error: not able to write filesize to output file: %d\n",
		 rc);
      }
      close (output_fd);
      output_fd = -1;
    }
  }
  exit(0);
}

  
/* --------------------------------------------------------------------
 *
 *  Function:	play_service
 *
 *  Description:
 *	This routine is called async from the POLL signal handler.
 *      Care is taken not to enter the routine recursively. Typically
 *      the POLL signal was issued because there is space avail in
 *      the audio device. We read some data from the audio file and
 *      write this audio data to the device. When done playing the
 *      file, an EOF marker is written to the audio device.
 *      If an error occurs (underflow), the error is cleared and we
 *      continue playing.
 *
 *  Parameters: none
 *	
 *  Returns: void
 *	
 * -------------------------------------------------------------------- */
void play_service()
{
  int	start;
  int	end;
  int	outcnt;
  int	rtn;
  unsigned int error;
  
  static int in_play_service = 0;

  /* avoid reentrant code here   */
  if (in_play_service != 0){
    return;
  }
  in_play_service = 1;

  if (input_fd >= 0){
    if (aud_amnt == 0){
      /* fill buffer with data */
      aud_amnt  = read(input_fd, aud_data,
				 buffer_size);
      (char *) aud_ptr = (char *) aud_data;
    }

    /* Check for underflow: Sun's /dev/audio may stop
     * playing on underflow. This is detected here and
     * the error is cleared. The Osprey doesn't stop
     * playing on underflow, silence is played instead.  */
    oti_audio_get_play_error(Audio_fd, &error);
    if (error){
      error = 0;
      oti_audio_set_play_error(Audio_fd, &error);
    }
    
    if (aud_amnt == 0){
      /* done with file: do eof to mark end */
#if 0
      /* this can be done by zero length write: */
      oti_write(Audio_fd, aud_data, 0);
#else
      /* or this can be done by api call */
      oti_audio_play_eof(Audio_fd);
#endif
      close(input_fd);
      input_fd = -1;
    }
    
    while (aud_amnt > 0) {
      /* write as much data as possible */
      rtn = oti_write(Audio_fd, aud_ptr, aud_amnt);
      
      if (rtn > 0) {
        aud_amnt -= rtn;
        aud_ptr += rtn;
      } else {
        break;
      }
    }
  }
  
  in_play_service = 0;
}


/*------------------------------------------------------------------------
 *
 *  Function:	poll_handler
 *
 *  Description:
 *	Hanldes SIGPOLL signal. A SIGPOLL is generated by the audio device
 *	when input and output ready conditions change. This is used to play
 *      or record audio data in an async, efficient manner.
 *	
 *  Parameters:
 *	int signo: signal number
 *	
 *  Returns: void
 *	
 *------------------------------------------------------------------------*/
void poll_handler(int signo)
{
  /* must re-signal immediately in case another signal occurs */
  signal(signo, poll_handler);
  
  /* you can only make oti_xxx calls in signal handler when you
   * know that you are not making such calls anywhere else.     */
  if (poll_enabled) {
    if (aud_mode == PLAY) play_service ();
    else if (aud_mode == REC) record_service();
  }
}


/*------------------------------------------------------------------------
 *
 *  Function:	get_audioctl_name
 *
 *  Description:
 *	Given an audio device name, determine the control device name
 *	
 *  Parameters:
 *	char* adevice_ctl_name_ptr: ptr to place control device name into
 *      char* adevice_name_ptr:     ptr to audio device name
 *	
 *  Returns: void
 *	
 *  Side Effects:
 *	Modifies buffer pointed to by adevice_ctl_name_ptr. 
 *	
 *  Notes:
 *	adevice_ctl_name_ptr must point to a reasonable sized buffer
 *	
 *------------------------------------------------------------------------*/

void get_audioctl_name(char* adevice_ctl_name_ptr, char* adevice_name_ptr)
{
  int i;
  char  c = '\0';
  char* s;

  /*
   *    Given         Produce      Example:
   *   xxxxaudio -> xxxxxaudioctl  /dev/audio -> /dev/audioctl
   *   xxxxo1kN  -> xxxxo1kctlN    /dev/o1k0  -> /dev/o1kctl0
   */

  strcpy(adevice_ctl_name_ptr, adevice_name_ptr);
  i = strlen(adevice_name_ptr);
  
  if (strstr(adevice_name_ptr,"o1k") != NULL){
    s = &adevice_ctl_name_ptr[i-1];
    c = *s;      
  } else {
    s = &adevice_ctl_name_ptr[i];
  }

  *s++ = 'c';
  *s++ = 't';
  *s++ = 'l';
  *s++ = c;
  *s++ = '\0';

}

 
void rec_main(int argc, char* argv[])
{
  int   i;
  int   rc;


  strcpy(info, "");
  
  for (i=1; i<argc; i++){
    switch (argv[i][1])
    {
      case 'D':
      {
        aud_device = argv[++i];
        break;
      }
      case 'f':
      {
        format = atoi(argv[++i]);
        break;
      }
      case 'r':
      {
        rate = atoi(argv[++i]);
        break;
      }
      case 'c':
      {
        channels = atoi(argv[++i]);
        break;
      }
      case 'p':
      {
        port = atoi(argv[++i]);
        break;
      }
      case 'g':
      {
        igain = atof(argv[++i]);
        printf("gain %f\n", igain);
        break;
      }
      case 'm':
      {
        mgain = atof(argv[++i]);
        break;
      }
      case 'i':
      {
        strncpy(info, argv[++i],120);
        break;
      }
      case 'b':
      {
        buffer_size = atoi(argv[++i]);
        break;
      }
      case 'h':
      {
        hash_amt = atoi(argv[++i]);
        hash_cnt = hash_amt;
        break;
      }
      case 'o':
      {
        strcpy(filename, argv[++i]);
        output_fd = open(filename, O_WRONLY | O_CREAT | O_TRUNC, 0666);
        if (output_fd < 0)
          fprintf(stderr,"could not create %s\n", filename);
        break;
      }
      case '?':
      default:
      {
        usage(argv[0]);
        exit(0);
      }
    }
  }

  if (output_fd < 0){
    usage(argv[0]);
    exit(0);
  }

  /*
   * setup signal handlers
   */
  signal(SIGINT,  g_exit);
  signal(SIGPOLL, poll_handler);

  printf("initializing audio device...\n");

  /*
   * This call must be made before any other
   */
  if ((rc = oti_audio_init(NULL, NULL)) < 0){
    fprintf(stderr,"oti_audio_init failed %d\n",rc);
    g_exit(0);
  }
fprintf (stderr, "1:AUD_LINE_IN %d\n", AUDIO_LINE_IN);
  
  /*
   * Open audio device and setup POLL signal
   */
  if ((Audio_fd = oti_open(aud_device, O_RDONLY | O_NDELAY))<0){
    fprintf(stderr,"could not open osprey audio device\n");
    exit(0);
  }
  if (oti_ioctl(Audio_fd, I_SETSIG, (void*)(S_INPUT | S_MSG)) < 0){
    perror("I_SETSIG ioctl");
    g_exit(0);
  }
  
  /*
   * Open audio control device and setup POLL signal
   */
  get_audioctl_name(audctl_device, aud_device);
  if ((Audioctl_fd = oti_open(audctl_device, O_RDWR))<0){
    fprintf(stderr,"could not open audio ctl device\n");
    g_exit(0);
  }   
  if (oti_ioctl(Audioctl_fd, I_SETSIG, (void*)(S_MSG)) < 0){
    oti_close(Audioctl_fd);
    perror("I_SETSIG ioctl");
    g_exit(0);
  }

  /*
   * Set the desired input port and gains
   */
  oti_audio_set_record_port(Audio_fd, &port);
  oti_audio_set_record_gain(Audio_fd, &igain);
  oti_audio_set_monitor_gain(Audio_fd, &mgain);  

  /*
   * Set up the desired format
   */
  switch (format)
  {
    case 0: hdr.encoding = AUDIO_ENCODING_ULAW;   break;
    case 1: hdr.encoding = AUDIO_ENCODING_ALAW;   break;
    case 2: hdr.encoding = AUDIO_ENCODING_LINEAR; break;
    case 3: hdr.encoding = AUDIO_ENCODING_LINEAR; break;
    case 4: hdr.encoding = AUDIO_ENCODING_G722;   break;
    case 5: hdr.encoding = AUDIO_ENCODING_G728;   break;
    default: hdr.encoding = AUDIO_ENCODING_ULAW;
  }

  switch (rate)
  {
    case 8000:  hdr.sample_rate =  8000; break;
    case 11025: hdr.sample_rate = 11025; break;
    case 16000: hdr.sample_rate = 16000; break;
    case 22050: hdr.sample_rate = 22050; break;
    default:    hdr.sample_rate =  8000; 
  }

  switch (channels)
  {
    case 1:  hdr.channels = 1; break;
    case 2:  hdr.channels = 2; break;
    default: hdr.channels = 1; break;
  }

  switch(hdr.encoding)
  {
    case AUDIO_ENCODING_ULAW:
    case AUDIO_ENCODING_ALAW:    
    {
      hdr.samples_per_unit = 1;
      hdr.bytes_per_unit = 1;
      break;
    }
    case AUDIO_ENCODING_LINEAR:
    {
      hdr.samples_per_unit = 1;
      if (format == 2)
        hdr.bytes_per_unit = 1; /* 8 bit PCM */
      else
        hdr.bytes_per_unit = 2; /* 16 bit PCM */
      break;
    }
    case AUDIO_ENCODING_G722:
    {
      /* G722 fixed rate and channels */
      hdr.sample_rate = 8000;
      hdr.channels    = 1;
      hdr.samples_per_unit = 1;
      hdr.bytes_per_unit = 1;
      break;
    }
    case AUDIO_ENCODING_G728:
    {
      /* G728 fixed rate and channels */
      hdr.sample_rate = 1600;
      hdr.channels    = 1;
      hdr.samples_per_unit = 8;
      hdr.bytes_per_unit = 10;
      break;
    }
  }
  
  hdr.data_size = AUDIO_UNKNOWN_SIZE;
  hdr.endian = AUDIO_ENDIAN_BIG;

  /*
   * Set the desired buffer size
   */
  AUDIO_INITINFO(&tmpinfo);
  tmpinfo.record.buffer_size = buffer_size;
  tmpinfo.play.buffer_size = buffer_size;
  oti_audio_setinfo(Audio_fd, &tmpinfo);

  /*
   * Write audio format to file and then set this format
   */
 fprintf(stderr, "Recording: encoding=%d sample_rate=%d samples_per_unit=%d\n",
	 hdr.encoding, hdr.sample_rate, hdr.samples_per_unit);
  printf("           bytes_per_unit=%d channels=%d endian=%d\n",
         hdr.bytes_per_unit, hdr.channels, hdr.endian);
    
  if ((rc = audio_write_filehdr(output_fd, &hdr, info, 
                                ((unsigned)strlen(info)))) != AUDIO_SUCCESS){
    fprintf(stderr,"unable to write audio header %d\n",rc);
    g_exit(0);
  }

  oti_audio_set_record_config(Audio_fd, &hdr);
       
  recording  = 1;

  /*
   * Pause the input and flush the read queue
   */
  (void) oti_audio_pause_record(Audio_fd);
  AUDIO_INITINFO(&tmpinfo);
  tmpinfo.record.pause = 0;
  tmpinfo.record.samples = 0;
  tmpinfo.record.error = 0;
  oti_audio_setinfo(Audio_fd, &tmpinfo);
  (void) oti_audio_flush_record(Audio_fd);

  /*
   * Recording happens in the background via async POLL signals
   */
  poll_enabled = 1;  
  while (1){
    poll(NULL,0,5);
  }
}
  

/* -----------------------------------------------------------------
 *
 *
 *
 * ----------------------------------------------------------------- */
int odo_play ()
{
  int      rc;

  char     audctl_device[80];

  unsigned int value;
  
  Audio_hdr hdr;
  Audio_info tmpinfo;

  /* Open audio device and setup POLL signal */  
  if ((Audio_fd = oti_open(aud_device, O_WRONLY | O_NDELAY))<0){
    fprintf(stderr,"could not open osprey audio device\n");
    g_exit(0);
  }
  if (oti_ioctl(Audio_fd, I_SETSIG, (void*)(S_OUTPUT | S_MSG)) < 0){
    perror("I_SETSIG ioctl");
    g_exit(0);
  }
  
  /* Open audio control device and setup POLL signal  */
  get_audioctl_name(audctl_device, aud_device);
  if ((Audioctl_fd = oti_open(audctl_device, O_RDWR))<0){
    fprintf(stderr,"could not open audio ctl device\n");
    g_exit(0);
  }    
  if (oti_ioctl(Audioctl_fd, I_SETSIG, (void*)(S_MSG)) < 0){
    perror("I_SETSIG ioctl");
    g_exit(0);
  }

  /* Set the desired gains */
  oti_audio_set_play_gain(Audio_fd, &ogain);
  oti_audio_set_monitor_gain(Audio_fd, &mgain);

  /* Set the desired buffer size  */
  AUDIO_INITINFO(&tmpinfo);  
  tmpinfo.record.buffer_size = buffer_size;
  tmpinfo.play.buffer_size = buffer_size;
  tmpinfo.play.pause = 0;
  tmpinfo.play.samples = 0;
  tmpinfo.play.error = 0;
  oti_audio_setinfo(Audio_fd, &tmpinfo);
  
  /* Get the audio file format  */
  if ((rc = audio_read_filehdr(input_fd, &hdr, info, 
			       strlen(info))) != AUDIO_SUCCESS){
    oti_close(Audio_fd);
    fprintf(stderr,"unable to read audio header from %s\n", filename);
    g_exit(0);
  }
  
  printf("playing file %s %s\n", filename, info);
  
  printf("   encoding=%d sample_rate=%d samples_per_unit=%d\n",
         hdr.encoding, hdr.sample_rate, hdr.samples_per_unit);
  printf("   bytes_per_unit=%d channels=%d endian=%d\n",
         hdr.bytes_per_unit, hdr.channels, hdr.endian);

  /* using audio_write_filehdr always seems to nuke endian select */
  hdr.endian = AUDIO_ENDIAN_BIG;
  oti_audio_set_play_config(Audio_fd, &hdr);

  value = 0;
  oti_audio_set_play_eof(Audio_fd, &value);
  
  poll_enabled = 1;
  playing  = 1;
fprintf (stderr, "play.1: Audio_fd = %f\n", Audio_fd);
  /* SIGPOLL kicks off play_service() */
  (void) kill(getpid(), SIGPOLL);
fprintf (stderr, "play.2: Audio_fd = %f\n", Audio_fd);
  /* Playing happens in the background via async POLL signals.
   * Wait here for EOF to be processed  */
  while (playing){
    poll(NULL,0,5);
    poll_enabled = 0;
    oti_audio_get_play_eof(Audio_fd, &value);
    poll_enabled = 1;    
    if (value > 0)
      playing = 0;
  }

  /* Wait for drain  */
  poll(NULL,0,20);
  
  g_exit(0);
  
  return 0;
}
  

/* ----------------------------------------------------------------------
 * Function   : odo_set_param ()
 * Definition : set and switch varables.
 * ---------------------------------------------------------------------- */
int odo_set_param ()
{

  /* Set up the desired format  */
  switch (format)
  {
    case 0: hdr.encoding = AUDIO_ENCODING_ULAW;   break;
    case 1: hdr.encoding = AUDIO_ENCODING_ALAW;   break;
    case 2: hdr.encoding = AUDIO_ENCODING_LINEAR; break;
    case 3: hdr.encoding = AUDIO_ENCODING_LINEAR; break;
    case 4: hdr.encoding = AUDIO_ENCODING_G722;   break;
    case 5: hdr.encoding = AUDIO_ENCODING_G728;   break;
    default:hdr.encoding = AUDIO_ENCODING_ULAW;
  }

  switch (rate)
  {
    case 8000:  hdr.sample_rate =  8000; break;
    case 11025: hdr.sample_rate = 11025; break;
    case 16000: hdr.sample_rate = 16000; break;
    case 22050: hdr.sample_rate = 22050; break;
    default:    hdr.sample_rate =  8000; 
  }

  switch (channels)
  {
    case 1:  hdr.channels = 1; break;
    case 2:  hdr.channels = 2; break;
    default: hdr.channels = 1; break;
  }

  switch(hdr.encoding)
  {
    case AUDIO_ENCODING_ULAW:
    case AUDIO_ENCODING_ALAW:    
    {
      hdr.samples_per_unit = 1;
      hdr.bytes_per_unit = 1;
      break;
    }
    case AUDIO_ENCODING_LINEAR:
    {
      hdr.samples_per_unit = 1;
      if (format == 2)
        hdr.bytes_per_unit = 1; /* 8 bit PCM */
      else
        hdr.bytes_per_unit = 2; /* 16 bit PCM */
      break;
    }
    case AUDIO_ENCODING_G722:
    {
      /* G722 fixed rate and channels */
      hdr.sample_rate = 8000;
      hdr.channels    = 1;
      hdr.samples_per_unit = 1;
      hdr.bytes_per_unit = 1;
      break;
    }
    case AUDIO_ENCODING_G728:
    {
      /* G728 fixed rate and channels */
      hdr.sample_rate = 1600;
      hdr.channels    = 1;
      hdr.samples_per_unit = 8;
      hdr.bytes_per_unit = 10;
      break;
    }
  }
  
  hdr.data_size = AUDIO_UNKNOWN_SIZE;
  hdr.endian = AUDIO_ENDIAN_BIG;

  /* Set the desired buffer size  */
  AUDIO_INITINFO(&tmpinfo);
  tmpinfo.record.buffer_size = buffer_size;
  tmpinfo.play.buffer_size = buffer_size;
  oti_audio_setinfo(Audio_fd, &tmpinfo);

return 0;
}

  

/* ----------------------------------------------------------------------
 * Function   : odo
 * Definition :
 * ---------------------------------------------------------------------- */

int odo_record ()
{
  int rc;

  /* Open audio device and setup POLL signal  */
  if((Audio_fd = oti_open(aud_device, O_RDONLY | O_NDELAY)) < 0){
    fprintf(stderr,"could not open osprey audio device\n");
    exit(0);
  }
  if (oti_ioctl(Audio_fd, I_SETSIG, (void*)(S_INPUT | S_MSG)) < 0){
    perror("I_SETSIG ioctl");
    g_exit(0);
  }
  
  /* Open audio control device and setup POLL signal  */
  get_audioctl_name(audctl_device, aud_device);
  if ((Audioctl_fd = oti_open(audctl_device, O_RDWR))<0){
    fprintf(stderr,"could not open audio ctl device\n");
    g_exit(0);
  }   
  if (oti_ioctl(Audioctl_fd, I_SETSIG, (void*)(S_MSG)) < 0){
    oti_close(Audioctl_fd);
    perror("I_SETSIG ioctl");
    g_exit(0);
  }

fprintf(stderr,"1:\n");
  /* Set the desired input port and gains   */
  oti_audio_set_record_port(Audio_fd, &port);
fprintf(stderr,"2:\n");
  oti_audio_set_record_gain(Audio_fd, &igain);
fprintf(stderr,"3:\n");
  oti_audio_set_monitor_gain(Audio_fd, &mgain);  
fprintf(stderr,"4:\n");

  if (odo_set_param()) {
    fprintf(stderr, "Unable to configure parameters\n");
    exit (1);
  }

  /* Write audio format to file and then set this format   */
  printf("Recording: encoding=%d sample_rate=%d samples_per_unit=%d\n",
	 hdr.encoding, hdr.sample_rate, hdr.samples_per_unit);
  printf("           bytes_per_unit=%d channels=%d endian=%d\n",
	 hdr.bytes_per_unit, hdr.channels, hdr.endian);
    
  if ((rc = audio_write_filehdr(output_fd, &hdr, info, 
			  ((unsigned)strlen(info)))) != AUDIO_SUCCESS){
    fprintf(stderr,"unable to write audio header %d\n",rc);
    g_exit(0);
  }

  oti_audio_set_record_config(Audio_fd, &hdr);
       
  recording  = 1;

  /* Pause the input and flush the read queue  */
  (void) oti_audio_pause_record(Audio_fd);
  AUDIO_INITINFO(&tmpinfo);
  tmpinfo.record.pause = 0;
  tmpinfo.record.samples = 0;
  tmpinfo.record.error = 0;
  oti_audio_setinfo(Audio_fd, &tmpinfo);
  (void) oti_audio_flush_record(Audio_fd);

  /* Recording happens in the background via async POLL signals   */
  poll_enabled = 1;  
  while (1){
    poll(NULL,0,5);
  }

  return 0;
}
  


void main (int argc, char **argv)
{
  int rc;

  /* Select optional parameters */
  if (select_param (argc, argv)) {
    fprintf (stderr, "Unable to read parameter options\n");
    exit (1);
  }

  /* setup signal handlers  */
  signal(SIGINT,  g_exit);
  signal(SIGPOLL, poll_handler);

  fprintf(stderr, "initializing audio device...\n");

  /* This call must be made before any other   */
  if ((rc = oti_audio_init(NULL, NULL)) < 0){
    fprintf(stderr,"oti_audio_init failed %d\n",rc);
    g_exit(0);
  }

  if (aud_mode == PLAY) {
    if (odo_play ()) {
      fprintf (stderr, "Unable to play audio sound\n");
      exit (1);
    }
  } /* end of audio playing mode */
  else if (aud_mode == REC) {
    if (odo_record ()) {
      fprintf (stderr, "Unable to record audio sound\n");
      exit (1);
    }
  } /* end of audio recording mode */
}
