

#include "odo.h"

void usage(char* pgm)
{
  fprintf(stderr,"usage: %s [options] -i/o filename\n",pgm);
  fprintf(stderr,"Common Options:                                Defaults\n");
  fprintf(stderr, " <-D device>   : device name                  /dev/o1k0\n");
  fprintf(stderr, " <-b size>     : buffer size                   1000\n");
  fprintf(stderr, "Recording Options:                             \n");  
  fprintf(stderr, " <-c channels> : channels (1,2)                1\n");
  fprintf(stderr, " <-f format>   : encoding formt                0\n");
  fprintf(stderr, "                 (0=ULAW,  1=ALAW, 2=PCM8,\n");
  fprintf(stderr, "                  3=PCM16, 4=G722, 5=G728)\n");
  fprintf(stderr, " <-g gain>     : input gain (0..1.0)           .125\n");
  fprintf(stderr, " <-h amount>   : print hash marks after amount 0\n");  
  fprintf(stderr, " <-m mgain>    : monitor gain (0..1)           0\n");
  fprintf(stderr, " <-p port>     : input port (0=LINE,1=MIC)     0\n");
  fprintf(stderr, " <-r rate>     : sample rate                   8000\n");
  fprintf(stderr, "                 (8000, 11025, 16000, 22050)   \n");
  fprintf(stderr, " <-s info>     : audio file string info        \"\"\n");
  fprintf(stderr, "Playing Options:                               \n");
  fprintf(stderr, " <-v gain>     : output gain (0...1.0)         1.0\n");
}


/* ---------------------------------------------------------------
 *  Function: select_param ()
 *
 *  Definition: selecting optional parameters
 * ---------------------------------------------------------------  */
int select_param (int argc, char **argv, ODO *odo)
{
  int i;

  strcpy(odo->info, "");

  for (i=1; i<argc; i++){
    switch (argv[i][1])
      {
      case 'D':
	{
	  odo->dev.aud_device = argv[++i];
	  break;
	}
      case 'f':
	{
	  odo->dev.format = atoi(argv[++i]);
	  break;
	}
      case 'r':
	{
	  odo->dev.rate = atoi(argv[++i]);
	  break;
	}
      case 'c':
	{
	  odo->dev.channels = atoi(argv[++i]);
	  break;
	}
      case 'p':
	{
	  odo->dev.port = atoi(argv[++i]);
	  break;
	}
      case 'g':
	{
	  odo->dev.igain = atof(argv[++i]);
	  fprintf(stderr, "Mic gain %f\n", odo->dev.igain);
	  break;
	}
      case 'm':
	{
	  odo->dev.mgain = atof(argv[++i]);
	  break;
	}
      case 's':
	{
	  strncpy(odo->info, argv[++i],120);
	  break;
	}
      case 'b':
	{
	  odo->buf_sz = atoi(argv[++i]);
	  if (odo->buf_sz > MAX_BUFFER_SIZE){
	    fprintf(stderr,"Desired buffer size %d too large. Using %d\n",
		    odo->buf_sz, MAX_BUFFER_SIZE);  
	    odo->buf_sz = MAX_BUFFER_SIZE;
	  }
	  break;
	}
      case 'h':
	{
	  odo->rec.hash_amt = atoi(argv[++i]);
	  odo->rec.hash_cnt = odo->rec.hash_amt;
	  break;
	}
      case 'i':
	{
	  odo->aud_mode = PLAY;
	  strcpy(odo->filename, argv[++i]);
	  if (audio_isaudiofile(odo->filename)) {
	    odo->play.input_fd = open(odo->filename, O_RDONLY, 0666);
	    if (odo->play.input_fd < 0)
	      fprintf(stderr,"could not open %s\n", odo->filename);
	  } 
	  else {
	    fprintf(stderr,"file %s is not an audio file\n", odo->filename);
	  }
	  break;
	}
      case 'o':
	{
	  odo->aud_mode = REC;
	  strcpy(odo->filename, argv[++i]);
	  odo->rec.output_fd = open(odo->filename, 
				   O_WRONLY | O_CREAT | O_TRUNC, 0666);
	  if (odo->rec.output_fd < 0)
	    fprintf(stderr,"could not create %s\n", odo->filename);
	  break;
	}
      case 'v':
	{
	  odo->dev.ogain = atof(argv[++i]);
	  break;
	}
      case '?':
      default:
	{
	  usage(argv[0]);
	  exit(0);
	}
      }
  }

  if ((odo->rec.output_fd || odo->play.input_fd) < 0){
    usage(argv[0]);
    exit(0);
  }

  return 0;
}


/*------------------------------------------------------------------------
 *
 *  Function:	record_service
 *
 *  Description:
 *	This routine is called async from the POLL signal handler.
 *      Care is taken not to enter the routine recursively. Typically
 *      the POLL signal was issued because there is data avail in
 *      the audio device. We read some data from the audio device and
 *      write this audio data to the file. 
 *      If an error occurs (overflow), the error is cleared and we
 *      continue playing.
 *
 *  Parameters: none
 *	
 *  Returns: void
 *	
 *------------------------------------------------------------------------*/

void record_service(ODO *odo)
{
  int	read_size;
  int	rtn, wrtn;
  unsigned int error;

  static int in_record_service = 0;

  /* avoid recursive calls  */
  if (in_record_service != 0){
    return;
  }
  in_record_service = 1;

  if (!odo->rec.recording) {
    return;
  }
  
  do {
    /* Determine how much we can read */
    if (oti_ioctl(odo->aud_fd, FIONREAD, &read_size) < 0) {
      perror("record_service error: oti_ioctl FIONREAD");
      break;
    }

    /* Check for underflow: Sun's /dev/audio might stop
     * recording on overflow. This is detected here and
     * the error is cleared. The Osprey doesn't stop
     * recording on overflow.                            */
    oti_audio_get_record_error(odo->aud_fd, &error);
    if (error){
      error = 0;
      oti_audio_set_record_error(odo->aud_fd, &error);
    }
    
    if ((read_size == 0))
      break;

    read_size = (read_size < MAX_BUFFER_SIZE) ? read_size : MAX_BUFFER_SIZE;
    
    if (read_size == 0)
      break;

    rtn = oti_read(odo->aud_fd, odo->aud_data, read_size);
    
    if (rtn < 0) {
      break;
    }

    /* write audio data to file */
    wrtn = write(odo->rec.output_fd, (void*)odo->aud_data, rtn);
    if (wrtn != rtn){
      fprintf(stderr,
	      "record_service error: only wrote %d bytes out of %d bytes\n",
	      wrtn, rtn);
    }
    odo->rec.byte_count += wrtn;

    /* print out hash marks if desired */
    if (odo->rec.hash_amt){
      if (odo->rec.byte_count > odo->rec.hash_cnt){
        printf("#"); 
	fflush(stdout);
        odo->rec.hash_cnt += odo->rec.hash_amt;
      }
    }
    
  } while (0);

  in_record_service = 0;
}


/*------------------------------------------------------------------------
 *
 *  Function:	stop_record
 *
 *  Description:
 *	Stops recording by pausing the audio device and then flushing it.
 *
 *  Parameters: none
 *	
 *  Returns: void
 *	
 *------------------------------------------------------------------------*/
void stop_record(ODO *odo)
{
  /* Pause the device */
  oti_audio_pause_record(odo->aud_fd);
  oti_audio_flush_record(odo->aud_fd);
}


/*------------------------------------------------------------------------
 *
 *  Function:	poll_handler
 *
 *  Description:
 *	Hanldes SIGPOLL signal. A SIGPOLL is generated by the audio device
 *	when input and output ready conditions change. This is used to play
 *      or record audio data in an async, efficient manner.
 *	
 *  Parameters:
 *	int signo: signal number
 *	
 *  Returns: void
 *	
 *------------------------------------------------------------------------*/
void poll_handler(ODO *odo, int signo)
{
  /* must re-signal immediately in case another signal occurs */
  signal(signo, poll_handler);
  
  /* you can only make oti_xxx calls in signal handler when you
   * know that you are not making such calls anywhere else.     */
  if (odo->poll_enabled) {
    if (odo->aud_mode == PLAY) play_service (odo);
    else if (odo->aud_mode == REC) record_service(odo);
  }
}


/*------------------------------------------------------------------------
 *
 *  Function:	get_audioctl_name
 *
 *  Description:
 *	Given an audio device name, determine the control device name
 *	
 *  Parameters:
 *	char* adevice_ctl_name_ptr: ptr to place control device name into
 *      char* adevice_name_ptr:     ptr to audio device name
 *	
 *  Returns: void
 *	
 *  Side Effects:
 *	Modifies buffer pointed to by adevice_ctl_name_ptr. 
 *	
 *  Notes:
 *	adevice_ctl_name_ptr must point to a reasonable sized buffer
 *	
 *------------------------------------------------------------------------*/

void get_audioctl_name(char* adevice_ctl_name_ptr, char* adevice_name_ptr)
{
  int i;
  char  c = '\0';
  char* s;

  /*
   *    Given         Produce      Example:
   *   xxxxaudio -> xxxxxaudioctl  /dev/audio -> /dev/audioctl
   *   xxxxo1kN  -> xxxxo1kctlN    /dev/o1k0  -> /dev/o1kctl0
   */

  strcpy(adevice_ctl_name_ptr, adevice_name_ptr);
  i = strlen(adevice_name_ptr);
  
  if (strstr(adevice_name_ptr,"o1k") != NULL){
    s = &adevice_ctl_name_ptr[i-1];
    c = *s;      
  } else {
    s = &adevice_ctl_name_ptr[i];
  }

  *s++ = 'c';
  *s++ = 't';
  *s++ = 'l';
  *s++ = c;
  *s++ = '\0';

}


/* ----------------------------------------------------------------------
 * Function   : odo_set_param ()
 * Definition : set and switch varables.
 * ---------------------------------------------------------------------- */
int odo_set_param (ODO *odo)
{

  /* Set up the desired format  */
  switch (odo->dev.format)
  {
    case 0: odo->hdr.encoding = AUDIO_ENCODING_ULAW;   break;
    case 1: odo->hdr.encoding = AUDIO_ENCODING_ALAW;   break;
    case 2: odo->hdr.encoding = AUDIO_ENCODING_LINEAR; break;
    case 3: odo->hdr.encoding = AUDIO_ENCODING_LINEAR; break;
    case 4: odo->hdr.encoding = AUDIO_ENCODING_G722;   break;
    case 5: odo->hdr.encoding = AUDIO_ENCODING_G728;   break;
    default: odo->hdr.encoding = AUDIO_ENCODING_ULAW;
  }

  switch (odo->dev.rate)
  {
    case 8000:  odo->hdr.sample_rate =  8000; break;
    case 11025: odo->hdr.sample_rate = 11025; break;
    case 16000: odo->hdr.sample_rate = 16000; break;
    case 22050: odo->hdr.sample_rate = 22050; break;
    default:    odo->hdr.sample_rate =  8000; 
  }

  switch (odo->dev.channels)
  {
    case 1:  odo->hdr.channels = 1; break;
    case 2:  odo->hdr.channels = 2; break;
    default: odo->hdr.channels = 1; break;
  }

  switch(odo->hdr.encoding)
  {
    case AUDIO_ENCODING_ULAW:
    case AUDIO_ENCODING_ALAW:    
    {
      odo->hdr.samples_per_unit = 1;
      odo->hdr.bytes_per_unit = 1;
      break;
    }
    case AUDIO_ENCODING_LINEAR:
    {
      odo->hdr.samples_per_unit = 1;
      if (odo->dev.format == 2)
        odo->hdr.bytes_per_unit = 1; /* 8 bit PCM */
      else
        odo->hdr.bytes_per_unit = 2; /* 16 bit PCM */
      break;
    }
    case AUDIO_ENCODING_G722:
    {
      /* G722 fixed rate and channels */
      odo->hdr.sample_rate = 8000;
      odo->hdr.channels    = 1;
      odo->hdr.samples_per_unit = 1;
      odo->hdr.bytes_per_unit = 1;
      break;
    }
    case AUDIO_ENCODING_G728:
    {
      /* G728 fixed rate and channels */
      odo->hdr.sample_rate = 1600;
      odo->hdr.channels    = 1;
      odo->hdr.samples_per_unit = 8;
      odo->hdr.bytes_per_unit = 10;
      break;
    }
  }
  
  odo->hdr.data_size = AUDIO_UNKNOWN_SIZE;
  odo->hdr.endian = AUDIO_ENDIAN_BIG;

  /* Set the desired buffer size  */
  AUDIO_INITINFO(&odo->tmpinfo);
  odo->tmpinfo.record.buffer_size = odo->buf_sz;
  odo->tmpinfo.play.buffer_size = odo->buf_sz;
  oti_audio_setinfo(odo->aud_fd, &odo->tmpinfo);

return 0;
}


/* ----------------------------------------------------------------------
 * Function   : odo
 * Definition :
 * ---------------------------------------------------------------------- */

int odo_record (ODO *odo)
{
  int rc;

  /* Open audio device and setup POLL signal  */
  if((odo->aud_fd = oti_open(odo->dev.aud_device, O_RDONLY | O_NDELAY)) < 0){
    fprintf(stderr,"could not open osprey audio device\n");
    exit(0);
  }
  if (oti_ioctl(odo->aud_fd, I_SETSIG, (void*)(S_INPUT | S_MSG)) < 0){
    perror("I_SETSIG ioctl");
    g_exit(odo, 0);
  }
  
  /* Open audio control device and setup POLL signal  */
  get_audioctl_name(odo->dev.audctl_device, odo->dev.aud_device);
  if ((odo->audctl_fd = oti_open(odo->dev.audctl_device, O_RDWR))<0){
    fprintf(stderr,"could not open audio ctl device\n");
    g_exit(odo, 0);
  }   
  if (oti_ioctl(odo->audctl_fd, I_SETSIG, (void*)(S_MSG)) < 0){
    oti_close(odo->audctl_fd);
    perror("I_SETSIG ioctl");
    g_exit(odo, 0);
  }

fprintf(stderr,"1:\n");
  /* Set the desired input port and gains   */
  oti_audio_set_record_port(odo->aud_fd, &(odo->dev.port));
fprintf(stderr,"2:\n");
  oti_audio_set_record_gain(odo->aud_fd, &(odo->dev.igain));
fprintf(stderr,"3:\n");
  oti_audio_set_monitor_gain(odo->aud_fd, &(odo->dev.mgain));  
fprintf(stderr,"4:\n");

  if (odo_set_param(odo)) {
    fprintf(stderr, "Unable to configure parameters\n");
    exit (1);
  }

  /* Write audio format to file and then set this format   */
  printf("Recording: encoding=%d sample_rate=%d samples_per_unit=%d\n",
	 odo->hdr.encoding, odo->hdr.sample_rate, odo->hdr.samples_per_unit);
  printf("           bytes_per_unit=%d channels=%d endian=%d\n",
	 odo->hdr.bytes_per_unit, odo->hdr.channels, odo->hdr.endian);
    
  if ((rc = audio_write_filehdr(odo->rec.output_fd, &odo->hdr, odo->info, 
			  ((unsigned)strlen(odo->info)))) != AUDIO_SUCCESS){
    fprintf(stderr,"unable to write audio header %d\n",rc);
    g_exit(odo, 0);
  }

  oti_audio_set_record_config(odo->aud_fd, &odo->hdr);
       
  odo->rec.recording  = 1;

  /* Pause the input and flush the read queue  */
  (void) oti_audio_pause_record(odo->aud_fd);
  AUDIO_INITINFO(&odo->tmpinfo);
  odo->tmpinfo.record.pause = 0;
  odo->tmpinfo.record.samples = 0;
  odo->tmpinfo.record.error = 0;
  oti_audio_setinfo(odo->aud_fd, &odo->tmpinfo);
  (void) oti_audio_flush_record(odo->aud_fd);

  /* Recording happens in the background via async POLL signals   */
  odo->poll_enabled = 1;  
  while (1){
    poll(NULL,0,5);
  }

  return 0;
}
  
  
/* --------------------------------------------------------------------
 *
 *  Function:	play_service
 *
 *  Description:
 *	This routine is called async from the POLL signal handler.
 *      Care is taken not to enter the routine recursively. Typically
 *      the POLL signal was issued because there is space avail in
 *      the audio device. We read some data from the audio file and
 *      write this audio data to the device. When done playing the
 *      file, an EOF marker is written to the audio device.
 *      If an error occurs (underflow), the error is cleared and we
 *      continue playing.
 *
 *  Parameters: none
 *	
 *  Returns: void
 *	
 * -------------------------------------------------------------------- */
void play_service(ODO *odo)
{
  int	start;
  int	end;
  int	outcnt;
  int	rtn;
  unsigned int error;
  
  static int in_play_service = 0;

  /* avoid reentrant code here   */
  if (in_play_service != 0){
    return;
  }
  in_play_service = 1;

  if (odo->play.input_fd >= 0){
    if (odo->play.aud_amnt == 0){
      /* fill buffer with data */
      odo->play.aud_amnt  = read(odo->play.input_fd, odo->aud_data,
				 odo->buf_sz);
      (char *) odo->play.aud_ptr = (char *) odo->aud_data;
    }

    /* Check for underflow: Sun's /dev/audio may stop
     * playing on underflow. This is detected here and
     * the error is cleared. The Osprey doesn't stop
     * playing on underflow, silence is played instead.  */
    oti_audio_get_play_error(odo->aud_fd, &error);
    if (error){
      error = 0;
      oti_audio_set_play_error(odo->aud_fd, &error);
    }
    
    if (odo->play.aud_amnt == 0){
      /* done with file: do eof to mark end */
#if 0
      /* this can be done by zero length write: */
      oti_write(odo->aud_fd, odo->aud_data, 0);
#else
      /* or this can be done by api call */
      oti_audio_play_eof(odo->aud_fd);
#endif
      close(odo->play.input_fd);
      odo->play.input_fd = -1;
    }
    
    while (odo->play.aud_amnt > 0) {
      /* write as much data as possible */
      rtn = oti_write(odo->aud_fd, odo->play.aud_ptr, odo->play.aud_amnt);
      
      if (rtn > 0) {
        odo->play.aud_amnt -= rtn;
        odo->play.aud_ptr += rtn;
      } else {
        break;
      }
    }
  }
  
  in_play_service = 0;
}

/*------------------------------------------------------------------------
 *
 *  Function:	stop_play
 *
 *  Description:
 *	Stops playing by pausing the audio device and then flushing it.
 *
 *  Parameters: none
 *	
 *  Returns: void
 *	
 *------------------------------------------------------------------------*/
void stop_play(ODO *odo)
{
  /* Pause the device */
  oti_audio_pause_play(odo->aud_fd);
  oti_audio_flush_play(odo->aud_fd);
}

/*------------------------------------------------------------------------
 *
 *  Function:	g_exit
 *
 *  Description:
 *	Hanldes signals which would cause process to exit (like Ctl-C).
 *      Stops playing and the closes various file descriptors. This
 *      routine can be called directly.
 *
 *  Parameters:
 *	int signo: signal number
 *	
 *  Returns: void
 *	
 *------------------------------------------------------------------------*/
void g_exit(ODO *odo, int signo)
{
  int rc;

  odo->poll_enabled = 0;
  
  if (odo->aud_mode == PLAY) {
    stop_play(odo);
    odo->play.playing = 0;
  } else if (odo->aud_mode == REC) {
    stop_record (odo);
    odo->rec.recording = 0;
  }

  if (odo->aud_fd != -1){
    oti_close(odo->aud_fd);
    odo->aud_fd = -1;
  }
  
  if (odo->audctl_fd != -1){
    oti_close(odo->audctl_fd);
    odo->audctl_fd = -1;
  }
  
  if (odo->aud_mode == PLAY) {
    if (odo->play.input_fd != -1){
      close(odo->play.input_fd);
      odo->play.input_fd = -1;
    }
  }else if (odo->aud_mode == REC) {
    if (odo->rec.output_fd != -1) {
      if((rc = audio_rewrite_filesize(odo->rec.output_fd, 
				      odo->rec.byte_count)) != AUDIO_SUCCESS) {
	fprintf (stderr, 
	       "g_exit error: not able to write filesize to output file: %d\n",
		 rc);
      }
      close (odo->rec.output_fd);
      odo->rec.output_fd = -1;
    }
  }
  exit(0);
}


/* -----------------------------------------------------------------
 *
 *
 *
 * ----------------------------------------------------------------- */
int odo_play (ODO *odo)
{
  int      rc;

  char     audctl_device[80];

  unsigned int value;
  
  Audio_hdr hdr;
  Audio_info tmpinfo;

  /* Open audio device and setup POLL signal */  
  if ((odo->aud_fd = oti_open(odo->dev.aud_device, O_WRONLY | O_NDELAY))<0){
    fprintf(stderr,"could not open osprey audio device\n");
    g_exit(odo, 0);
  }
  if (oti_ioctl(odo->aud_fd, I_SETSIG, (void*)(S_OUTPUT | S_MSG)) < 0){
    perror("I_SETSIG ioctl");
    g_exit(odo, 0);
  }
  
  /* Open audio control device and setup POLL signal  */
  get_audioctl_name(audctl_device, odo->dev.aud_device);
  if ((odo->audctl_fd = oti_open(audctl_device, O_RDWR))<0){
    fprintf(stderr,"could not open audio ctl device\n");
    g_exit(odo, 0);
  }    
  if (oti_ioctl(odo->audctl_fd, I_SETSIG, (void*)(S_MSG)) < 0){
    perror("I_SETSIG ioctl");
    g_exit(odo, 0);
  }

  /* Set the desired gains */
  oti_audio_set_play_gain(odo->aud_fd, &odo->dev.ogain);
  oti_audio_set_monitor_gain(odo->aud_fd, &odo->dev.mgain);

  /* Set the desired buffer size  */
  AUDIO_INITINFO(&odo->tmpinfo);  
  odo->tmpinfo.record.buffer_size = odo->buf_sz;
  odo->tmpinfo.play.buffer_size = odo->buf_sz;
  odo->tmpinfo.play.pause = 0;
  odo->tmpinfo.play.samples = 0;
  odo->tmpinfo.play.error = 0;
  oti_audio_setinfo(odo->aud_fd, &odo->tmpinfo);
  
  /* Get the audio file format  */
  if ((rc = audio_read_filehdr(odo->play.input_fd, &odo->hdr, odo->info, 
			       strlen(odo->info))) != AUDIO_SUCCESS){
    oti_close(odo->aud_fd);
    fprintf(stderr,"unable to read audio header from %s\n", odo->filename);
    g_exit(odo, 0);
  }
  
  printf("playing file %s %s\n", odo->filename, odo->info);
  
  printf("   encoding=%d sample_rate=%d samples_per_unit=%d\n",
         odo->hdr.encoding, odo->hdr.sample_rate, odo->hdr.samples_per_unit);
  printf("   bytes_per_unit=%d channels=%d endian=%d\n",
         odo->hdr.bytes_per_unit, odo->hdr.channels, odo->hdr.endian);

  /* using audio_write_filehdr always seems to nuke endian select */
  odo->hdr.endian = AUDIO_ENDIAN_BIG;
  oti_audio_set_play_config(odo->aud_fd, &odo->hdr);

  value = 0;
  oti_audio_set_play_eof(odo->aud_fd, &value);
  
  odo->poll_enabled = 1;
  odo->play.playing  = 1;
fprintf (stderr, "play.1: odo->aud_fd = %f\n", odo->aud_fd);
  /* SIGPOLL kicks off play_service() */
  (void) kill(getpid(), SIGPOLL);
fprintf (stderr, "play.2: odo->aud_fd = %f\n", odo->aud_fd);
  /* Playing happens in the background via async POLL signals.
   * Wait here for EOF to be processed  */
  while (odo->play.playing){
    poll(NULL,0,5);
    odo->poll_enabled = 0;
    oti_audio_get_play_eof(odo->aud_fd, &value);
    odo->poll_enabled = 1;    
    if (value > 0)
      odo->play.playing = 0;
  }

  /* Wait for drain  */
  poll(NULL,0,20);
  
  g_exit(odo, 0);
  
  return 0;
}
  
  
