/* ************************************************************************* 
 *                       codec.c
 *--------------------------------------------------------------------------
 *         rights reserved by Bo-kyun Na (bkna@top.cis.syr.edu)
 ***************************************************************************
 *  This file includes functions saving the captured XIL compressed image 
 * seqence (CIS) into a file.
 * *************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <memory.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <xil/xil.h>

#include "memmap.h"
#include "vdo.h"

/* - - - - - - - - - - - - - - - - - - - - - - */
/* - - - - - - - - - - - - - - - - - - - - - - */
void write_frames(FILE *df, XilCis cis)
{
  Xil_unsigned8 *data;
  int nframes, nbytes;
  
  while (xil_cis_has_frame(cis)) {
    data = (void *) xil_cis_get_bits_ptr(cis, &nbytes, &nframes);
    
    if (df)
      fwrite(data, 1, nbytes, df);
  }
}

/* - - - - - - - - - - - - - - - - - - - - - - */

int compressor (XilSystemState state, IMAGE *img, CIS *cis,
		VDOTYPE *vdo, IMGPROC *imgproc)
{
  FILE *df;
  char pattern[32];

  int i;
  int c;
  int format;
  int frames;

  cis->h261.bitrate = 256;
  cis->h261.mquant  = 10;  
  
  cis->mpeg.bitrate = 2000;
  cis->mpeg.rate_control = 1;
  cis->mpeg.fixed_quant = 10;

  strcpy(pattern, "IIII");
                       
  df = fopen(cis->cisfile, "w");
  if (df == NULL) {
    fprintf(stderr, "failed to open output movie file.\n");
    exit(1);
  }

  /* get image attribute */
  xil_get_device_attribute(img->handle, "FORMAT_V", (void **) &format);
  if (format == 0) {
    fprintf(stderr, "Unknown video format, exiting.\n");
    xil_close(state);
    exit(1);
  }

  /* compression of image */
  if (cis->do_cis == H261) {
    xil_set_device_attribute(img->handle, "H261_BIT_RATE", 
			     (void *) cis->h261.bitrate);
    xil_set_device_attribute(img->handle, "H261_MAX_QUANT",
			     (void *) cis->h261.mquant);
  }
  
  if (cis->do_cis == MPEG) {
    xil_set_device_attribute(img->handle, "MPEG_RATE_CONTROL", 
			     (void *) cis->mpeg.rate_control);
    xil_set_device_attribute(img->handle, "MPEG_FIXED_QUANT", 
			     (void *) cis->mpeg.fixed_quant);
  }

  if (vdo->image_skip)
    xil_set_device_attribute(img->handle, "IMAGE_SKIP",
                             (void *) vdo->image_skip);
  
  xil_get_info(img->handle, &(img->width), &(img->height), &(img->nbands), 
	       &(img->datatype));

  if (cis->do_cis == H261) {
    if (imgproc->scale.doscale == 2){
      img->width = 352;
      img->height = 288;
    } else if (imgproc->scale.doscale == 4) {
      img->width = 352/2;
      img->height = 288/2;
    } else {
      fprintf(stderr,
	    "Invalid scale factor for H261 Compression: only 2 and 4 legal\n");
      xil_close(state);
      exit(1);
    }
    img->scaled = xil_create(state, img->width, img->height, img->nbands,
			     XIL_BYTE);
  } else {
    if (imgproc->scale.doscale) {
      img->width = (int) (((float) img->width / imgproc->scale.doscale) + 0.5);
      img->height = (int) \
	(((float) img->height / imgproc->scale.doscale) + 0.5);
      img->scaled = xil_create(state, img->width, img->height, img->nbands, 
			       XIL_BYTE);
    }
  }

  /* MPEG requires CIF (352 width) compression */
  if (cis->do_cis == MPEG) {
      if (imgproc->scale.doscale != 2) {
          fprintf(stderr, "ERROR: scalefactor is: %d, should be 2 for Mpeg1\n",
                  imgproc->scale.doscale);
          xil_close(state);
          exit(1);
      }
      img->width  = 352; /* Osprey Mpeg1 compression is 601 not square */
      img->height = 240;
      img->scaled = xil_create(state, img->width, img->height, img->nbands, 
			       XIL_BYTE);
  }

  /* Create a cis to put the movie frames in. */
  cis->cis = xil_cis_create(state, cis->cis_type);

  /*
   * set the buffers to make sure that you don't keep information you
   * do not need.  H261 needs more frames.
   */
  if (cis->do_cis == H261) {
    xil_cis_set_keep_frames(cis->cis, 3);
    xil_cis_set_max_frames(cis->cis, 3);
  } else if (cis->do_cis == MPEG) {
      xil_cis_set_keep_frames(cis->cis, 6);
      xil_cis_set_max_frames(cis->cis, 6);
  } else {
    xil_cis_set_keep_frames(cis->cis, 1);
    xil_cis_set_max_frames(cis->cis, 1);
  }

  if (cis->do_cis == MPEG) {
      int bitrate = 1000 * cis->mpeg.bitrate;
      xil_cis_set_attribute(cis->cis, "COMPRESSOR_BITS_PER_SECOND", 
			    (void*)bitrate);
  }

  /*
   * If JPEG, set an attribute needed for fast decompression of the
   * resulting movie.
   */
  if (cis->do_cis == JPEG)
      xil_cis_set_attribute(cis->cis, "ENCODE_411_INTERLEAVED", (void *) TRUE);

  for (frames = 0; frames < vdo->total_frames; frames ++) {
printf ("frames = %d, total frames = %d\n", frames, vdo->total_frames);    
    if (cis->do_cis == H261 || cis->do_cis == MPEG){
      if (format == 2) /* NTSC */
        xil_scale(img->handle, img->scaled, "nearest",
                  img->width/640.0, img->height/480.0);
      else
        xil_scale(img->handle, img->scaled, "nearest",
                  img->width/768.0, img->height/576.0);        
    } 
    else if (imgproc->scale.doscale) {
      xil_scale(img->handle, img->scaled, "nearest",
                1.0 / (float) imgproc->scale.doscale, 
		1.0 / (float) imgproc->scale.doscale);
    }
    else {
      img->scaled = img->handle;
    }
    
    xil_compress(img->scaled, cis->cis);
    xil_cis_sync(cis->cis);
    xil_toss(img->scaled);
    
    write_frames(df, cis->cis);
  }
  
  /* done, empty buffer if any frames remain */
  xil_cis_flush(cis->cis);
  write_frames(df, cis->cis);
  
  xil_close(state);
  
  return 0;
}


/* - - - - - - - - - - - - - - - - - - - - - - */
/* memory-mapping-related functions */
/* - - - - - - - - - - - - - - - - - - - - - - */
void openfile(MFILE* memfile, char* fname)
{
    int fd;
    Xil_unsigned8 *fp;
    struct stat buf;
  
    if ((fd = open(fname, O_RDONLY)) < 0) {
        /* XILERROR(OPEN); */
        fprintf(stderr, "ERROR: failed to open %s file.\n", fname );
        exit(-1);
    }
    fstat(fd, &buf);
    fp = (Xil_unsigned8 *) mmap(0, (size_t)buf.st_size, PROT_READ,
        MAP_SHARED, fd, (off_t)0);
    if ((int)fp == -1) {
        /* XILFATALERROR(MMAP); */
        fprintf(stderr, "ERROR: mmap failed\n");
        exit(-1);
    }
    memfile->dptr = fp;
    memfile->start = fp;
    memfile->fd = fd;
    memfile->len = (int) buf.st_size;
}

void init_memfile( MFILE* memfile )
{
    memfile->dptr = NULL;
    memfile->start = NULL;
    memfile->fd = 0;
    memfile->len = 0;
    memfile->currentFrame = 0;
    memfile->mstart = NULL;
    memfile->mlen = 0;
}

void detach_file( MFILE* memfile )
{
    close(memfile->fd);
    if (munmap((caddr_t) memfile->start, memfile->len) == -1) {
        /* XILFATALERROR(MUNMAP); */
        fprintf(stderr, "ERROR: munmap failed in detach_file\n");
    }
    memfile->start = 0;
    memfile->len = 0;
    memfile->fd = 0;
}

void attach_file(MFILE *memfile, char *fname)
{
    if (memfile->start)
        detach_file( memfile);
  
    openfile(memfile,fname);
  
    memfile->mstart = memfile->dptr;
    memfile->mlen = memfile->len;
}

/*----------------------------------------------------------------------
* Description:
*
*   This file contains routines specific to JPEG colormap
*   creation and installation.
*----------------------------------------------------------------------*/
XilLookup create_cmap(XilSystemState State, XilCis cis,
    Display *display, Window window, int screen,
    Colormap *cmap, char *type, XilIndexList *ilist, 
    XilLookup yuvtorgb, XilLookup colorcube)
{
    unsigned long tmp_colors[256], pixels[256], mask;
    XColor cdefs[256];
    int top_colors = 2;	/* colormap indices 255 and 254 */
    int i, j, t;
    Xil_unsigned8 *data;
    XilLookup lut;
    int cmapsize;

    /* Get colormap size */
    if (type == "JPEG")
      cmapsize = xil_lookup_get_num_entries(yuvtorgb);

    /* Create an X colormap for the cis.*/
    *cmap = XCreateColormap(display, window, DefaultVisual(display, screen), 
        AllocNone);

    /* Allocate X Colormap cells
     *
     * If we do not need the entire colormap, allocate `cmapsize'
     * entries just below the top of the colormap.  Here's how:
     *    Temporarily allocate some entries at the front of the cmap.  
     *    Don't allocate the top_colors (the top two color indices 
     *	  are often used by other applications). 
     *    Allocate the needed entries in the next cmap section
     *    Free the temporary entries.
     * This allows the X-Window manager to use them and reduces the
     * chances of of your window colormap flashing.
     */

    if (cmapsize < 256-top_colors) {
        if (!XAllocColorCells(display, *cmap, 0, &mask, 0,
	        tmp_colors, 256 - cmapsize - top_colors)) {
            fprintf(stderr, " XAllocColorCells for cmap_create failed(1)\n");
            exit(1);
        }
    }
  
    if (!XAllocColorCells(display, *cmap, 0, &mask, 0, pixels, cmapsize)) {
        fprintf(stderr, " XAllocColorCells for cmap_create failed(2)\n");
        exit(1);
    }
    /* The remaining code assumes that the values returned in pixels[0] through
     * pixels[cmapsize-1] are a contiguous range.
     */

    /* Free the unused colors in the front */
    if(cmapsize < 256-top_colors)
      XFreeColors(display, *cmap, tmp_colors, 256 - cmapsize - top_colors, 0);

    /* Allocate memory to hold colormap data. */
    if ((data = (Xil_unsigned8 *)
	 malloc(sizeof(Xil_unsigned8) * cmapsize * 3) ) == NULL ) {
      fprintf(stderr, "xilcis_color: out of memory for cmap data create\n");
      exit(1);
    }  

    /* Get the entries for the colormap. The method depends on the compression
     * type.  For CELL, get the entries in the current default colormap.
     * For JPEG, get the entries from the standard lookup table
     * yuv_to_rgb.
     */

    if(type == "JPEG") {
        xil_lookup_get_values(yuvtorgb, xil_lookup_get_offset(yuvtorgb),
                              cmapsize, data);
        xil_lookup_set_offset(colorcube, (unsigned int)pixels[0]);
        for (i = 0, t = 0; i < cmapsize; i++, t += 3) {
            cdefs[i].pixel = pixels[i];
            cdefs[i].flags = DoRed | DoGreen | DoBlue;
            cdefs[i].blue = data[t] << 8;  
            cdefs[i].green = data[t+1] << 8; 
            cdefs[i].red = data[t+2] << 8;  
        }
        XStoreColors(display, *cmap, cdefs, cmapsize);
        lut = yuvtorgb;
    }

    /* Install colormap before cursor is placed in window */
    XInstallColormap(display, *cmap);
    XSync(display, False);

    XSetWindowColormap(display, window, *cmap);

    free(data);

    return(lut);
}


/* ----------------------------------------------
 *  Attempt to provide rate controlled playback  
 * ---------------------------------------------- */
void rate_control(int rate)
{
  static double t0 = 0;
  static double t1 = 0;
  double delay;
  
  if (!rate)
    return;

  /* rate is in fps: determine ns delay */
  delay = (1000000.0/rate);
  
  t1 = gethrtime()/1000.0;

  /*printf("%.2f - %.2f = %.2f < %.2f\n",t1,t0,t1-t0,delay);*/
  while (t1-t0 < delay){
    poll(0,0,2);
    t1 = gethrtime()/1000.0;
  }
  t0 = t1;
}


/* ------------------------------------------------------------
 *
 *
 * ------------------------------------------------------------ */
void find_quit(Display* display, Atom quitAtom)
{
  XEvent event;
  
  if(XPending(display) > 0) {
    XNextEvent(display, &event);
    if (event.xany.type == ClientMessage &&
        ((XClientMessageEvent *)&event)->data.l[0] == quitAtom)
      exit(0);
  }
}


/* ---------------------------------------------------------------
 * Create an X-Window to put the movie in 
 * --------------------------------------------------------------- */
int display_cis (XilSystemState state, IMAGE *img, XWIN *xwin, 
		 IMGPROC *imgproc)
{
  xwin->screen_num = DefaultScreen(xwin->xdisplay);
  xwin->xwindow = XCreateSimpleWindow(xwin->xdisplay, 
				      RootWindow(xwin->xdisplay, 
						 xwin->screen_num), 0, 0,
				      img->width, img->height, 0, 
				      BlackPixel(xwin->xdisplay, 
						 xwin->screen_num),
				      WhitePixel(xwin->xdisplay, 
						 xwin->screen_num));
  if (!xwin->xwindow) {
    fprintf(stderr, "Unable to create window\n");
    exit(1);
  }

  /* put window title bar */
  if (xwin_titlebar (img, xwin)) {
    fprintf(stderr, "Can't make title bar on the window\n");
    exit(1);
  }

  /* Visualize of window */
  xwin->visual = DefaultVisual(xwin->xdisplay, xwin->screen_num);
  xwin->visualid = XVisualIDFromVisual(xwin->visual);
  xwin->vis_template.visualid = xwin->visualid;
  xwin->vis_template.depth = 24;
  xwin->vis_template.class = TrueColor;
  XGetVisualInfo(xwin->xdisplay, 
		 VisualIDMask | VisualDepthMask | VisualClassMask,
		 &xwin->vis_template, &xwin->matching_visuals);
  if (xwin->matching_visuals == 1)
    xwin->default_vis = TRUECOLOR;
  else {
    xwin->vis_template.depth = 8;
    xwin->vis_template.class = PseudoColor;
    XGetVisualInfo(xwin->xdisplay,
		   VisualIDMask | VisualDepthMask | VisualClassMask,
		   &xwin->vis_template, &xwin->matching_visuals);
    if (xwin->matching_visuals != 1) {
      fprintf(stderr, "Default visual for this screen must be an "); 
      fprintf(stderr, "8-bit PseudoColor or 24-bit TrueColor visual\n");
      exit(1);
    }
  }
      
  if (xwin->closedownAtom = XInternAtom(xwin->xdisplay, 
					"WM_DELETE_WINDOW", False))
    XSetWMProtocols(xwin->xdisplay, xwin->xwindow, &xwin->closedownAtom, 1);
    
  /* Need to get window exposure and button press events. */
  XSelectInput(xwin->xdisplay, xwin->xwindow, ExposureMask | ButtonPressMask);
    
  /* Make this window visible */
  XMapWindow(xwin->xdisplay, xwin->xwindow);
    
  /* Wait for an expose event to make sure window has been properly
     displayed */
  while (1) {
    XNextEvent(xwin->xdisplay, &xwin->event);
    if (xwin->event.xany.type == Expose &&
	xwin->event.xexpose.window == xwin->xwindow) break;
  }
      
  /* Create an XIL display image */
  img->display = xil_create_from_window(state, xwin->xdisplay, 
					      xwin->xwindow);
  if (!img->display) {
    /* XIL sends error message to stderr if xil_create_from_window fails */
    exit(1);
  }
}


/* -------------------------------------------------------------------
 * Function: dcompress () 
 * Usage   : decompression of image data
 *
 *
 * ------------------------------------------------------------------ */
int dcompressor (XilSystemState state, IMAGE *img, IMGPROC *imgproc, 
		 XWIN *xwin, CIS *cis)
{
  /* ----- Program Variables ----- */
  int i, j;
  MFILE *memfile; /* The MFILE structure is defined in memmap.h */

  int frame_count = -1;
  enum history_state ignore_history = OFF;
    
  /* ----- XIL variables ----- */
  XilDevice device = NULL;

  /*
   * Currently, the Osprey-1x00 decompression molecules default to using
   * the /dev/o1k0 device. An XIL decompression-only application is not
   * required to create a MMACo1k device in order to decompress using
   * this device. However, if this device is in use by another process,
   * them the h/w decompression molecules will fail. The current work-
   * around for this is to create a MMACo1k device in the application,
   * set the DEVICE_NAME attribute and finally create a device image
   * from this device. These are the same steps that are used in a
   * compression application. In a future release, the Osprey-1x00
   * decompression molecules will look for a free o1k device and use
   * it for decompression.
   */

  if (xwin->devname != NULL){
    if (!(device = xil_device_create(state, "MMACo1k"))) {
      fprintf(stderr, "Unable to create a device object\n");
      xil_close(state);
      exit(1);
    }
    xil_device_set_value(device, "DEVICE_NAME", (void *) xwin->devname);
    img->handle = xil_create_from_device(state, "MMACo1k", device);
    xil_device_destroy(device);
  }
    
  xwin->xdisplay = XOpenDisplay(NULL);
  if (xwin->xdisplay == NULL) {
    fprintf(stderr, "Unable to to connect to X server.\n");
    exit(1);
  }
    
  /* Memory-map file containing Compressed Image Sequence (CIS) data */
  if( (memfile = (MFILE *) malloc(sizeof(MFILE)) ) == NULL ) {
    fprintf(stderr, "Out of memory for memfile create\n");
    exit(1);
  }  
  init_memfile(memfile);
  attach_file(memfile, cis->cisfile);


  /* Create the CIS */
  cis->cis = xil_cis_create(state, cis->cis_type);
  if (!cis->cis) {
    /* XIL sends an error message to stderr if xil_cis_create fails */
    exit(1);
  }

  /* Set decompression quality for faster speed for Mpeg1  */
  if(cis->cis_type == "Mpeg1")
    xil_cis_set_attribute(cis->cis, "DECOMPRESSOR_QUALITY", (void *)90);

  /* Give the CIS a pointer to the data */
  xil_cis_put_bits_ptr(cis->cis, memfile->mlen, frame_count, 
		       memfile->mstart, NULL);

  /* Get information from the CIS for dimensions */
  img->outputtype = xil_cis_get_output_type(cis->cis);
  xil_imagetype_get_info(img->outputtype, &img->width, &img->height,
			 &img->nbands, &img->datatype);
  if (img->nbands != 3) {
    fprintf(stderr,
	    "Program cannot play JPEG movies made from %d band images\n",
	    img->nbands);
    exit(1);
  }

  if (!display_cis (state, img, xwin, imgproc)) {
    printf ("Image can not be displayed on Xwindow");
    exit (1);
  }

  { /* decompression of JPEG and H.261 formats */
        
    /*    XilLookup colorcube;
	  XilDitherMask dmask;
	  float scale[3], offset[3]; 
    */

    XilLookup xil_cmap;
    XilColorspace ycc, rgb;
    XilImage imageYCC = NULL;
        
    /* imageYCC is the XilImage that the cis will be decompressed into. */
    imageYCC = xil_create(state, img->width, img->height, img->nbands,
			  img->datatype);
    if (imageYCC == NULL) {
      /* XIL sends error message to stderr if xil_create fails */
      exit(1);
    }
        
    if (xwin->default_vis == PSEUDOCOLOR) {
      imgproc->scale.rescale_values[0] = 255.0 / (235.0 - 16.0);
      imgproc->scale.rescale_values[1] = 255.0 / (240.0 - 16.0);
      imgproc->scale.rescale_values[2] = 255.0 / (240.0 - 16.0);
      imgproc->scale.offset_values[0] = -16.0*imgproc->scale.rescale_values[0];
      imgproc->scale.offset_values[1] = -16.0*imgproc->scale.rescale_values[1];
      imgproc->scale.offset_values[2] = -16.0*imgproc->scale.rescale_values[2];
            
      imgproc->colorcube = xil_lookup_get_by_name(state, "cc855");
      imgproc->mask = xil_dithermask_get_by_name(state, "dm443");
            
      /* This routine is in the file: xilcis_color.c. Inside this routine
	 the colormap is installed through X, and the color cube offset
	 is set to the cmap offset. */
      xil_cmap = create_cmap(state, cis->cis, xwin->xdisplay, xwin->xwindow,
			     DefaultScreen(xwin->xdisplay), &xwin->x_cmap,
			     "JPEG", NULL, 
			     xil_lookup_get_by_name(state, "yuv_to_rgb"),
			     imgproc->colorcube);
            
      while (xil_cis_has_frame(cis->cis)) {
                
	/* Begin XIL Molecule... */
	rate_control(cis->rate);
	xil_decompress(cis->cis, imageYCC );
	xil_rescale(imageYCC, imageYCC, imgproc->scale.rescale_values, 
		    imgproc->scale.offset_values);
	xil_ordered_dither(imageYCC, img->display, imgproc->colorcube, 
			   imgproc->mask);
	/* End of XIL Molecule */
	find_quit(xwin->xdisplay, xwin->closedownAtom);
      }
    }
        
    if (xwin->default_vis == TRUECOLOR) {
      ycc = xil_colorspace_get_by_name(state, "ycc601");
      rgb = xil_colorspace_get_by_name(state, "rgb709");
      xil_set_colorspace(imageYCC, ycc);
      xil_set_colorspace(img->display, rgb);
      while (xil_cis_number_of_frames(cis->cis)>1) {
                
	/* Begin XIL molecule */
	rate_control(cis->rate);
	xil_decompress(cis->cis, imageYCC);
	xil_color_convert(imageYCC, img->display);
	/* End of XIL molecule */
	find_quit(xwin->xdisplay, xwin->closedownAtom);
      }
    }
        
    /* Determine number of frames played */
    frame_count = xil_cis_get_read_frame(cis->cis) - 1;
        
    /* Wait for user to press button. Also, refresh the window if there
       is an exposure event. */
    if (!imgproc->exit_on_done)
      while (1) {
	XNextEvent(xwin->xdisplay, &xwin->event);
            
	/* Exit JPEG decompression when user presses button in window */
	if (xwin->event.xany.type == ButtonPress) break;
	
	if (xwin->event.xany.type == ClientMessage &&
	    ((XClientMessageEvent *)&(xwin->event))->data.l[0] == \
	    xwin->closedownAtom)
	  break;
            
	if (xwin->event.xany.type == Expose) {
                
	  /* Go to the last frame of the movie.
	   * For CellB and H261, must use the ignore history attribute to
	   * backup 1 frame.  This frame is "special" because it is the
	   * last decompressed frame and our history buffer is 1 deep.
	   * So we are able to re-decompress/display the last frame with
	   * valid history.
	   */
	  if (cis->do_cis == H261){
	    ignore_history = ON;
	    xil_cis_set_attribute(cis->cis,"IGNORE_HISTORY",
				  (void *)ignore_history);
	    xil_cis_seek(cis->cis, -1, 2);
	    ignore_history = OFF;
	    xil_cis_set_attribute(cis->cis,"IGNORE_HISTORY", 
				  (void *)ignore_history);
	  }
	  else 
	    xil_cis_seek(cis->cis, -1, 2);
	  
	  rate_control(cis->rate);
	  if (xwin->default_vis == PSEUDOCOLOR) {
	    /* Begin XIL Molecule */
	    xil_decompress(cis->cis, imageYCC);
	    xil_rescale(imageYCC, imageYCC, imgproc->scale.rescale_values, 
			imgproc->scale.offset_values);
	    xil_ordered_dither(imageYCC, img->display, imgproc->colorcube,
			       imgproc->mask);
	    /* End of XIL Molecule */
	  }
	  else {
	    xil_decompress(cis->cis, imageYCC);
	    xil_color_convert(imageYCC, img->display);
	  }    
	}
      }
    if ( imageYCC )
      xil_destroy( imageYCC );
    /* end of decompress to display */
  }
    
  /* Clean up */
  if (img->display)
    xil_destroy(img->display);
  xil_cis_destroy(cis->cis);
  xil_close(state);
  XCloseDisplay(xwin->xdisplay);
  /*  exit(0); */

  return 1;
}

/* ----------------------------------------------------------------
 *
 *
 * ---------------------------------------------------------------- */
int xwin_titlebar (IMAGE *img, XWIN *xwin)
{
  XSizeHints hints;
  int port;
  char *format;
      
  hints.flags = 0;		/* random position */

  xil_get_device_attribute(img->handle, "FORMAT", (void **) &format);
  xil_get_device_attribute(img->handle, "PORT_V", 
			   (void **) &(xwin->port));

  /* window title when the display window opened  */
  if ((void *) xwin->titlebar == 0 )
    sprintf(xwin->titlebar, "%s: Port %d [%s]", 
	    (xwin->devname ? xwin->devname : "/dev/o1k0"), 
	    xwin->port, format);
    
  /* This Function informs the standard info of the window to win manager */
  XSetStandardProperties(xwin->xdisplay, xwin->xwindow, xwin->titlebar, 
			 xwin->titlebar, 0, xwin->argv, *(xwin->argc), &hints);
  return 0;
}
