/* Copyright (c) 1997 MultiMedia Access Corporation - Osprey Systems Division. 
 * 1150 S.E. Maynard Rd., Suite 110, Cary, NC 27511 USA. All Rights Reserved.
 *
 * This product and related documentation is protected by copyright laws and
 * international treaties; however, the user of this software is granted a 
 * nonexclusive, royalty free license to modify, make derivitive works, use,
 * distribute, or otherwise incorporate this software in products which use
 * or incorporate the MMAC product.
 *
 * MMAC reserves the right to change any products herein at any time and 
 * without notice. MMAC does not assume any liability arising from the 
 * application or use of this product, except as expressly agreed to in 
 * writing by MMAC. Use and purchase of this product does not convey a license
 * under any patent rights, copyrights, trademark rights, or any other 
 * intellectual property rights of MMAC, Sun Microsystems, or others. 
 */

/*-----------------------------------------------------------------------
 * Filename:     xil_record.c
 *
 * Description:  This xil_record example program records video images 
 *               and writes it into a file. This example uses the Sun
 *               Rasterfile format as an illustration. However, the
 *               developer may choose to implement other formats.
 *               The example formats that are supported by this program 
 *               are:
 *               a) GRAYSCALE
 *               b) DITHERED8 (A 24bit color image that is 
 *                             dithered to an 8bit color image)
 *               c) RGB24BIT  (A 24bit color image)
 *               d) RGB32BIT  (A 32bit color image)
 *
 *-----------------------------------------------------------------------*/

#include    <stdio.h>
#include    <stdlib.h>
#include    <sys/types.h>
#include    <sys/stat.h>
#include    <fcntl.h>
#include    <math.h>
#include    "xil_record.h"

/************************************************
 * Main
 ************************************************/
void main(int argc, char *argv[])
{
    /* Xil objects */
    XilDevice device = NULL;  
    XilSystemState state = NULL;
    XilImage dev_image = NULL; 
    XilImage child_dev_image = NULL;
    XilImage dst_image = NULL;
    XilImage child_dst_image = NULL;
    XilImage scaled_image = NULL;
    XilImage rescaled_image = NULL;
    XilDataType datatype;
    XilMemoryStorage storage;
    Xil_boolean status;
    XilDitherMask mask;
    XilLookup colorcube855 = NULL;
    XilLookup yuv_to_rgb = NULL;
    XilColorspace cspace_ycc601;
    XilColorspace cspace_rgb709;
    Xil_unsigned8 cmap_data[MAX_CMAP_LEN * 3];
    
    /* raster file structure used for this example */
    struct rasterfile ras_hdr;
    
    /* Conversion table for formats to number of bands */
    const int formatToBands[] = {1,1,3,4};
    
    /* command line arguments */
    const char getoptargs[]= "w:h:g:f:o:D:T:n:I:i:";
    
    /* image info */
    unsigned int nbands;
    unsigned int dev_width;
    unsigned int dev_height;
    int frame_size;
    
    /* rescale variables */
    float rescale_values[3];
    float offset_values[3];
    
    /* scale variables */
    float width_scale_factor;
    float height_scale_factor; 

    /* Program Variables */
    int cmap_len = CMAP_SIZE;
    unsigned char *cmap[3] = {NULL,NULL,NULL};
    int frames;
    double time;
    double wait_time;
    struct timeval t0, t1;
    int i,j;
    int input;
    
    /* Default values */
    char file_name[80] = "";
    char *devname = DEVNAME;
    char *devtype = DEVTYPE;   
    int format = GRAYSCALE_FORMAT;
    int outfile_fd = -1;
    int scale = 0;
    int port_num = 1;
    int frame_count = 100;
    int image_skip = 0;
    int desired_width = 0;
    int desired_height = 0;
    double gamma_val = 0.7;
    
    /* Handle user input */
    while ((input = getopt(argc, argv, getoptargs)) != EOF) { 
	switch (input) {
	case 'D':
	    devname = optarg;
	    break;
	case 'f':
	    format = atoi(optarg);
	    break;
	case 'g':
	    gamma_val = atof(optarg);
	    break;
	case 'h':
	    desired_height = atoi(optarg);
	    break;
	case 'i':
	    image_skip = atoi(optarg);
	    break;
	case 'n':
	    frame_count = atoi(optarg);
	    break;
	case 'o':
	    strcpy(file_name, optarg);
	    break;
	case 'I':
	    port_num = atoi(optarg);
	    break;
	case 'w':
	    desired_width = atoi(optarg);
	    break;
	case 'T':
	    devtype=optarg;
	    break;
	default:
	    fprintf(stderr, "%s:  unknown flag\n", argv[0]);
	    usage(argv[0]);
	    exit(1);
	    break;
	}
    }
    
    /*--------------------
     * Open an XIL session
     *--------------------*/
    if ((state = xil_open()) == NULL) {
	perror("Unable to open an Xil Session\n");
	exit(1);
    }
    
    /*---------------
     * Initialization 
     *---------------*/
    switch (format) {
    case DITHERED8_FORMAT:
	/* Setting up the rescale and the offset values. */
	rescale_values[0] = 255.0 / (235.0 - 16.0);
	rescale_values[1] = 255.0 / (240.0 - 16.0);
	rescale_values[2] = 255.0 / (240.0 - 16.0);
	offset_values[0] = -16.0 * rescale_values[0];
	offset_values[1] = -16.0 * rescale_values[1];
	offset_values[2] = -16.0 * rescale_values[2];
	
	/* Getting the colorcube for dithering, colormap and mask. */
	colorcube855 = xil_lookup_get_by_name(state, "cc855");
	yuv_to_rgb = xil_lookup_get_by_name(state, "yuv_to_rgb");
	mask = xil_dithermask_get_by_name(state, "dm443");
	break;
	
    case RGB24BIT_FORMAT:
    case RGB32BIT_FORMAT:
	/* Getting the colorspaces. */
	cspace_ycc601 = xil_colorspace_get_by_name(state, "ycc601");
	cspace_rgb709 = xil_colorspace_get_by_name(state, "rgb709");
	break;
	
    default:
	break;
    } 
    
    /*---------------------------------------------
     * Creating a device attached to a Capture Card
     *---------------------------------------------*/
    if ((device = xil_device_create(state, devtype)) == NULL) {
	printf ("Unable to create a device object. \n");
	xil_close(state);
	exit(1);
    }
    
    /*------------------------
     * Setting some attributes
     *------------------------*/
    xil_device_set_value(device, "DEVICE_NAME", (void *) devname);
    xil_device_set_value(device, "PORT_V", (void *) port_num);
    xil_device_set_value(device, "IMAGE_SKIP", (void *) image_skip);
    
    /*-------------------------- 
     * Creating the device image
     *--------------------------*/
    if ((dev_image = xil_create_from_device(state, devtype, device)) == NULL) {
	perror("Failed to open capture device\n");
	xil_close(state);
	exit(1);
    }
    
    /*-------------------
     * Destroy the device
     *-------------------*/
    xil_device_destroy(device);
    
    /*---------------------------------------- 
     * Getting information of the device image
     *----------------------------------------*/
    xil_get_info(dev_image,&dev_width,&dev_height,&nbands,&datatype);
    
    /*-----------------------------------------
     * Checking the width and height dimensions
     *-----------------------------------------*/
    if (!desired_width) desired_width = dev_width;
    if (!desired_height) desired_height = dev_height;
    
    if ((desired_width > dev_width)||(desired_height > dev_height)) {
	printf(" Output width/height is too big.\n");
	if(device) xil_device_destroy(device);     	
	if(dev_image) xil_destroy(dev_image);
	if(state) xil_close(state);
	exit(1);
    }
    
    if ((desired_width != dev_width) || (desired_height != dev_height)) {
	width_scale_factor = (float)desired_width / dev_width;
	height_scale_factor = (float)desired_height / dev_height;
	scale = 1; /* Initialize the scale flag. */
    }
    
    /*-------------------------------
     * Creating the destination image
     *-------------------------------*/
    if ((dst_image = xil_create(state,desired_width,desired_height,
				formatToBands[format],XIL_BYTE)) == NULL){
	perror("Could not create destination image");
	goto onerror;
    }
    
    /* Calculating the frame size of the data that will be written to file. */
    frame_size = desired_width * formatToBands[format] * desired_height;
    
    /*-----------------------------------------------------------------
     * Initialize the Header Info.
     * Please note that this example uses the Sun Rasterfile format but
     * the developer may choose to implement other image formats. 
     *-----------------------------------------------------------------*/
    initSunRasterHeaderInfo(&ras_hdr,desired_width,desired_height,frame_size,
			    cmap_len,format);
    
    /*-----------------------------------------------
     * Setting up colormaps for the respective format
     *-----------------------------------------------*/
    if (format == GRAYSCALE_FORMAT) {
	if (createGrayScaleCMap(cmap,cmap_len,gamma_val) < 0 ) {
	    perror("Unable to create the GrayScale image colormap.\n");
	    goto onerror;
	}
    }
    else if (format == DITHERED8_FORMAT){
	if (createDitheredCMap(colorcube855, yuv_to_rgb,cmap,cmap_len) < 0) {
	    perror("Unable to create the Dithered image colormap.\n");
	    goto onerror;
	} 
    }
    
    /*---------------------
     * Open the output file
     *---------------------*/
    if ((outfile_fd = open(file_name, 
			   (O_CREAT | O_TRUNC | O_WRONLY), 0666)) <= 0) {
	fprintf(stderr, "%s:  can't open file (%d)\n", argv[0], file_name);
	goto onerror;
    }
    
    /*----------------------------------- 
     * Write the file header and colormap
     *-----------------------------------*/
    if (writeSunRasterHeaderInfo(outfile_fd, &ras_hdr) < 0) {
	fprintf(stderr, "%s:  unable to write header\n", argv[0]);
	goto onerror;
    }
    if (ras_hdr.ras_maplength) {
	if (writeColorMapInfo(outfile_fd,cmap,ras_hdr.ras_maplength/3) < 0) {
	    fprintf(stderr, "%s: unable to write colormap\n", argv[0]);
	    goto onerror;
	}
    }
        
    /*---------------------------------------------------------------
     * Creating a scaled_image of the desired dimensions so as to 
     * store the scaled down image when xil_scale operation is called.
     *----------------------------------------------------------------*/
    if (scale)
	scaled_image = xil_create(state,desired_width,desired_height,3,XIL_BYTE);
    
    /*------------------------------------------------------
     * Creating the temporary images that will be needed 
     * by the respective formats and setting the colorspaces.
     *------------------------------------------------------*/
    if (format == GRAYSCALE_FORMAT)
	/* Creating child device image of band 0 from the 
	 * 3 banded device image. */
	child_dev_image = xil_create_child(dev_image,0,0,dev_width,
					   dev_height,0,1);
    else if (format == DITHERED8_FORMAT)
	/* Creating a 3 banded rescaled image for use 
	 * during a rescale operation. */
	rescaled_image = xil_create(state,desired_width,desired_height,
				    3,XIL_BYTE);
    else if (format == RGB24BIT_FORMAT) {
	/* If scaling is to be done, set the colorspace of the scaled 
	 * image. Otherwise, set the colorspace of the device image. */
	if(scale)
	    xil_set_colorspace(scaled_image, cspace_ycc601);
	else
	    xil_set_colorspace(dev_image, cspace_ycc601);
	
	/* With respect to setting the colorspace of the source image,
	 * we also need to set the colorspace of the dst_image. */
	xil_set_colorspace(dst_image, cspace_rgb709);
	
    }    
    else if (format == RGB32BIT_FORMAT) {
	/* Creating a child dst image from band1 to band3 
	 * from a 4-banded dst_image. */
	child_dst_image = xil_create_child(dst_image,0,0,desired_width,
					   desired_height,1,3);
	/* if scaling is to be done, set the colorspace of the scaled
	 * image. Otherwise, set the colorspace of the device image. */
	if (scale)
	    xil_set_colorspace(scaled_image, cspace_ycc601);
	else
	    xil_set_colorspace(dev_image, cspace_ycc601);
	
	/* With respect to setting the source image, we 
	 * also need to set the colorspace of the dst_image. */
	xil_set_colorspace(child_dst_image, cspace_rgb709);
    }
    
    /*-------------------------------------------------------
     * Processing the writing of the image to the opened file
     *-------------------------------------------------------*/
    for (frames = 0; frames < frame_count; frames++) {
	switch(format){
	case GRAYSCALE_FORMAT:
	    /* If scaling is to be done, then scale band 0 of device image
	     * to the 1 banded dst image.
	     * Otherwise, just copy the band 0 of the device image directly
	     * to the 1 banded dst image. */
	    if(scale) 
		xil_scale(child_dev_image,dst_image,"nearest",width_scale_factor,
			  height_scale_factor);
	    else
		xil_copy(child_dev_image,dst_image);
	    break;
	    
	case DITHERED8_FORMAT:
	    /* If scaling is to be done, scale the 3 banded device image
	     * to the scaled_image that was previously created with the 
	     * desired dimensions, then perform a pixel by pixel rescaling
	     * of the scale_image to the rescaled_image. Finally,
	     * dither the 3 banded image to a 1 banded image for storage. */ 
	    if (scale) { 
		xil_scale(dev_image, scaled_image, "nearest", 
			  width_scale_factor, height_scale_factor);  	    
		xil_rescale(scaled_image,rescaled_image,
			    rescale_values,offset_values);
		xil_ordered_dither(rescaled_image,dst_image,colorcube855,mask);
		xil_toss(scaled_image);
		xil_toss(rescaled_image);
	    }
	    /* Since no scaling is involved, just perform a pixel by pixel 
	     * rescaling of the device image to the rescaled_image and
	     * finally dither the 3 banded image to a 1 banded image. */
	    else {
		xil_rescale(dev_image,rescaled_image,rescale_values,
			    offset_values);
		xil_ordered_dither(rescaled_image,dst_image,colorcube855,mask);
		xil_toss(rescaled_image);
	    }
	    break;
	    
	case RGB24BIT_FORMAT:
	    /* If scaling is to be done, scale the 3 banded device image
	     * to the scaled_image that was previously created with the 
	     * desired dimensions. 
	     * Then color convert the scaled image from ycc to rgb.
	     * Otherwise, simply do a color conversion.  */ 
	    if (scale) {
		xil_scale(dev_image, scaled_image, "nearest", 
			  width_scale_factor, height_scale_factor);
		xil_color_convert(scaled_image, dst_image);
		xil_toss(scaled_image);
	    }
	    else 
		xil_color_convert(dev_image, dst_image);
	    break;
	    
	case RGB32BIT_FORMAT:
	    /* If scaling is to be done, scale the 3 banded image to the
	     * scaled_image which was previously created with the desired
	     * dimensions. Then color convert the scaled image from ycc to rgb.
	     * Otherwise, simply do a color convert from the 3 banded device
	     * image to the previously created child dst image of band 1 
	     * to band 3. */
	    if (scale) {
		xil_scale(dev_image, scaled_image, "nearest", 
			  width_scale_factor, height_scale_factor);
		xil_color_convert(scaled_image, child_dst_image);  
		xil_toss(scaled_image);
	    }
	    else
		xil_color_convert(dev_image, child_dst_image);  
	    break;
	    
	default:
	    break;
	}
      
	/*-------------------------------
	 * Getting a pointer to the image
	 *-------------------------------*/
	if (xil_export(dst_image) == XIL_FAILURE) {
	    perror(" Could not export image.\n");
	    goto onerror;
	}
	if ((status = xil_get_memory_storage(dst_image, &storage)) == FALSE) {
	    perror("Could not get image storage info.\n");
	    goto onerror;
	}
	/* Writing the data image to the opened file. */
	if ((storage.byte.pixel_stride == formatToBands[format]) &&
	    (storage.byte.scanline_stride == desired_width*formatToBands[format])) {
	    if (write(outfile_fd, storage.byte.data, frame_size) < 0){
		fprintf(stderr, "%s:  Unable to write image.\n", argv[0]);
		goto onerror;
	    }
	} 
	else {
	    Xil_unsigned8* cur_line = storage.byte.data;
	    Xil_unsigned8* pixel;
	    int w, h;
	    for (h = 0; h < desired_height; h++) {
		pixel = cur_line;
		for (w = 0; w < desired_width; w++ ) {
		    if (write(outfile_fd, storage.byte.data, frame_size) < 0){
			fprintf(stderr, "%s:  Unable to write image.\n", argv[0]);
			goto onerror;
		    }
		    pixel += storage.byte.pixel_stride;
		}
		cur_line += storage.byte.scanline_stride;
	    }
	}
	xil_import(dst_image,FALSE);

    if (frames == 0) /* get start time after 1st pass */
        gettimeofday(&t0, NULL);

    
    }  
    
    gettimeofday(&t1, NULL);
    
    /*-----------------------------------------------
     * Calculating the time taken to write the frames
     *-----------------------------------------------*/
    time = (t1.tv_sec - t0.tv_sec) + (t1.tv_usec - t0.tv_usec) * 1e-6;
    printf("%d frames in %g seconds, %g frames/second\n",
	   frames, time, (frames - 1) / time);
    
    /*------------------------------------
     * Free the buffer and do more Cleanup
     *------------------------------------*/
 onerror:
    for ( i = 0; i < 3; i++)
	if (cmap[i]) free(cmap[i]);
    if (colorcube855) xil_lookup_destroy(colorcube855);
    if (yuv_to_rgb) xil_lookup_destroy(yuv_to_rgb);
    if (outfile_fd > 0) close(outfile_fd);
    if (dev_image) xil_destroy(dev_image);
    if (dst_image) xil_destroy(dst_image);
    if (child_dev_image) xil_destroy(child_dev_image);
    if (child_dst_image) xil_destroy(child_dst_image);
    if (rescaled_image) xil_destroy(rescaled_image);
    if (scaled_image) xil_destroy(scaled_image);
    xil_close(state);
}

/*==========================================================
 * Function:    initSunRasterHeaderInfo
 * Description: Initializing the sun raster file information
 *              for the selected format. 
 *==========================================================*/
void initSunRasterHeaderInfo(struct rasterfile* ras_hdr,
			     int frame_width,
			     int frame_height,
			     int frame_size,
			     int cmap_len,
			     int format)
{
    /*-------------------------------
     * Some common header information
     *-------------------------------*/
    ras_hdr->ras_magic = RAS_MAGIC;
    ras_hdr->ras_width = frame_width;
    ras_hdr->ras_height = frame_height;
    
    switch (format) {
    case DITHERED8_FORMAT:
    case GRAYSCALE_FORMAT:
	ras_hdr->ras_depth = 8;
	ras_hdr->ras_length = frame_size;
	ras_hdr->ras_type = RT_STANDARD;
	ras_hdr->ras_maptype = RMT_EQUAL_RGB;
	ras_hdr->ras_maplength = (cmap_len * 3);
	break;
	
    case RGB24BIT_FORMAT:
	ras_hdr->ras_depth = 24;
	ras_hdr->ras_length = frame_size;
	ras_hdr->ras_type = RT_STANDARD;
	ras_hdr->ras_maptype = RMT_NONE;
	ras_hdr->ras_maplength = 0;
	break;
	
    case RGB32BIT_FORMAT:
	ras_hdr->ras_depth = 32;
	ras_hdr->ras_length = frame_size;
	ras_hdr->ras_type = RT_STANDARD;
	ras_hdr->ras_maptype = RMT_NONE;
	ras_hdr->ras_maplength = 0;
	break;

    default:
	break;
    }
}

/*==============================================================
 * Function:    writeSunRasterHeaderInfo 
 * Description: Writes the sun raster header info into the file.
 *              return -1 for failure.
 *              return 0 or positive integer for success.
 *==============================================================*/
int writeSunRasterHeaderInfo(int fd, 
			     struct rasterfile* ras_hdr) 
{
    return write(fd,ras_hdr,sizeof(*ras_hdr));
}

/*=======================================================
 * Function:    writeColorMapInfo 
 * Description: Writes the color map info to the file.
 *              return -1 for failure.
 *              return 0 or positive integer for success.
 *=======================================================*/
int writeColorMapInfo(int fd, 
		      unsigned char* cmap[],
		      int cmap_len)
{
    int i, status=0;
    
    if (cmap_len) {
	for (i = 0; i < 3; i++) {
	    if ((status = write(fd, cmap[i], cmap_len)) < 0)
		return status;
	}
    }
    return status;
}

/*=============================================
 * Function:    createGrayScaleCMap 
 * Description: Creates the grayscale colormap.
 *              return -1 for failure.
 *              return 0 for success.
 *=============================================*/
int createGrayScaleCMap(unsigned char* cmap[],
			int cmap_len, 
			float gamma)
{
    int i;
    for (i = 0; i < 3; i++) {
	/* Allocating memory for the cmap buffer. */
	if ((cmap[i] = (unsigned char *)malloc(cmap_len)) == NULL)
	    return FAILURE;
    }
    /* Storing the colormap into the buffer. */
    for (i = 0; i < cmap_len; i++){
	cmap[0][i] = cmap[1][i] = cmap[2][i] = (int)(0.5+255.0*pow(i/255.0,gamma));
    }
    return SUCCESS;
}

/*==========================================================
 * Function:    createDitheredCMap 
 * Description: Creating the color map for Dithered8 format.
 *              return -1 for failure.
 *              return 0 for failure.
 *==========================================================*/
int createDitheredCMap(XilLookup colorcube855, 
		       XilLookup yuv_to_rgb,
		       unsigned char* cmap[],
		       int cmap_len)
{
    Xil_unsigned8 cmap_data[MAX_CMAP_LEN * 3];
    int ncolors;
    int base;
    int i,j;
    /*
     * Get a standard 8x5x5 colorcube to use with xil_ordered_dither,
     * and copy the lookup offset from yuv_to_rgb to this colorcube.
     */
    xil_lookup_set_offset(colorcube855, xil_lookup_get_offset(yuv_to_rgb));
    
    /* Getting the number of entries for the colormap. */
    ncolors = xil_lookup_get_num_entries(yuv_to_rgb);
    base = xil_lookup_get_offset(yuv_to_rgb);
    
    /* Copying the look-up table values into the buffer cmap_data. */
    xil_lookup_get_values(yuv_to_rgb, base, ncolors, cmap_data);
    
    for (i = 0; i < 3; i++) {
	/* Allocating memory for the cmap buffer. */
	if ((cmap[i] = (unsigned char *)malloc(cmap_len)) == NULL)
	    return FAILURE;
    }
    
    /* Initialize the buffer to zeroes. */
    memset(cmap[0], 0, cmap_len);
    memset(cmap[1], 0, cmap_len);
    memset(cmap[2], 0, cmap_len);
    
    /* Storing the yuv_to_rgb colormap into the buffers 
     * in the order of RGB before writing it into the file. */
    for (i=0,j=0; i < ncolors; i++,j+=3) {
	cmap[0][i+base] = cmap_data[j+2];
	cmap[1][i+base] = cmap_data[j+1];
	cmap[2][i+base] = cmap_data[j];
    }
    return SUCCESS;
}

/*==============================================
 * Function:    usage
 * Description: Prints the user input interface.
 *==============================================*/
void usage(char *progName)
{
    fprintf(stderr, "usage:  %s [options] -o <filename>\n", progName);
    fprintf(stderr, "Legal Options:                               Default\n");
    fprintf(stderr, "   -b <bright>   : brightness (0<b<256)     not set \n");
    fprintf(stderr, "   -c <cont>     : contrast   (0<c<256)     not set \n");
    fprintf(stderr, "   -D <dev name> : device name             /dev/o1k0\n"); 
    fprintf(stderr, "   -T <dev type> : device type               MMACo1k\n");     
    fprintf(stderr, "   -f <format>   : output format           GRAYSCALE\n");
    fprintf(stderr, "      (GRAYSCALE=0,DITHERED8=1,RGB24BIT=2,RGB32BIT=3\n"); 
    fprintf(stderr, "   -g <gamma>    : set gamma value              0.7 \n");
    fprintf(stderr, "   -I <port>     : input video port (0|1|2)        1\n");
    fprintf(stderr, "   -w <width>    : output width                  640\n");
    fprintf(stderr, "   -h <height>   : output height                 480\n");
    fprintf(stderr, "   -n <count>    : frame count;                  100\n"); 
    fprintf(stderr, "   -i <skip>     : frames to skip between capture  0\n");
    fprintf(stderr, "   -G            : disable degamma table\n");
}

