/* $Id: loopback.cpp 1.1 1999/01/25 17:30:45 scl Exp $ */
/****************************************************************
 ** File: File Long Name ($RCSfile: loopback.cpp $)
 **
 ** ($Revision: 1.1 $)
 **
 ** Abstract:
 **   Contains the implementations for the loopback operations.
 **   For this loopback example, we have chosen to use the first
 **   instance of the device for video capture & video compression
 **   and chosen the last instance of the device for decompression.
 **   Note: If you only have 1 osprey1000 card installed in your system, 
 **   then the same card will be used for the loopback operations.
 **   
 ** Copyright (c) 1998, MultiMedia Access Corp
 **                     Osprey Technologies Division
 ****************************************************************/
#include <iostream.h>
#include <string.h>
#include <stdio.h>
#include "loopback.h"

/****************************************************************
 ** Method: Loopback  (CLoopback::CLoopback)
 **
 ** Abstract:
 **  Constructor
 **
 ** Error Conditions: None
 **
 ***************************************************************/
CLoopback::CLoopback()
{
  pipeline  = NULL;
  pipeline2 = NULL;
  localWin  = NULL;
  decompWin = NULL;

  localview  = FALSE;
  farview    = FALSE;
  splitpipeline = FALSE;

  // Open the OPI system.
  opi = new OPISystem();
  
  // setDebugLevel will control the amount of debug information 
  // provided by OPI system. Currently, the only effect 
  // this has is disabling detection of streaming timeouts
  // if passed a non-zero value.
  opi->setDebugLevel(1);
}

/****************************************************************
 ** Method: Loopback  (CLoopback::~CLoopback)
 **
 ** Abstract:
 **  Destructor
 **
 ** Error Conditions: None
 **
 ***************************************************************/
CLoopback::~CLoopback()
{
  if(pipeline) {
    pipeline->stop();
    delete pipeline;
  }
  if(splitpipeline && pipeline2) {
    pipeline2->stop();
    delete pipeline2;
  }
  
  if (localWin)
    delete localWin;
  
  if (decompWin)
    delete decompWin;
  
  if (opi)
    delete opi;
}

OPIStream* inVidStream = NULL;

void CALLBACK LoopbackBuffer(OPIBuffer *buffer, void *encodePipeline)
{
  OPIBuffer *decodeBuffer;
#if 1
  // Get a buffer from decode stream
  do {
    decodeBuffer = inVidStream->getBuffer();
    if (!decodeBuffer) 
      OPISleep(5);
  } while (!decodeBuffer);

  // Copy data into decode buffer
  memcpy(decodeBuffer->data, buffer->data, buffer->fsize);
  decodeBuffer->fsize = buffer->fsize;
  // Send buffer to decode stream
  inVidStream->addData(decodeBuffer);
#endif

  // Return buffer from encode stream
  ((OPIPipeline*)encodePipeline)->getOutStream(NULL)->returnBuffer(buffer);
}

void CALLBACK LoopbackDoneBuffer(OPIBuffer *buffer, void *decodePipeline)
{
  ((OPIPipeline*)decodePipeline)->getInStream()->returnBuffer(buffer);
}

/****************************************************************
 ** Method: Loopback  (CLoopback::Loopback)
 **
 ** Abstract:
 **   This loopback function uses the first instance of the device type
 **   to perform video capture & video compression and then uses the last 
 **   instance of the device type to perform decompression.
 **   If you only have 1 osprey1000 card installed in your system, 
 **   then the same card will be used for the loopback operations.
 **
 ** Error Conditions: 
 **   No o1k device installed.
 **   Could not create encode pipeline.
 **   Could not create OPI window.
 **   Could not map to pipeline. 
 **
 ***************************************************************/
int CLoopback::Loopback()
{

  cout << "local/far view: " << localview << "/" << farview << " split: " << splitpipeline << "\n";
  printf("local/far %d/%d split %d\n",localview, farview, splitpipeline);

  // Check whether o1k device is present within system and get pointer.
  OPIDevice* dp = opi->getDeviceByName("o1k");
  if(dp==NULL) {
    cerr << "No o1k device installed\n";
    return APP_ERROR;
  }

  // Start creating the compression pipeline.
  if ((pipeline = new OPIPipeline(opi)) == NULL) {
    cerr << "Could not create encode pipeline.\n";
    return APP_ERROR;
  }
  
  // Use the first instance of the device type for capturing uncompressed
  // video.
  // Get the pointer to the first instance of the device.
  OPIDeviceInstance* inst = dp->getInstanceByIndex(0);

  // Get the name of the first instance of the device.
  char* selected_device = inst->getName();

  // Setting up the external format variables for the first pipeline stage.
  OPIFormatVideo externalFmt;
  externalFmt.compression   = VIDEO_FORMAT;
  externalFmt.deviceName    = selected_device;
  
  // Setting up uncompressed format variables for the first pipeline stage.
  OPIFormatVideo uncompressedFmt;
  uncompressedFmt.width       = VIDEO_WIDTH;
  uncompressedFmt.height      = VIDEO_HEIGHT;
  uncompressedFmt.depth       = 16;
  if (localview || farview)
    uncompressedFmt.compression = "RGB565";  
  else 
    uncompressedFmt.compression = "YUV422"; 
  uncompressedFmt.deviceName  = externalFmt.deviceName;
  
  // Creating the first pipeline stage which converts the 
  // captured image into RGB565.
  OPIStage* sh;
  sh = pipeline->firstStage(CAP_VID_CAPTURE, &externalFmt, &uncompressedFmt);
  
  // Setting up the compression format variables for the compression stage.
  OPIFormatVideo compressedFmt;
  compressedFmt.compression = "H263P";
  compressedFmt.width       = VIDEO_WIDTH;
  compressedFmt.height      = VIDEO_HEIGHT;
  compressedFmt.depth       = 16;
  compressedFmt.deviceName  = externalFmt.deviceName;
  
  // Creating the compression stage of the pipeline to convert RGB565 to compressed h263p.
  OPIStage* videoCompressionStage;
  videoCompressionStage = pipeline->addStage(sh,CAP_VID_COMPRESSION, &uncompressedFmt, &compressedFmt);

  OPIFormatVideo windowFmt;  
  // Setting up the window format variables for the window display for the 
  // uncompressed image that was captured by the first device instance.
  windowFmt.compression    = "WINDOW";
  windowFmt.externalOutput = FALSE;
  windowFmt.width          = WINDOW_WIDTH;
  windowFmt.height         = WINDOW_HEIGHT;
  windowFmt.depth          = 16;

  if (localview){
    // Create a separate OPI window to display local raw video.
    if((localWin = new OPIWindow(10,0,windowFmt.width, windowFmt.height, "Local raw video")) == NULL) {
      cerr << "Could not create opi window.\n";
      return APP_ERROR;
    }
    windowFmt.param = (void *)localWin;
    
    // Create branch from first pipeline stage for viewing captured video.
    pipeline->addStage(sh, CAP_VID_OUTPUT, &uncompressedFmt, &windowFmt);
  }

  // Set the uncompressedFmt device name to the last instance of the device type.
  // Note: If you have only 1 osprey1000 card installed in your system, then the device 
  //       name would be the same as that which was used previously.
  //       Else if you have more than 1 osprey1000 card installed in your system, 
  //       then the device name would be the last osprey1000 card found in your system.

  // Getting the number of instances of the device
  int num_of_instances = dp->instances.getNumItems();		
  
  // Get the pointer to the last instance of the device.
  inst = dp->getInstanceByIndex(num_of_instances - 1);

  // Get the name of the last instance of the device.
  selected_device = inst->getName();

  uncompressedFmt.deviceName = selected_device;
  
  // Creating the decompression stage of the pipeline.

  OPIStage* videoDecompressionStage;
  
  if (!splitpipeline){
    pipeline2 = pipeline;
    videoDecompressionStage = pipeline->addStage(videoCompressionStage, CAP_VID_DECOMPRESSION, &compressedFmt, &uncompressedFmt);
  } else {
    if ((pipeline2 = new OPIPipeline(opi)) == NULL) {
      cerr << "Could not create decode pipeline.\n";
      return APP_ERROR;
    }
    videoDecompressionStage = pipeline2->firstStage(CAP_VID_DECOMPRESSION, &compressedFmt, &uncompressedFmt);
  }
  
  if (farview){
    // Creating the window to display the decompressed view.
    if((decompWin = new OPIWindow(10,300,windowFmt.width,windowFmt.height,"Decompressed video")) == NULL) {
      cerr << "Could not create OPI window.\n";
      return APP_ERROR;
    }
    windowFmt.param = (void *)decompWin;
    
    // Creating the viewing stage of the pipeline for the viewing of the decompressed video.
    pipeline2->addStage(videoDecompressionStage, CAP_VID_OUTPUT, &uncompressedFmt, &windowFmt);
  } else {
    windowFmt.compression    = "NTSC";
    windowFmt.externalOutput = TRUE;
    windowFmt.width          = WINDOW_WIDTH;
    windowFmt.height         = WINDOW_HEIGHT;
    windowFmt.depth          = 16;
    pipeline2->addStage(videoDecompressionStage, CAP_VID_OUTPUT, &uncompressedFmt, &windowFmt);
  }

  // Assigning the actual hardware and software to 
  // each of the required actions. 
  if (!pipeline->mapToDevices()) { 
    cerr << "Could not map to pipeline.\n";
    return APP_ERROR;
  }
  if (splitpipeline){
    if (!pipeline2->mapToDevices()) { 
      cerr << "Could not map to pipeline2.\n";
      return APP_ERROR;
    }
  }

  // Getting the specific compression device and setting some of the
  // capabilities of the compression device.
  OPICompressionDeviceH263 *cdH263;
  cdH263 = (OPICompressionDeviceH263 *)pipeline->getDeviceFromHandle(videoCompressionStage);
  cdH263->setQuality(max_q);
  cdH263->setBitRate(bitrate);
  
  // Getting the specific capture device and setting the framerate and the overlay rate.
  OPICaptureDeviceVideo *cap = (OPICaptureDeviceVideo *)pipeline->getDeviceFromHandle(videoCompressionStage);
  cap->setFrameRate(33);       // frame rate is measured in msec per frame
  cap->setOverlayRate(66);   // Set frame rate for local video
  
  // Finish the construction of the pipeline.
  pipeline->finishConstruction();
  if (splitpipeline) {
    pipeline2->finishConstruction();

    // set callback on encode pipeline
    pipeline->getOutStream(NULL)->setDataReadyCallback(&LoopbackBuffer, pipeline);

    // set callback on decode pipeline
    inVidStream = pipeline2->getInStream();
    inVidStream->setDataDoneCallback(&LoopbackDoneBuffer, pipeline2);
  }

  // Start the pipeline.
  pipeline->start();
  if (splitpipeline)
    pipeline2->start();

  return APP_OK;
}

