/* $Id: opi_viddsp.cpp 1.8 1998/07/24 18:26:58 ott Initial_Build $ */
/*****************************************************************************
 ** File: OPI Capture and Display ($RCSfile: opi_viddsp.cpp $)
 **
 ** ($Revision: 1.8 $)
 **
 ** Abstract:
 **   This OPI example demonstrates video display. A video display pipeline
 **   is created and is then started. The pipeline captures video and 
 **   displaysthis video in a window.
 **
 ** Revision History:
 **   $Log: opi_viddsp.cpp $
 **   Revision 1.8  1998/07/24 18:26:58  ott
 **   Minor simplification to rqFrameRate.
 **   Revision 1.7  1998/07/22 18:02:28  ott
 **   Expanded and standardized command line options.
 **   Minor changes to eliminate warning messages when compiling.
 **   Revision 1.6  1998/07/12 18:05:24  brettb
 **   changed -f color_format to -c to match other example programs.
 **   Revision 1.5  1998/07/02 15:32:28  brettb
 **   coding standards.
 **
 ** Copyright (c) 1998, MultiMedia Access Corp. - Osprey Systems Division
 ****************************************************************************/

#include <stdio.h>
#include <stropts.h>
#include <stdlib.h>
#include <string.h>
#include "opi.h"
#include <X11/Xutil.h>

OPIPipeline* capturePipeline = NULL;

static OPISystem*   msys   = NULL;
static OPIWindow*   window = NULL;

// Set default values here
static char *colorFormat = "RGB32";
static char *capFormat   = "NTSC";
static int  colorDepth   = 24;
static int  maxWidth     = 640; 
static int  maxHeight    = 480;
static int  sizeDivisor  = 2;
static int  nframes      = 100;
static int  rqFrameRate  = 33; // framerate
static char *rqDevName   = NULL;

static char *getVidCapDevname(OPISystem *msys);
static void  setFrameRate(OPIDeviceInstance* dev, int rate);
static int   StartCodec(char* color_format);
static int   StopCodec(void);

/*****************************************************************************
 ** Function: Stop Codec (StopCodec)
 **
 ** Abstract:
 **   This function stops the OPI pipeline and shuts down opi.
 **
 ** Error Conditions: None
 **
 ****************************************************************************/
static int StopCodec(void)
{
  
  // shutdown the pipeline
  if (capturePipeline)
  {
    capturePipeline->stop();  
    delete capturePipeline;
    capturePipeline = NULL;
  }
  
  // shut down OPI 
  if (msys)
  {
    delete msys; 
    msys = NULL;
  }
  
  return 0;
}

/*****************************************************************************
 ** Function: Get Video Capture Device Name ( getVidCapDevname)
 **
 ** Abstract:
 **   Look for the requested devicename,  if none requested then
 **   return first availible.  if none availible then return NULL;
 **
 ** Error Conditions: None
 **
 ****************************************************************************/
char *getVidCapDevname(OPISystem *msystem)
{

  // try o1k device
  OPIDevice *dpO1k = msystem->getDeviceByName("o1k");
  if ((dpO1k != NULL) && (dpO1k->instances.getNumItems() > 0))
  {
    return  dpO1k->instances.getItemName(0);
  }

  // try o1c device
  OPIDevice *dpO1c = msystem->getDeviceByName("o1c");
  if ((dpO1c != NULL) && (dpO1c->instances.getNumItems() > 0))
  {
    return dpO1c->instances.getItemName(0);
  }

  return  NULL;
}

/*****************************************************************************
 ** Function: Set Frame Rate ( setFrameRate)
 **
 ** Abstract:
 **   Sets the frame rate for the capture stage.
 **
 ** Error Conditions: 
 **
 ****************************************************************************/
static void  setFrameRate(OPIDeviceInstance* dev, int rate)
{
  OPICaptureDeviceVideo *o1c = (OPICaptureDeviceVideo *) dev;
  o1c->setFrameRate(rate);
  rqFrameRate = o1c->getFrameRate();
}


/*****************************************************************************
 ** Function: Start Codec ( StartCodec)
 **
 ** Abstract:
 **   Initialzes OPI,
 **
 ** Error Conditions: 
 **   Could not create an OPISystem
 **   Could not assign OPI video capture device
 **   Could not create an OPIPipeline
 **   Could not create viewing window
 **   Could not MAP pipeline
 **
 ****************************************************************************/
int StartCodec()
{
  char *devName;

  // Create a new OPI System
  if ((msys = new OPISystem()) == NULL){
    fprintf(stderr,"Could not create an OPISystem\n");
    return -1;
  }
  
  // get video capture device name
  if (rqDevName)
    devName = rqDevName;
  else
    devName = getVidCapDevname(msys);
  
  printf("Capture Device Name is '%s'\n", devName);
    
  // Create the pipeline
  if ((capturePipeline = new OPIPipeline(msys)) == NULL) {
    fprintf(stderr,"Could not create an OPIPipeline\n");
    return -1;
  }

  // Set the format for the input of capture
  OPIFormatVideo inFmt;
  inFmt.compression   = capFormat;
  inFmt.deviceName    = devName;
  inFmt.externalInput = TRUE;

  // Set the format for the output of capture  
  OPIFormatVideo outFmt;
  outFmt.width  = maxWidth  / sizeDivisor;
  outFmt.height = maxHeight / sizeDivisor;
  outFmt.depth  = colorDepth;
  outFmt.compression = colorFormat;
  
  // Create the first stage in our capture pipeline
  OPIStage*    captureStage;
  captureStage = capturePipeline->firstStage(CAP_VID_CAPTURE, &inFmt, &outFmt);
  
  // Also create branch of the pipeline for viewing captured video
  inFmt.width  = outFmt.width;
  inFmt.height = outFmt.height;
  inFmt.depth  = outFmt.depth;
  inFmt.compression  = colorFormat;
  outFmt.compression = "WINDOW";
  outFmt.externalOutput = FALSE;

  // create a window 
  window = new OPIWindow(0,0,inFmt.width,inFmt.height,"Raw Video"); 
  outFmt.param = window;

  if (!window)
  {
    fprintf(stderr,"Could not create viewing window\n");
    return -1;
  }

  // Add the display stage
  OPIStage*    displayStage;
  displayStage = capturePipeline->addStage(captureStage,
					   CAP_VID_OUTPUT, &inFmt, &outFmt);

  // Now MAP the pipeline to the device
  if (!capturePipeline->mapToDevices()) {  
    fprintf(stderr,"Could not MAP pipeline\n");
    return -1;
  }

  // Set Frame rate
  setFrameRate(capturePipeline->getDeviceFromHandle(captureStage),
		 rqFrameRate);
  
  // Finish the pipeline construction
  capturePipeline->finishConstruction();

  // Start the pipeline
  capturePipeline->start();

  return 0;
}

/*****************************************************************************
 ** Function: Display Program Usage ( usage )
 **
 ** Abstract:
 **   Prints out the valid command line arguments and then exits.
 **
 ** Error Conditions: None
 **
 ****************************************************************************/
void usage(char* appName)
{
  fprintf(stderr, "usage: %s [options]\n",appName);
  fprintf(stderr, "Legal Options |       Description           | Defaults \n");
  fprintf(stderr, "--------------|-----------------------------|----------\n");
  fprintf(stderr, "-d devname    | device name                 |   o1c0   \n");
  fprintf(stderr, "-c format     | color format                |   RGB32  \n");
  fprintf(stderr, "-v capformat  | capture format              |   NTSC   \n");
  fprintf(stderr, "-z depth      | color depth                 |   24     \n");
  fprintf(stderr, "-w width      | width                       |   320    \n");
  fprintf(stderr, "-h height     | height                      |   240    \n");
  fprintf(stderr, "-s sizediv    | size divisor                |   2      \n");
  fprintf(stderr, "-n framecount | number of frames to capture |   100    \n");
  fprintf(stderr, "-r framerate  | requested frame rate        |   33     \n");
  fprintf(stderr, "-o options    | list command line options   |          \n");
  exit(0);
}

/*****************************************************************************
 ** Function: Start Video capture and display ( main)
 **
 ** Abstract:
 **   This program demonstrates a simple capture and display using
 **   OPI.
 **
 ** Error Conditions: None
 **   StartCodec failed. Closing down
 **
 ****************************************************************************/
int main(int argc, char** argv)
{
  int   c;

  // Process command line arguments
  while ((c = getopt(argc, argv, "d:c:v:z:w:h:s:n:r:o:")) != EOF)
  {
    switch (c)
    {
      case 'd': rqDevName    = optarg;                              break;
      case 'c': colorFormat  = optarg;                              break;
      case 'v': capFormat    = optarg;                              break;
      case 'z': colorDepth   = atoi(optarg);                        break; 
      case 'w': maxWidth     = atoi(optarg);    sizeDivisor = 1;    break;
      case 'h': maxHeight    = atoi(optarg);    sizeDivisor = 1;    break;
      case 's': sizeDivisor  = atoi(optarg);                        break;
      case 'n': nframes      = atoi(optarg);                        break;
      case 'r': rqFrameRate  = atoi(optarg);                        break;
      case 'o': usage(argv[0]);
      default: usage(argv[0]);
      break;
    }
  }

  // Create/Start OPI Pipelines
  if (StartCodec() != 0)
  {
    fprintf(stderr,"StartCodec failed. Closing down\n");
    StopCodec(); 
    exit(0);
  }

  // Sleep until desired number of frames acquired
  while (nframes--) 
    OPISleep(rqFrameRate);

  // Stop OPI Pipelines 
  StopCodec();

}








 
