/* ************************************************************************* 
 *                       codec.c
 *--------------------------------------------------------------------------
 *         rights reserved by Bo-kyun Na (bkna@top.cis.syr.edu)
 ***************************************************************************
 *  This file includes functions saving the captured XIL compressed image 
 * seqence (CIS) into a file.
 * *************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <xil/xil.h>

#include "vdo.h"

/* - - - - - - - - - - - - - - - - - - - - - - */
/* - - - - - - - - - - - - - - - - - - - - - - */
void write_frames(FILE *df, XilCis cis)
{
  Xil_unsigned8 *data;
  int nframes, nbytes;
  
  while (xil_cis_has_frame(cis)) {
    data = (void *) xil_cis_get_bits_ptr(cis, &nbytes, &nframes);
    
    if (df)
      fwrite(data, 1, nbytes, df);
  }
}

/* - - - - - - - - - - - - - - - - - - - - - - */

int compressor (XilSystemState state, IMAGE *img, CIS *cis,
		VDOTYPE *vdo, IMGPROC *imgproc)
{
  FILE *df;
  char pattern[32];

  int i;
  int c;
  int format;
  int frames;

  cis->h261.bitrate = 256;
  cis->h261.mquant  = 10;  
  
  cis->mpeg.bitrate = 2000;
  cis->mpeg.rate_control = 1;
  cis->mpeg.fixed_quant = 10;

  strcpy(pattern, "IIII");
                       
  df = fopen(cis->cisfile, "w");
  if (df == NULL) {
    fprintf(stderr, "failed to open output movie file.\n");
    exit(1);
  }

  /* get image attribute */
  xil_get_device_attribute(img->handle, "FORMAT_V", (void **) &format);
  if (format == 0) {
    fprintf(stderr, "Unknown video format, exiting.\n");
    xil_close(state);
    exit(1);
  }

  /* compression of image */
  if (cis->do_cis == H261) {
    xil_set_device_attribute(img->handle, "H261_BIT_RATE", 
			     (void *) cis->h261.bitrate);
    xil_set_device_attribute(img->handle, "H261_MAX_QUANT",
			     (void *) cis->h261.mquant);
  }
  
  if (cis->do_cis == MPEG) {
    xil_set_device_attribute(img->handle, "MPEG_RATE_CONTROL", 
			     (void *) cis->mpeg.rate_control);
    xil_set_device_attribute(img->handle, "MPEG_FIXED_QUANT", 
			     (void *) cis->mpeg.fixed_quant);
  }

  if (vdo->image_skip)
    xil_set_device_attribute(img->handle, "IMAGE_SKIP",
                             (void *) vdo->image_skip);
  
  xil_get_info(img->handle, &(img->width), &(img->height), &(img->nbands), 
	       &(img->datatype));

  if (cis->do_cis == H261) {
    if (imgproc->scale.doscale == 2){
      img->width = 352;
      img->height = 288;
    } else if (imgproc->scale.doscale == 4) {
      img->width = 352/2;
      img->height = 288/2;
    } else {
      fprintf(stderr,"Invalid scale factor for H261 Compression: only 2 and 4 legal\n");
      xil_close(state);
      exit(1);
    }
    img->scaled = xil_create(state, img->width, img->height, img->nbands,
			     XIL_BYTE);
  } else {
    if (imgproc->scale.doscale) {
      img->width = (int) (((float) img->width / imgproc->scale.doscale) + 0.5);
      img->height = (int) (((float) img->height / imgproc->scale.doscale) + 0.5);
      img->scaled = xil_create(state, img->width, img->height, img->nbands, 
			       XIL_BYTE);
    }
  }

  /* MPEG requires CIF (352 width) compression */
  if (cis->do_cis == MPEG) {
      if (imgproc->scale.doscale != 2) {
          fprintf(stderr, "ERROR: scalefactor is: %d, should be 2 for Mpeg1\n",
                  imgproc->scale.doscale);
          xil_close(state);
          exit(1);
      }
      img->width  = 352; /* Osprey Mpeg1 compression is 601 not square */
      img->height = 240;
      img->scaled = xil_create(state, img->width, img->height, img->nbands, 
			       XIL_BYTE);
  }

  /* Create a cis to put the movie frames in. */
  cis->cis = xil_cis_create(state, cis->cis_type);

  /*
   * set the buffers to make sure that you don't keep information you
   * do not need.  H261 needs more frames.
   */
  if (cis->do_cis == H261) {
    xil_cis_set_keep_frames(cis->cis, 3);
    xil_cis_set_max_frames(cis->cis, 3);
  } else if (cis->do_cis == MPEG) {
      xil_cis_set_keep_frames(cis->cis, 6);
      xil_cis_set_max_frames(cis->cis, 6);
  } else {
    xil_cis_set_keep_frames(cis->cis, 1);
    xil_cis_set_max_frames(cis->cis, 1);
  }

  if (cis->do_cis == MPEG) {
      int bitrate = 1000 * cis->mpeg.bitrate;
      xil_cis_set_attribute(cis->cis, "COMPRESSOR_BITS_PER_SECOND", 
			    (void*)bitrate);
  }

  /*
   * If JPEG, set an attribute needed for fast decompression of the
   * resulting movie.
   */
  if (cis->do_cis == JPEG)
      xil_cis_set_attribute(cis->cis, "ENCODE_411_INTERLEAVED", (void *) TRUE);
    
  for (frames = 0; frames < vdo->total_frames; frames ++) {
    if (cis->do_cis == H261 || cis->do_cis == MPEG){
      if (format == 2) /* NTSC */
        xil_scale(img->handle, img->scaled, "nearest",
                  img->width/640.0, img->height/480.0);
      else
        xil_scale(img->handle, img->scaled, "nearest",
                  img->width/768.0, img->height/576.0);        
    } 
    else if (imgproc->scale.doscale) {
      xil_scale(img->handle, img->scaled, "nearest",
                1.0 / (float) imgproc->scale.doscale, 
		1.0 / (float) imgproc->scale.doscale);
    }
    else {
      img->scaled = img->handle;
    }
    
    xil_compress(img->scaled, cis->cis);
    xil_cis_sync(cis->cis);
    xil_toss(img->scaled);
    
    write_frames(df, cis->cis);
  }
  
  /* done, empty buffer if any frames remain */
  xil_cis_flush(cis->cis);
  write_frames(df, cis->cis);
  
  xil_close(state);
  
  return 0;
}
