/*------------------------------------------------------------------------
 *
 *  File:	o1k_timer.cc
 *  Revision:	1.2
 *  Last Mod:	17:36:58, 03/01/97
 *
 *  Description:
 *     Implementation of O1kTimer object
 *     
 *----------------------------------------------------------------------*/
#pragma ident	"@(#)o1k_timer.cc	1.2\t97/03/01  "
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/time.h>
#include <fcntl.h>
#include <malloc.h>
#include <memory.h>
#include <unistd.h>

#include "o1k_timer.h"

/*------------------------------------------------------------------------
 *
 *  Function:	O1kTimer::O1kTimer
 *
 *  Description:
 *     Allocate storage for time differences and data values
 *
 *  Parameters:
 *     char* time_name  : name of timer
 *     int ntimes       : number of items to keep track of
 *
 *  Notes:
 *     Failure of creating object is not flagged
 *     
 *----------------------------------------------------------------------*/

O1kTimer::O1kTimer(char* time_name, int ntimes)
{
  num_times = ntimes;

  if ((tdiff = (double*) new double[ntimes]) == NULL)
    fprintf(stderr,"O1kTimer: error could not allocate tdiff\n");
  if ((dval = (int*) new int[ntimes]) == NULL)
    fprintf(stderr,"O1kTimer: error could not allocate dval\n");

  index = 0;
  dval_flag = 0;
  strncpy(name, time_name, 20);
  
  gettimeofday(&t0, NULL); 
}

/*------------------------------------------------------------------------
 *
 *  Function:	O1kTimer::~O1kTimer()
 *
 *  Description:
 *     Delete allocated O1kTimer data
 *     
 *----------------------------------------------------------------------*/

O1kTimer::~O1kTimer()
{
  if (tdiff){
    delete tdiff;
    tdiff = NULL;
  }
  if (dval){
    delete dval;
    dval = NULL;
  }
}

/*------------------------------------------------------------------------
 *
 *  Function:	O1kTimer::timeEnd()
 *
 *  Description:
 *     Set end time
 *     
 *----------------------------------------------------------------------*/

void O1kTimer::timeEnd()
{
  gettimeofday(&t1, NULL);
}

/*------------------------------------------------------------------------
 *
 *  Function:	O1kTimer::timeEnd()
 *
 *  Description:
 *     Set end time and data value
 *
 *  Parameters:
 *     int val   : data value
 *
 *----------------------------------------------------------------------*/

void O1kTimer::timeEnd(int val)
{
  gettimeofday(&t1, NULL);
  dval[index] = val;
  dval_flag = 1;
}

/*------------------------------------------------------------------------
 *
 *  Function:	O1kTimer::timeDiff()
 *
 *  Description:
 *     Set timediff and index to next storage location. If
 *     locations all filled, then output time differences and
 *     data values (if used).
 *     
 *----------------------------------------------------------------------*/

void  O1kTimer::timeUpdate()
{
  tdiff[index] = (t1.tv_sec - t0.tv_sec) + (t1.tv_usec - t0.tv_usec) * 1e-6;
  index++;
  
  if (index == num_times){
    int i;
    
    printf("%s times: ",name);
    for (i=0; i<num_times; i++)
      printf("%.3g ",tdiff[i]);
    printf("\n");

    if (dval_flag){
      printf("  data: ");
      for (i=0; i<num_times; i++)
        printf("%d ",dval[i]);
      printf("\n");
    }
    
    index = 0;
  }
}

/*------------------------------------------------------------------------
 *
 *  Function:	O1kTimer::timeDiff()
 *
 *  Description:
 *     return time diff from start to end time in msecs
 *     
 *----------------------------------------------------------------------*/

double O1kTimer::timeDiff()
{
  double tmp;
  tmp = (t1.tv_sec - t0.tv_sec) + (t1.tv_usec - t0.tv_usec) * 1e-6;
  return tmp;
}

/*------------------------------------------------------------------------
 *
 *  Function:	O1kTimer::timeStart()
 *
 *  Description:
 *     Set start time
 *     
 *----------------------------------------------------------------------*/

void O1kTimer::timeStart()
{
  gettimeofday(&t0, NULL);
}



