/* Copyright (c) 1997 MultiMedia Access Corporation - Osprey Systems Division. 
 * 1150 S.E. Maynard Rd., Suite 110, Cary, NC 27511 USA. All Rights Reserved.
 *
 * This product and related documentation is protected by copyright laws and
 * international treaties; however, the user of this software is granted a 
 * nonexclusive, royalty free license to modify, make derivitive works, use,
 * distribute, or otherwise incorporate this software in products which use
 * or incorporate the MMAC product.
 *
 * MMAC reserves the right to change any products herein at any time and 
 * without notice. MMAC does not assume any liability arising from the 
 * application or use of this product, except as expressly agreed to in 
 * writing by MMAC. Use and purchase of this product does not convey a license
 * under any patent rights, copyrights, trademark rights, or any other 
 * intellectual property rights of MMAC, Sun Microsystems, or others. 
 */

/*-------------------------------------------------------------------
 * File:         xil_play.c
 *
 * Description:  This xil_play program reads video images from a file
 *               and plays it back in an xwindow display. This example 
 *               uses the Sun Rasterfile format as an illustration. 
 *               However, the developer may choose to implement other 
 *               formats. The example formats that are supported by this 
 *               example program are:
 *               a) GRAYSCALE
 *               b) DITHERED8 (A 24bit color image that is 
 *                             dithered to an 8bit color image)
 *               c) RGB24BIT  (A 24bit color image)
 *               d) RGB32BIT  (A 32bit color image)
 *
 *-------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "xil_play.h"


/****************************************
 * Main
 ****************************************/
void main(int argc, char *argv[])
{
    /* Xil Objects */
    XilSystemState state;
    XilImage image = NULL;
    XilImage child_image = NULL;
    XilImage display = NULL;
    XilImage tmp_image = NULL;
    XilLookup lookup_table = NULL;
    XilLookup color_map = NULL;
    XilDitherMask mask;
    XilDataType datatype;
    XilMemoryStorage storage;
    Xil_boolean status;
    Xil_unsigned8 *data = NULL;
    
    /* Xwindow Objects */
    Display *xdisplay;
    Window xwindow;
    XEvent event; 
    
    /* Conversion table for formats to number of bands */
    const int formatToBands[] = {1,1,3,4};
    
    /* Command line arguments */
    const char getoptargs[]= "p:f:i:x:y:r:";
    
    /* Image info variables */
    int width;
    int height;
    int cmap_size;
    int src_image_size;
    int nbands;
    
    /* raster file structure used for this example */
    struct rasterfile ras_hdr;
    
    /* Program Variables */
    unsigned char *cmap[3] = {NULL,NULL,NULL};
    int cmap_len; 
    int in_file;
    int nread;
    int input;
    float time;
    struct timeval t0, t1;
    int display_depth;
    int m,t;
    int i,j;
    
    /* Default values */
    char filename [80]="";
    int counter = 0;
    int rate = 0;
    int winx = 0;
    int winy = 0;
    int xyflag = 0;
    int format =  GRAYSCALE_FORMAT;
    int pause = -1;
    int first_time = 1;
    int pressed = 0;

    /* Handling input from user */
    while ((input = getopt(argc, argv, getoptargs)) != EOF) {
	switch (input) {
	case 'i':
	    strncpy(filename, optarg, 80);
	    break;
	case 'x':
	    winx = atoi(optarg);
	    xyflag = 1;
	    break;
	case 'y':
	    winy = atoi(optarg);
	    xyflag = 1;
	    break;
	case 'f':
	    format = atoi(optarg);
	    break;
	case 'p':
	    pause = atof(optarg);
	    break;
	case 'r':
	    rate = atoi(optarg);
	    break;
	default:
	    fprintf(stderr, "%s:  unknown flag\n", argv[0]);
	    usage(argv[0]);
	    exit(1);
	    break;
	}
    }
    
    /*---------------
     *Opening the XIL 
     *---------------*/
    if((state = xil_open()) == NULL){
	perror("Unable to open an Xil Session.\n");  
	exit(1);
    }
    
    /*---------------
     * Initialization
     *---------------*/
    if(format == RGB32BIT_FORMAT || format == RGB24BIT_FORMAT) {
	color_map = xil_lookup_get_by_name(state,"cc496");
	mask = xil_dithermask_get_by_name(state,"dm443");
    }
    
    /*-------------------- 
     *Open the header file 
     *--------------------*/
    if ((in_file = open(filename, O_RDONLY)) < 0){
	perror("Failed to Open Input File.\n");
	goto cleanup;
    }
    
    /*-------------------------------------------------
     * Reading the header information from the file.
     * Calculate the size of the cmap buffer
     * Reading the colormap information from the file.
     *-------------------------------------------------*/
    if (readSunRasterHeaderInfo(in_file, &width, &height, 
				&src_image_size,&cmap_size) < 0){
	perror("Error Reading file.\n");
	goto cleanup;
    }
    
    cmap_len = cmap_size/3;
    
    if (readColorMapInfo(in_file,cmap,cmap_len) < 0) {
	perror("Unable to read the colormap.\n");
	goto cleanup;
    }
    
    /*---------------------------------------------
     * Xlib Creation & creating the Xwindow display
     *---------------------------------------------*/
    if((xdisplay = XOpenDisplay(NULL)) == NULL)	goto cleanup;
    
    /* Getting the display_depth */
    display_depth = DefaultDepth(xdisplay,DefaultScreen(xdisplay));
    
    /* Creating an xwindow */
    if ((xwindow = XCreateSimpleWindow(xdisplay, DefaultRootWindow(xdisplay),
				       winx,winy,width,height,0,0,0)) == NULL) {
	perror (" Unable to create window.\n");
	goto cleanup;
    }
    	
    /* Setting the titlebar for the xwindow display */
    XSetStandardProperties(xdisplay,xwindow,filename,"xil_play",0,NULL,0,0);
    
    /* if the user tries to shut down the window using the menu */
    if (closedownAtom = XInternAtom(xdisplay, "WM_DELETE_WINDOW", False))
	XSetWMProtocols(xdisplay, xwindow, &closedownAtom, 1);	    

    /* List window events used */
    XSelectInput(xdisplay,xwindow,ExposureMask | KeyPressMask | ButtonPressMask);
    
    /* Make the window visible */
    XMapWindow(xdisplay, xwindow);
    
    /* Wait for the window to be mapped(an Expose event) */
    do
	XNextEvent(xdisplay,&event);
    while(event.xany.type != Expose);
    
    /* Creating the display image */
    display = xil_create_from_window(state, xdisplay, xwindow);
    
    /* Creating the source image*/
    if ((image = xil_create(state,width,height,
			    formatToBands[format],XIL_BYTE)) == NULL){
	perror("Could not create destination image.\n");
	goto cleanup;
    } 
    
    /*------------------------------------------------------------------
     * Creating the tmp images and creating a lookup
     * data that will be needed for the processing
     * of the image data when display depth is 
     * either 8 bit or 24 bit.
     * Setting the respective format's colormap when display depth is 8.
     *------------------------------------------------------------------*/
    if (display_depth == 8) {
	if (format == GRAYSCALE_FORMAT || format == DITHERED8_FORMAT)
	    setColorMap(xdisplay,xwindow,format,NULL,cmap,cmap_len);
	else if (format == RGB24BIT_FORMAT) {
	    /* Create a 1 banded tmp image. */
	    tmp_image = xil_create(state,width,height,1,XIL_BYTE);
	    setColorMap(xdisplay,xwindow,format,color_map,NULL,0);
	}
	/* Create a child image from band 1 to band 3 from the source image.
	 * Create a 1 banded image for use in a display depth of 8 bit. */
	else if (format == RGB32BIT_FORMAT) {
	    child_image = xil_create_child(image,0,0,width,height,1,3); 	  
	    tmp_image = xil_create(state,width,height,1,XIL_BYTE);
	    setColorMap(xdisplay,xwindow,format,color_map,NULL,0);
	}
    }
    /* display depth of 24 */
    else {
	/* Creating a 3 banded tmp image for use in the lookup operation.
	 * Creating the lookup data for use in the creation of a lookup
	 * table. */
	if ((format == DITHERED8_FORMAT)|| (format == GRAYSCALE_FORMAT)) {
	    tmp_image = xil_create(state, width, height,3,XIL_BYTE);
	    if((data = (Xil_unsigned8*)malloc(MAXCOLORS*3)) == NULL) {
		perror("Fail to allocate memory to data buffer.\n");
		goto cleanup;
	    }
	    for (m = 0, t = 0; m < MAXCOLORS; m++, t += 3){
		data[t+2] =  cmap[0][m];
		data[t+1] = cmap[1][m];
		data[t] = cmap[2][m];
	    }
	    /* Creating the lookup table */
	    if ((lookup_table = xil_lookup_create(state,XIL_BYTE,XIL_BYTE,3,
						  MAXCOLORS,0,data)) == NULL) {
		perror("NULL lookup table.\n");
		goto cleanup;
	    }
	}
	/* Creating a child source image from band 1 to band 3
	 * Which will be used to copy into a 3 banded display image*/
	else if (format == RGB32BIT_FORMAT) {
	    child_image = xil_create_child(image,0,0,width,height,1,3); 
	}
    }
    
    /*---------------------------
     * Processing the image data  
     *---------------------------*/
    do{
	/* Getting a pointer to the data */
	if (xil_export(image) == XIL_FAILURE){
	    perror(" Could not export image.\n");
	    goto cleanup;
	}
	if ((status = xil_get_memory_storage(image, &storage)) == FALSE) {
	    perror("Could not get image storage info.\n");
	    goto cleanup;
	}
	/* Reading the data image from the file */
	if ((storage.byte.pixel_stride == formatToBands[format]) &&
	    (storage.byte.scanline_stride == width * formatToBands[format])) {
	    nread = read(in_file, storage.byte.data, src_image_size);
	    if((nread < 0) || (nread < src_image_size)) 
		goto done;
	} 
	/* Otherwise, read the data image pixels by pixels */
	else {
	    Xil_unsigned8* cur_line = storage.byte.data;
	    Xil_unsigned8* pixel;
	    int w, h;
	    for (h = 0; h < height; h++) {
		pixel = cur_line;
		for (w = 0; w < width; w++ ) {
		    nread = read(in_file, pixel, formatToBands[format]);
		    if ((nread < 0) || (nread < formatToBands[format])) {
			goto done;
		    }
		  pixel += storage.byte.pixel_stride;
		}
		cur_line += storage.byte.scanline_stride;
	    }
	}
	xil_import(image,TRUE);
	
	/*---------------------------------------
	 * Start processing the image for display
	 *---------------------------------------*/
	switch(format) {
	case GRAYSCALE_FORMAT:
	case DITHERED8_FORMAT:
	    if(display_depth == 8) {
		/* Copying the source image of 1 band to
		 * the 1 banded display.*/
		xil_copy(image,display);	
	    }
	    else if(display_depth == 24) {
		/* Performing a lookup for the 1 banded source and 
		 * storing it into a 3 banded tmp image which will be 
		 * copied to the display of 3 band. */
		xil_lookup(image, tmp_image, lookup_table);
		xil_copy(tmp_image, display);
		xil_toss(tmp_image);
	    }
	    break;
	    
	case RGB24BIT_FORMAT:
	    if(display_depth == 8) {
		/* Performing an order dithering to dither the 3 banded
		 * source image into a 1 banded tmp_image which will then
		 * be copied into the display of 1 band. */
		xil_ordered_dither(image,tmp_image,color_map,mask);
		xil_copy(tmp_image,display);
		xil_toss(tmp_image);  
	    }
	    else if(display_depth == 24) {
		/* No special processing needed when displaying 
		 * a 3 banded source image to a 3 banded display. */
		xil_copy(image, display);
		xil_toss(image);
	    }
	    break;
	    
	case RGB32BIT_FORMAT:
	    if(display_depth == 8) {
		/* Setting the color map.
		 * Performing an order dithering to dither the 3 banded 
		 * child source image into a 1 banded tmp_image which will
		 *  then be copied into the display of 1 band. */
		xil_ordered_dither(child_image,tmp_image,color_map,mask);
		xil_copy(tmp_image,display);
		xil_toss(tmp_image);  
	    }
	    else if (display_depth == 24) {
		/* Using the child image of band 1 to band 3 created from 
		 * the 4 banded source image, simply copy the 3 banded child
		 * source image into the display of 3 band.*/
		xil_copy(child_image, display);
		xil_toss(child_image);
	    }
	    break;
	    
	default:
	    break;
	}
	/* The xwindow display will terminate when a user presses the mouse
	 * button or any key in the keyboard or selects window quit
	 * while the video is being played back.Then it calculates the time 
	 * taken to read the frames and prints the results to the xterm.
	 */	 
	while(XPending(xdisplay)) {
	    XNextEvent(xdisplay, &event);
	    if (event.xany.type == ClientMessage &&
		((XClientMessageEvent *) & event)->data.l[0] ==
		closedownAtom) {
		pressed++;
		goto done;
	    }
	    else if ((event.xany.type==ButtonPress)||
		     (event.xany.type==KeyPress)) {
		pressed++;
		counter++;
		goto done;
	    }
	}
	
	/* Initializing the time. */
	if(first_time){
	    gettimeofday(&t0, NULL);
	    first_time = 0;
	}
    
    /* Determine the number of frames to playback. */
    rateControl(rate);
    
    if (pause >= 0)
	poll(0,0,pause);
    
    /* Increment the counter. */
    counter++;
    }while((nread > 0)|| (nread > src_image_size));  	
    
 done:
    /* When the last image frame is read, the xwindow display
     * remains mapped and waits for the user to press the mouse
     * button or the keyboard or select window quit
     * to terminate the xwindow display. Then it calculates the
     * time taken to read the frames and prints the results to 
     * the xterm. */
    while(!pressed) {
	XNextEvent(xdisplay, &event);
	if ((event.xany.type == ClientMessage &&
	    ((XClientMessageEvent *) & event)->data.l[0] == closedownAtom) ||
	    (event.xany.type == ButtonPress) || (event.xany.type == KeyPress))
	    pressed++;
    }
	
    /* Calculate the time taken to read the frames. */
    gettimeofday(&t1, NULL);
    time = (t1.tv_sec - t0.tv_sec) + (t1.tv_usec - t0.tv_usec) * 1e-6;
    printf("%d frames in %g seconds, %g frames/second\n",
	   (counter), time, (counter-1) / time);
 cleanup:
    for ( i = 0; i < 3; i++)
	if (cmap[i]) free(cmap[i]);
    if (data) free(data);
    if (image) xil_destroy(image);
    if (child_image) xil_destroy(child_image);
    if (tmp_image) xil_destroy(tmp_image);
    if (display) xil_destroy(display);
    if (lookup_table) xil_lookup_destroy(lookup_table);
    if (color_map) xil_lookup_destroy(color_map);
    close(in_file); 
    xil_close(state);
    XCloseDisplay(xdisplay);    
}

/*==================================================
 * Function:    setColorMap 
 * Description: Setting the colormap to the Xwindow.
 *==================================================*/
void setColorMap(Display *xdisplay, 
		 Window xwindow, 
		 int format,
		 XilLookup colormap,
		 unsigned char *cmap[],
		 int cmap_len)
{
    XVisualInfo vinfo;
    unsigned long pixels[256];
    unsigned long pmask[1];
    unsigned long pixel[1];
    unsigned long mask[256];
    int j;
    int step;
    int ncolors;
    int base;
    int index;
    int i;
    Xil_unsigned8 cmap_data[MAXCOLORS * 3];
    XColor colors[MAXCOLORS];
    Colormap priv_color_map;
    
    if (format == RGB24BIT_FORMAT || format == RGB32BIT_FORMAT) {
	step = xil_lookup_get_output_nbands(colormap);
	ncolors = xil_lookup_get_num_entries(colormap);
	base = xil_lookup_get_offset(colormap);
	xil_lookup_get_values(colormap, base, ncolors, cmap_data);
    }
    else {
	ncolors = cmap_len;
	base = 0;
    }
    /* Creating a colormap. */
    priv_color_map = XCreateColormap(xdisplay, DefaultRootWindow(xdisplay),
				     DefaultVisual(xdisplay,
						   DefaultScreen(xdisplay)), 
				     AllocNone);
        
    /* Allocating the color cells. */
    if (XAllocColorCells(xdisplay, priv_color_map, TRUE, pmask, 0, pixels, 
			 base+ncolors) == 0){
	perror("XAllocColorCells failed\n");
	return;
    }
    
    /* Free the cells represented by pixels whose values 
     * are in the pixels array.*/
    for (pixel[0] = 0; pixel[0] < base; pixel[0]++) 
	XFreeColors(xdisplay, priv_color_map, pixel, 1, 0);
    
    /* Fill the colormap. */
    if (format == RGB24BIT_FORMAT || format == RGB32BIT_FORMAT) {
	for (i = 0, j = step - 3; i < ncolors; i++, j += step){
	    colors[i].pixel = base + i;
	    colors[i].flags = DoRed | DoGreen | DoBlue;
	    colors[i].blue = cmap_data[j] << 8;
	    colors[i].green = cmap_data[j + 1] << 8;
	    colors[i].red = cmap_data[j + 2] << 8;
	}
    }
    else {
	for (i = 0; i < ncolors; i ++) {
	    colors[i].pixel = base + i;
	    colors[i].red = cmap[0][i] << 8;
	    colors[i].green = cmap[1][i] << 8;
	    colors[i].blue = cmap[2][i] << 8;
	    colors[i].flags = DoRed | DoGreen | DoBlue;
	}
    }
    /* Storing the RGB colors into the colormap cells. */
    XStoreColors(xdisplay, priv_color_map, colors, ncolors);
    
    /* Setting the colormap to the window. */
    XSetWindowColormap(xdisplay, xwindow, priv_color_map);
    
    /* Installing the colormap to the display. */
    XInstallColormap(xdisplay, priv_color_map);
}

/*=================================================================
 * Function:    readSunRasterHeaderInfo 
 * Description: Reads the sun rasterfile header info from the file.
 *              return -1 for failure.
 *              return 0 or positive integer for success.
 *=================================================================*/
int readSunRasterHeaderInfo(int fd, 
			    int* width, 
			    int* height, 
			    int* src_image_size, 
			    int* cmap_size) 
{
    struct rasterfile ras_hdr;
    int status = 0;
    
    if ((status = read(fd, &ras_hdr, sizeof(ras_hdr))) < 0)
	return status;
    
    *width = ras_hdr.ras_width;
    *height = ras_hdr.ras_height;
    *src_image_size = ras_hdr.ras_length;
    *cmap_size = ras_hdr.ras_maplength;
    return status;
}

/*===================================================================
 * Function:    readColorMapInfo 
 * Description: Reads the color map information from the opened file.
 *              return -1 for failure.
 *              return 0 or positive integer for success.
 *===================================================================*/
int readColorMapInfo(int fd,
		     unsigned char* cmap[],
		     int cmap_len)
{
    int i;
    int status = 0;
    if (cmap_len) {
	for (i = 0; i < 3; i++) {
	    /* Allocating memory for the cmap buffer. */
	    if ((cmap[i] = (unsigned char *)malloc(cmap_len)) == NULL)
		return -1;
	    if ((status = read(fd, cmap[i], cmap_len)) < 0)
		return status;
	}
    }
    return status;
}

/*=========================================================
 * Function:    rateControl 
 * Description: Determines the number of frames to playback
 *              by waiting some amount of time before the 
 *              the next playback.
 *==========================================================*/
void rateControl(int rate)
{
    static double t0 = 0;
    static double t1 = 0;
    double delay;
    
    if(!rate)
	return;
    
    /* rate is in fps: determine ns delay. */
    delay = (1000000.0/rate);
    t1 = gethrtime()/1000.0;
    
    while(t1-t0 < delay) {
	poll(0,0,2);
	t1 = gethrtime()/1000.0;
    }
    t0 = t1;
}

/*==============================================
 * Function:    usage
 * Description: Prints the user input interface.
 *==============================================*/
void usage(char *name)
{
    fprintf(stderr,"Usage: %s [options] -i <filename>\n", name);
    fprintf(stderr, "Legal Options:                              Default\n");
    fprintf(stderr, "   -f <format>            : output format\n");
    fprintf(stderr, "    (GRAYSCALE=0,DITHERED8=1,RGB24BIT=2, RGB32BIT=3\n"); 
    fprintf(stderr, "   -x <x window location> :                (random)\n");
    fprintf(stderr, "   -y <y window location> :                (random)\n");
    fprintf(stderr, "   -p <pause>             : msec between frames  (0)\n");
    fprintf(stderr,"    -r <playback framerate>:                      (0)\n");                      
}

