/* $Id: loopback.cpp 1.6 1998/07/10 21:14:21 kimng Initial_Build $ */
/*************************************************************
 ** File: File Long Name ($RCSfile: loopback.cpp $)
 **
 ** ($ Revision: $)
 **
 ** Abstract:
 **   Contains the implementations for the loopback operations.
 **   For this loopback example, we have chosen to use the first
 **   instance of the device for audio capture & compression
 **   and chosen the last instance of the device for decompression.
 **   Note: If you only have 1 osprey1000 card installed in your system, 
 **   then the same card will be used for the loopback operations.
 ** 
 **   Copyright (c) 1998, MultiMedia Access Corp
 **                     Osprey Technologies Division
 *************************************************************/
#include <iostream.h>
#include <string.h>
#include "loopback.h"

/****************************************************************
 ** Method: Audio Loopback  (CLoopback::CLoopback())
 **
 ** Abstract:
 **  Constructor
 **
 ** Error Conditions: None
 **
 ***************************************************************/
CLoopback::CLoopback()
{
  pipeline  = NULL;
  devinfo   = NULL;
  
  // Open the OPI system.
  opi = new OPISystem();
  
  // setDebugLevel will control the amount of debug information 
  // provided by OPI system. Currently, the only effect 
  // this has is disabling detection of streaming timeouts
  // if passed a non-zero value.
  opi->setDebugLevel(1);
}

/****************************************************************
 ** Method: Audio Loopback  (CLoopback::~CLoopback())
 **
 ** Abstract:
 **  Destructor
 **
 ** Error Conditions: None
 **
 ***************************************************************/
CLoopback::~CLoopback()
{
  if(pipeline)
    {
      pipeline->stop();
      delete pipeline;
      pipeline = NULL;		
    }
  
  if (opi)
    {
      delete opi;
      opi = NULL;
    }
  
}

/****************************************************************
 ** Method: Audio Loopback  (CLoopback::Loopback())
 **
 ** Abstract:
 **   This loopback function uses the first instance of the device type
 **   to perform video capture & video compression and then uses the last 
 **   instance of the device type to perform decompression.
 **   If you only have 1 osprey1000 card installed in your system, 
 **   then the same card will be used for the loopback operations.  
 **
 ** Error Conditions: 
 **   No o1k device installed.
 **   Could not create pipeline.
 **   Could not map to pipeline.
 ***************************************************************/
int CLoopback::Loopback()
{
  int num_of_instances;
  
  // Getting the instances of the device type(o1k) that is installed in your system.
  if (getDeviceInfo(opi,num_of_instances) == APP_ERROR)
    return APP_ERROR;
  
  // Start creating the compression pipeline.
  if ((pipeline = new OPIPipeline(opi)) == NULL)
    {
      cerr << "Could not create encode pipeline.\n";
      return APP_ERROR;
    }
  
  // Using the first instance of the device type (ie. o1k0) for 
  // capturing the decompressed audio.
  // Setting up the capture format variables for the first pipeline stage.
  OPIFormatAudio captureFmt;
  captureFmt.compression   = "ANALOG";
  captureFmt.externalInput = TRUE;
  captureFmt.deviceName    = devinfo[0].inst_name;
  
  // Setting up decompressed format variables for the first pipeline stage.
  OPIFormatAudio decompressedFmt;
  decompressedFmt.compression = "PCM";
  
  OPIFormatAudio compressedFmt;
  compressedFmt.compression = "G728";
	
  OPIStage* sh = pipeline->firstStage(CAP_AUD_CAPTURE, &captureFmt, &decompressedFmt);
  OPIStage* compressionStage = pipeline->addStage(sh, CAP_AUD_COMPRESSION, &decompressedFmt, &compressedFmt);
	
  compressedFmt.deviceName = devinfo[num_of_instances-1].inst_name;
  
  // Creating the decompression stage of the pipeline.
  OPIStage* audioDecompressionStage;
  audioDecompressionStage = pipeline->addStage(compressionStage, CAP_AUD_DECOMPRESSION, &compressedFmt, &decompressedFmt);
  OPIStage* audioOutputStage = pipeline->addStage(audioDecompressionStage, CAP_AUD_OUTPUT, &decompressedFmt, &captureFmt);
  
	// Assigning the actual hardware and software to 
  // each of the required actions. If it returns false
  // then delete the pipeline and window.
  if (!pipeline->mapToDevices())
    {  
      delete pipeline;
      pipeline  = NULL;
      return APP_ERROR;
    }
  
  OPICaptureDeviceAudio *ca = (OPICaptureDeviceAudio*)pipeline->getDeviceFromHandle(audioDecompressionStage);
  //sample rate and channels are relevant for PCM & G.711.  Will be ignored for G.728 & G.722
  ca->setSampleRate(22050);
  ca->setChannels(2);
  
  // Finish the construction of the pipeline.
  pipeline->finishConstruction();
  
  // Start the pipeline.
  pipeline->start();
  
  return APP_OK;
}

/****************************************************************
 ** Method: Audio Loopback  (CLoopback::getDeviceInfo())
 **
 ** Abstract:
 **   Checks for the number of instances of the device type (o1k)
 **   in your system and store the instances into a devinfo structure.  
 **
 ** Error Conditions: 
 **   No o1k device installed.
 ***************************************************************/
int CLoopback::getDeviceInfo(OPISystem* opisys,int& num_of_instances)
{
  OPIDevice* dp;
  dp = opisys->getDeviceByName("o1k");
  
  if(dp==NULL) {
    cerr << "No o1k device installed\n";
    return APP_ERROR;
  }
  else
    {
      // Getting the number of instances of the device
      num_of_instances = dp->instances.getNumItems();		
      devinfo = new DeviceInfo[num_of_instances];
      
      // Going thru the loop to get the name of the instances 
      // of the device type. The instances name should begin with
      // o1k0 if that is the only instance of the device type in
      // your system.
      // Storing the name of the instance of the device type to a 
      // devinfo structure.
      char* name_of_device;
      for (int j = 0; j < num_of_instances; j++)
	{
	  name_of_device = dp->instances.getItemName(j);
	  strcpy(devinfo[j].inst_name, name_of_device);
	}
    }
  return APP_OK;
}
