/* $Id: loopback.cpp 1.10 1998/09/25 19:03:32 scl Initial_Build $ */
/****************************************************************
 ** File: File Long Name ($RCSfile: loopback.cpp $)
 **
 ** ($Revision: 1.10 $)
 **
 ** Abstract:
 **   Contains the implementations for the loopback operations.
 **   For this loopback example, we have chosen to use the first
 **   instance of the device for video capture & video compression
 **   and chosen the last instance of the device for decompression.
 **   Note: If you only have 1 osprey1000 card installed in your system, 
 **   then the same card will be used for the loopback operations.
 **   
 ** Copyright (c) 1998, MultiMedia Access Corp
 **                     Osprey Technologies Division
 ****************************************************************/
#include <iostream.h>
#include <string.h>
#include "loopback.h"

/****************************************************************
 ** Method: Loopback  (CLoopback::CLoopback)
 **
 ** Abstract:
 **  Constructor
 **
 ** Error Conditions: None
 **
 ***************************************************************/
CLoopback::CLoopback()
{
  pipeline  = NULL;
  localWin  = NULL;
  decompWin = NULL;
    
  // Open the OPI system.
  opi = new OPISystem();
  
  // setDebugLevel will control the amount of debug information 
  // provided by OPI system. Currently, the only effect 
  // this has is disabling detection of streaming timeouts
  // if passed a non-zero value.
  opi->setDebugLevel(1);
}

/****************************************************************
 ** Method: Loopback  (CLoopback::~CLoopback)
 **
 ** Abstract:
 **  Destructor
 **
 ** Error Conditions: None
 **
 ***************************************************************/
CLoopback::~CLoopback()
{
  if(pipeline)
    {
      pipeline->stop();
      delete pipeline;
    }
    
  if (localWin)
    delete localWin;
  
  if (decompWin)
    delete decompWin;
  
  if (opi)
    delete opi;
}

/****************************************************************
 ** Method: Loopback  (CLoopback::Loopback)
 **
 ** Abstract:
 **   This loopback function uses the first instance of the device type
 **   to perform video capture & video compression and then uses the last 
 **   instance of the device type to perform decompression.
 **   If you only have 1 osprey1000 card installed in your system, 
 **   then the same card will be used for the loopback operations.
 **
 ** Error Conditions: 
 **   No o1k device installed.
 **   Could not create encode pipeline.
 **   Could not create OPI window.
 **   Could not map to pipeline. 
 **
 ***************************************************************/
int CLoopback::Loopback()
{

  // Check whether o1k device is present within system and get pointer.
  OPIDevice* dp = opi->getDeviceByName("o1k");
  if(dp==NULL) 
    {
      cerr << "No o1k device installed\n";
      return APP_ERROR;
    }

  // Start creating the compression pipeline.
  if ((pipeline = new OPIPipeline(opi)) == NULL)
    {
      cerr << "Could not create encode pipeline.\n";
      delete pipeline;
      delete opi;
      return APP_ERROR;
    }
  
  // Use the first instance of the device type for capturing uncompressed
  // video.
  // Get the pointer to the first instance of the device.
  OPIDeviceInstance* inst = dp->getInstanceByIndex(0);

  // Get the name of the first instance of the device.
  char* selected_device = inst->getName();

  // Setting up the external format variables for the first pipeline stage.
  OPIFormatVideo externalFmt;
  externalFmt.compression   = VIDEO_FORMAT;
  externalFmt.deviceName    = selected_device;
  
  // Setting up uncompressed format variables for the first pipeline stage.
  OPIFormatVideo uncompressedFmt;
  uncompressedFmt.width       = VIDEO_WIDTH;
  uncompressedFmt.height      = VIDEO_HEIGHT;
  uncompressedFmt.depth       = 16;
  uncompressedFmt.compression = "RGB565";
  uncompressedFmt.deviceName  = externalFmt.deviceName;
  
  // Creating the first pipeline stage which converts the 
  // captured image into RGB565.
  OPIStage* sh;
  sh = pipeline->firstStage(CAP_VID_CAPTURE, &externalFmt, &uncompressedFmt);
  
  // Setting up the compression format variables for the compression stage.
  OPIFormatVideo compressedFmt;
  compressedFmt.compression = "IP64";
  compressedFmt.width       = VIDEO_WIDTH;
  compressedFmt.height      = VIDEO_HEIGHT;
  compressedFmt.depth       = 16;
  compressedFmt.deviceName  = externalFmt.deviceName;
  
  // Creating the compression stage of the pipeline to convert RGB565 to compressed IP64.
  OPIStage* videoCompressionStage;
  videoCompressionStage = pipeline->addStage(sh,CAP_VID_COMPRESSION, &uncompressedFmt, &compressedFmt);
  
  // Setting up the window format variables for the window display for the 
  // uncompressed image that was captured by the first device instance.
  OPIFormatVideo windowFmt;
  windowFmt.compression    = "WINDOW";
  windowFmt.externalOutput = FALSE;
  windowFmt.width          = WINDOW_WIDTH;
  windowFmt.height         = WINDOW_HEIGHT;
  windowFmt.depth          = 16;
  
  // Create a separate OPI window to display local raw video.
  if((localWin = new OPIWindow(10,0,windowFmt.width, windowFmt.height, "Local raw video")) == NULL)
    {
      cerr << "Could not create opi window.\n";
      return APP_ERROR;
    }
  windowFmt.param = (void *)localWin;
  
  // Create branch from first pipeline stage for viewing captured video.
  pipeline->addStage(sh, CAP_VID_OUTPUT, &uncompressedFmt, &windowFmt);
  
  // Set the uncompressedFmt device name to the last instance of the device type.
  // Note: If you have only 1 osprey1000 card installed in your system, then the device 
  //       name would be the same as that which was used previously.
  //       Else if you have more than 1 osprey1000 card installed in your system, 
  //       then the device name would be the last osprey1000 card found in your system.

  // Getting the number of instances of the device
  int num_of_instances = dp->instances.getNumItems();		
  
  // Get the pointer to the last instance of the device.
  inst = dp->getInstanceByIndex(num_of_instances - 1);

  // Get the name of the last instance of the device.
  selected_device = inst->getName();

  uncompressedFmt.deviceName = selected_device;
  
  // Creating the decompression stage of the pipeline.
  OPIStage* videoDecompressionStage;
  videoDecompressionStage = pipeline->addStage(videoCompressionStage, CAP_VID_DECOMPRESSION, &compressedFmt, &uncompressedFmt);
  
  // Creating the window to display the decompressed view.
  if((decompWin = new OPIWindow(10,300,windowFmt.width,windowFmt.height,"Decompressed video")) == NULL)
    {
      cerr << "Could not create OPI window.\n";
      return APP_ERROR;
    }
  windowFmt.param = (void *)decompWin;
  
  // Creating the viewing stage of the pipeline for the viewing of the decompressed video.
  pipeline->addStage(videoDecompressionStage, CAP_VID_OUTPUT, &uncompressedFmt, &windowFmt);
  
  // Assigning the actual hardware and software to 
  // each of the required actions. If it returns false
  // then delete the pipeline and window.
  if (!pipeline->mapToDevices())
    { 
      cerr << "Could not map to pipeline.\n";
      delete pipeline;
      delete localWin;
      delete decompWin;
      delete opi;
      return APP_ERROR;
    }
  
  // Getting the specific compression device and setting some of the
  // capabilities of the compression device.
  OPICompressionDeviceH261 *cdH261;
  cdH261 = (OPICompressionDeviceH261 *)pipeline->getDeviceFromHandle(videoCompressionStage);
  cdH261->setQuality(31);
  cdH261->setBitRate(1024);
  cdH261->setBlockRefreshPeriod(3000);
  
  // Getting the specific capture device and setting the framerate and the overlay rate.
  OPICaptureDeviceVideo *cap = (OPICaptureDeviceVideo *)pipeline->getDeviceFromHandle(videoCompressionStage);
  cap->setFrameRate(33);       // frame rate is measured in msec per frame
  cap->setOverlayRate(1000);   // Set frame rate for local video
  
  // Finish the construction of the pipeline.
  pipeline->finishConstruction();
  
  // Start the pipeline.
  pipeline->start();
	
  return APP_OK;
}

