/* $Id: record.cpp 1.9 1998/10/21 14:39:06 scl Initial_Build $ */
/*************************************************************
 ** File: File Long Name ($RCSfile: record.cpp $)
 **
 ** ($ Revision: $)
 **
 ** Abstract:
 **    Class Implementation for the record operations.
 ** 
 **   Copyright (c) 1998, MultiMedia Access Corp
 **                     Osprey Technologies Division
 *************************************************************/
#include <iostream.h>
#include <stdio.h>
#include "record.h"

/****************************************************************
 ** Method: Record H261  (CRecord::CRecord)
 **
 ** Abstract:
 **  Constructor
 **
 ** Error Conditions: None
 **
 ***************************************************************/
CRecord::CRecord()
{
  pipeline      = NULL;
  fd            = NULL;
  videofile     = NULL;
  window        = NULL;
  buffers_saved = 0;
  localview     = FALSE;
  
  // Open the OPI system.
  opi = new OPISystem();
  
  // setDebugLevel will control the amount of debug information 
  // provided by OPI system. Currently, the only effect 
  // this has is disabling detection of streaming timeouts
  // if passed a non-zero value.
  opi->setDebugLevel(1);	
}

/****************************************************************
 ** Method: Record H261  (CRecord::~CRecord)
 **
 ** Abstract:
 **  Destructor
 **
 ** Error Conditions: None
 **
 ***************************************************************/
CRecord::~CRecord()
{
  if (opi)
    delete opi;    
}
/****************************************************************
 ** Method: Record H261  (CRecord::Record)
 **
 ** Abstract:
 **   Opens a file for writing.
 **   Creates a pipeline to capture NTSC and converts it into RGB565.
 **   Creates another pipeline to convert RGB565 to compressed IP64.
 **   Creates a branch of pipeline to display the local view to an
 **   opi window while the video is being captured to a file.
 **
 ** Error Conditions: 
 **   Could not create pipeline.
 **  Could not create opi window.
 **  Could not map to pipeline.
 **
 ***************************************************************/
int CRecord::Record(int framecount)
{
  // Capture H.261 to a file
  if (videofile) 
    fd = fopen(videofile,"wb");
  
  // Set up the external format variables for the first stage of the pipeline.
  OPIFormatVideo externalFmt;
  externalFmt.compression   = "NTSC";
    
  // Set up the captured format variables for the first stage of the pipeline.
  OPIFormatVideo capturedFmt;
  capturedFmt.width         = 160;
  capturedFmt.height        = 120;
  capturedFmt.depth         = 16;
  capturedFmt.compression   = "RGB565";
    
  // Creating the pipeline.
  if ((pipeline = new OPIPipeline(opi)) == NULL)
    {
      cerr << "Could not Create pipeline\n";
      delete pipeline;
      return APP_ERROR;
    }	
  
  // Create the first stage of the pipeline.
  OPIStage* captureStage;
  captureStage = pipeline->firstStage(CAP_VID_CAPTURE, &externalFmt, &capturedFmt);
  
  // Set up the compressed format variables for the second stage of the pipeline.
  OPIFormatVideo compressedFmt;
  compressedFmt.compression   = "IP64";
  compressedFmt.width         = 160;
  compressedFmt.height        = 120;
  compressedFmt.depth         = 16;
    
  // Creating the 2nd stage of the pipeline.
  OPIStage* videoCompressionStage;
  videoCompressionStage = pipeline->addStage(captureStage, CAP_VID_COMPRESSION, &capturedFmt, &compressedFmt);
  
  if(localview == TRUE)
    {
      // Setting the window format variables that are needed for the
      // creation of the opi window for the local view while
      // video is being captured to the file.
      OPIFormatVideo windowFmt;
      windowFmt.compression    = "WINDOW";
      windowFmt.externalOutput = FALSE;
      windowFmt.width          = 160;
      windowFmt.height         = 120;
      windowFmt.depth          = 16;
      
      // Creating the OPI window.
      if((window = new OPIWindow(10,0,windowFmt.width, windowFmt.height, "Local View"))== NULL)
	{
	  cerr << "Could not create OPI window.\n";
	  return APP_ERROR;
	}
      windowFmt.param = (void*) window;
      
      // Creating a branch of pipeline for the viewing of the captured video.
      pipeline->addStage(captureStage,CAP_VID_OUTPUT, &capturedFmt, &windowFmt);
    }
  
  
  // Assigning the actual hardware and software to
  // each of the required actions in the pipeline. 
  // If it returns false then delete the pipeline and window.
  if (!pipeline->mapToDevices()) 
    {  
      cerr <<" Could not map to pipeline\n";
      if (window)
        delete window;
      delete pipeline;
      return APP_ERROR;
    }

  // Getting the specific compression device and the capture device.
  // Setting some of the attributes of the compression device and 
  // capture device.
  OPICompressionDeviceH261 *cdH261;
  cdH261 = (OPICompressionDeviceH261 *)pipeline->getDeviceFromHandle(videoCompressionStage);
  cdH261->setQuality(15);
  cdH261->setBitRate(256);
  cdH261->setBlockRefreshPeriod(3000);
  
  OPICaptureDeviceVideo *cap = (OPICaptureDeviceVideo *)pipeline->getDeviceFromHandle(captureStage);
  cap->setFrameRate(50);    // in milliseconds per frame
  cap->setOverlayRate(100); // set the frame rate for local view.
  
  // Finish the pipeline construction
  pipeline->finishConstruction();
  
  // Set up the callback on the output stream.
  pipeline->getOutStream()->setDataReadyCallback(&SaveBuffer, (void*)this);
  
  // Start the pipeline.
  pipeline->start();
  
  while (framecount > buffers_saved)
    OPISleep(10);

  // Stop the pipeline
  pipeline->stop();

  // Delete the window and then the pipeline
  if (window)
    delete window;
  delete pipeline;

  fclose(fd);

  return APP_OK;
}

/****************************************************************
 ** Method: Record H261  (CRecord::SaveBuffer)
 **
 ** Abstract:
 **   Writes the image data to the file.
 **   Increments the variable buffers_saved and returns the buffer.
 **
 ** Error Conditions: None
 **
 ***************************************************************/
void CALLBACK CRecord::SaveBuffer(OPIBuffer* buffer, void* userData)
{ 
  CRecord* recordPtr = (CRecord*) userData;
  fwrite(buffer->data, 1, buffer->fsize, recordPtr->fd);
  recordPtr->buffers_saved++;
  recordPtr->pipeline->getOutStream()->returnBuffer(buffer);
  
} 
