#include <stdio.h>
#include <string.h>
#include <tcl.h>
#include "opi.h"

extern "C"
{
int OpiCmd(ClientData clientData,
	   Tcl_Interp *interp,
	   int argc, char *argv[]);

// The initialization procedure that is called when the packet is loaded


int Opi_Init(Tcl_Interp *interp);
}  // extern "C"


class OPIData {
 private:
  OPISystem* sys;
  OPIDeviceInstance* di;
  int fake;
 public:
  OPIData();
  int init(Tcl_Interp *interp, int argc, char* argv[]);
  int dummy(Tcl_Interp *interp, int argc, char* argv[]);
  int control_device(Tcl_Interp *interp, int argc, char* argv[]);
  int get_attribute(Tcl_Interp *interp, int argc, char* argv[]);
  int get_minmax(Tcl_Interp *interp, int argc, char* argv[]);
  int set_attribute(Tcl_Interp *interp, int argc, char* argv[]);
  int num_devices(Tcl_Interp *interp, int argc, char* argv[]);
  int initialized() { return sys != NULL; }
};


// The "dummy" command basically says use whatever values you can that are
// legit, but if you ever get a request for a non-existent attribute then
// just go ahead and return "normal" values.

//  This allows you to test o1k control panels on a o1c machine.

OPIData::OPIData()
{
  sys = NULL;
  di = NULL;
  fake = 0;
}



int OPIData::init(Tcl_Interp *interp, int argc, char* argv[])
{
  sys = new OPISystem();
  if (sys == NULL) {
    interp->result = "OPI System returned NULL";
    return TCL_ERROR;
  }
  return TCL_OK;
}


int OPIData::dummy(Tcl_Interp *interp, int argc, char* argv[])
{
  fake = 1;
  return TCL_OK;
}


int OPIData::control_device(Tcl_Interp *interp, int argc, char* argv[])
{
  if (argc != 4) {
    interp->result = "Usage: opi control_device <devicetype> <devicename>";
    return TCL_ERROR;
  }
  di = sys->getControlInstance(argv[2], argv[3]);
  if (di == NULL) {
    sprintf(interp->result, "No device %s with instance %s", argv[2], argv[3]);
    return TCL_ERROR;
  }
  return TCL_OK;
}


int OPIData::get_attribute(Tcl_Interp *interp, int argc, char* argv[])
{
  if (di == NULL) {
    interp->result = "You must first set the control device before calling get_attribute";
    return TCL_ERROR;
  }
  if (argc != 3) {
    interp->result = "Usage: opi get_attribute <attribute_name>";
    return TCL_ERROR;
  }

  OPIAttributeInfo* oinfo;
  OPIAttribute attr;
  int error = TCL_OK;
  int val;
  if (di->getAttributeInfo(argv[2], &oinfo) == FALSE) {
    if (fake) {
      sprintf(interp->result, "%ld", 1);
      return TCL_OK;
    }
    sprintf(interp->result,"No attribute %s for this device", argv[2]);
    return TCL_ERROR;
  }

  di->getAttribute(argv[2], &attr);

  switch (oinfo->getAttributeType()) {
  case INT_ATTRIBUTE:
    sprintf(interp->result, "%ld", attr.ptr);
    break;
  case PTR_ATTRIBUTE:
    sprintf(interp->result, "%ld", attr.i);
    break;
  case STRING_ATTRIBUTE:
    sprintf(interp->result, "%s", attr.s);
    break;
  case DOUBLE_ATTRIBUTE:
    sprintf(interp->result, "%g", attr.d);
    break;
  }
  return TCL_OK;
}

int OPIData::get_minmax(Tcl_Interp *interp, int argc, char* argv[])
{
  if (di == NULL) {
    interp->result = "You must first set the control device before calling get_attribute";
    return TCL_ERROR;
  }
  if (argc != 3) {
    interp->result = "Usage: opi get_minmax <attribute_name>";
    return TCL_ERROR;
  }

  OPIAttributeInfo* oinfo;
  OPIAttribute attr;
  int error = TCL_OK;
  int val;
  if (di->getAttributeInfo(argv[2], &oinfo) == FALSE) {
    if (fake) {
      sprintf(interp->result, "%ld %ld", 0, 100);
      return TCL_OK;
    }
    sprintf(interp->result,"No attribute %s for this device", argv[2]);
    return TCL_ERROR;
  }

  switch (oinfo->getAttributeType()) {
  case INT_ATTRIBUTE:
    sprintf(interp->result, "%ld %ld", oinfo->getMin().i, oinfo->getMax().i);
    break;
  case PTR_ATTRIBUTE:
  case STRING_ATTRIBUTE:
  case DOUBLE_ATTRIBUTE:
    interp->result = "This attribute type does not have a min and max";
    return TCL_ERROR;
  }
  return TCL_OK;
}


int OPIData::set_attribute(Tcl_Interp *interp, int argc, char* argv[])
{
  if (di == NULL) {
    interp->result = "You must first set the control device before calling set_attribute";
    return TCL_ERROR;
  }
  if (argc != 4) {
    interp->result = "Usage: opi set_attribute <attribute_name> <attribute_value>";
    return TCL_ERROR;
  }

  OPIAttributeInfo* oinfo;
  OPIAttribute attr;
  int error = TCL_OK;
  int val;
  if (di->getAttributeInfo(argv[2], &oinfo) == FALSE) {
    if (fake)
      return TCL_OK;
    sprintf(interp->result,"No attribute %s for this device", argv[2]);
    return TCL_ERROR;
  }

  switch (oinfo->getAttributeType()) {
  case INT_ATTRIBUTE:
  case PTR_ATTRIBUTE:
    error = Tcl_GetInt(interp, argv[3], &val);
    attr.i = (long)val;
    break;
  case STRING_ATTRIBUTE:
    attr.s = strdup(argv[3]);
    break;
  case DOUBLE_ATTRIBUTE:
    error = Tcl_GetDouble(interp, argv[3], &attr.d);
    break;
  }
  if (error != TCL_OK) {
    return error;
  }

  di->setAttribute(argv[2], attr);
  return TCL_OK;
}


int OPIData::num_devices(Tcl_Interp *interp, int argc, char* argv[])
{
  if (argc != 3) {
    interp->result = "Usage: opi num_devices <devicetype>";
    return TCL_ERROR;
  }
  OPIDevice* dp = sys->getDeviceByName(argv[2]);
  if (dp == NULL) {
    sprintf(interp->result,"No device %s exists in the system", argv[2]);
    return TCL_ERROR;
  }
  sprintf(interp->result, "%d", dp->instances.getNumItems());
  return TCL_OK;
}


int Opi_Init(Tcl_Interp *interp)
{
  OPIData* data = new OPIData();
  // Register the command version of the opi interface
  Tcl_CreateCommand(interp, "opi", OpiCmd, (ClientData)data, (Tcl_CmdDeleteProc*)NULL);

  // Declare that we implement the opi package so scripts can do "package require opi"
  Tcl_PkgProvide(interp, "opi", "1.0");
  return TCL_OK;
  
}



// OpiCmd
//	Implement the opi tcl cmd set.
//	Subcommands include
//		init:  Initialized the OPISystem variable
//		control_device <devicetype> <devicename> : Opens a specific control device  (i.e. o1c0)
//		num_devices <devicetype> : returns the number of this type of device that exists
//		set_attribute <attribute_name> <attribute_value> : sets the given attribute
//		get_attribute <attribute_name> : gets the given attribute value
//		get_minmax <attribute_name> : return the min and max separated by a space
//		dummy: pretend you support attributes you really dont support.  Still requires
//			a real control device

int OpiCmd(ClientData clientData,
	   Tcl_Interp *interp,
	   int argc, char *argv[])
{
  OPIData* data = (OPIData*)clientData;
  
  if (argc < 2) {
    interp->result = "Usage: opi [init|control_device|num_devices|set_attribute]";
    return TCL_ERROR;
  }

  if (strcmp(argv[1], "dummy") == 0) {
    return data->dummy(interp, argc, argv);
  } else if (strcmp(argv[1], "init") == 0) {
    return data->init(interp, argc, argv);
  }

  if (!data->initialized()) {
    interp->result = "You must first call opi init before calling other opi subcommands";
    return TCL_ERROR;
  }

  
  if (strcmp(argv[1], "control_device") == 0) {
    return data->control_device(interp, argc, argv);
  } else if (strcmp(argv[1], "num_devices") == 0) {
    return data->num_devices(interp, argc, argv);
  } else if (strcmp(argv[1], "set_attribute") == 0) {
    return data->set_attribute(interp, argc, argv);
  } else if (strcmp(argv[1], "get_attribute") == 0) {
    return data->get_attribute(interp, argc, argv);
  } else if (strcmp(argv[1], "get_minmax") == 0) {
    return data->get_minmax(interp, argc, argv);
  }

  interp->result = "Unrecognized opi subcommand";
  return TCL_ERROR;

  return TCL_OK;
}


