/* $Id: main.cpp 1.4 1998/07/24 18:57:02 ott Release $ */
/**************************************************************
 ** File: Video Capture and Display ($RCSfile: main.cpp $)
 **
 ** ($Revision: 1.4 $)
 **
 ** Abstract:
 **   This OPI example demonstrates Video capture, store and
 **   display.  It is implemented in vidcapdsp.cpp.  Briefly
 **   it uses two opi pipelines: Capture and Display.
 **
 ** Copyright (c) 1998, MultiMedia Access Corp
 **                     Osprey Technologies Division
 *************************************************************/
#include <stdlib.h>
#include <stream.h>
#include "opi.h"
#include "vidcapdsp.h"

// Default values
#define  COLORFMT  "RGB565"
#define  CAPFMT    "NTSC"
#define  DEPTH     16
#define  MAXWIDTH  640
#define  MAXHEIGHT 480
#define  SIZEDIV   2
#define  NFRAMES   50
#define  VIDFILE   "video.raw"

/*****************************************************************************
 ** Function: Display Program Usage ( usage )
 **
 ** Abstract:
 **   Prints out the valid command line arguments and then exits.
 **
 ** Error Conditions: None
 **
 ****************************************************************************/
void usage(char* appName)
{
  fprintf(stderr, "usage: %s [options]\n",appName);
  fprintf(stderr, "Legal Options |       Description           | Defaults \n");
  fprintf(stderr, "--------------|-----------------------------|----------\n");
  fprintf(stderr, "-c format     | color format                |   RGB565  \n");
  fprintf(stderr, "-v capformat  | capture format              |   NTSC   \n");
  fprintf(stderr, "-z depth      | color depth                 |   16     \n");
  fprintf(stderr, "-w width      | width                       |   320    \n");
  fprintf(stderr, "-h height     | height                      |   240    \n");
  fprintf(stderr, "-s sizediv    | size divisor                |   2      \n");
  fprintf(stderr, "-n framecount | number of frames to capture |   50     \n");
  fprintf(stderr, "-f filename   | output file name            |   video.raw\n");
  fprintf(stderr, "-o options    | list command line options   |          \n");
  exit(0);
}

/**************************************************************
 ** Function: Video Capture and Display Main (main)
 **
 ** Abstract:
 **   This function used the CVidCapDsp class which is 
 **   implemented in vidcapdsp.cpp to display video and
 **   save it to a file.
 **
 ** Error Conditions: 
 **   OPIInternalError. Closing down.
 **   CheckForDevice failed. Closing down.
 **   StartDisplay failed. Closing down.
 **   StartCapture failed.  Closing down.
 **   Error capturing frames.
 **
 **************************************************************/

int main(int argc, char** argv)
{
  char *colorFormat = COLORFMT;
  char *capFormat   = CAPFMT;
  int   colorDepth  = DEPTH;
  int   width       = MAXWIDTH;
  int   height      = MAXHEIGHT;
  int   sizeDivisor = SIZEDIV;
  int   nframes     = NFRAMES;
  char *filename    = VIDFILE;

  FILE        *fd          = NULL;
  CVidCapDsp  *vidcapdsp   = NULL;
  char c;

  // Process command line arguments
  while ((c = getopt(argc, argv, "c:v:z:w:h:s:n:f:o:")) != EOF)
  {
    switch (c)
    {
      case 'c': colorFormat  = optarg;                              break;
      case 'v': capFormat    = optarg;                              break;
      case 'z': colorDepth   = atoi(optarg);                        break; 
      case 'w': width        = atoi(optarg);    sizeDivisor = 1;    break;
      case 'h': height       = atoi(optarg);    sizeDivisor = 1;    break;
      case 's': sizeDivisor  = atoi(optarg);                        break;
      case 'n': nframes      = atoi(optarg);                        break;
      case 'f': filename     = optarg;                              break;
      case 'o': usage(argv[0]);
      default: usage(argv[0]);
      break;
    }
  }

  width = width/sizeDivisor;
  height = height/sizeDivisor;

  // Build the class 
  try
  {
    vidcapdsp = new CVidCapDsp();
  }
  catch (OPIInternalError)
  {
    printf("OPIInternalError. Closing down.\n");
    return FALSE;
  }

  // Create/Start OPI Pipelines
  if (vidcapdsp->StartDisplay(colorFormat, colorDepth, width, height) != 0)
  {
    printf("StartDisplay failed. Closing down.\n");
    vidcapdsp->StopAll();
    delete vidcapdsp;
    exit(0);
  }

  // Create/Start OPI Pipelines
  if (vidcapdsp->StartCapture(capFormat, colorFormat, colorDepth,
                              width,height) != 0)
  {
    printf("StartCapture failed.  Closing down.\n");
    vidcapdsp->StopAll(); 
    delete vidcapdsp;
    exit(0); 
  }

  // Open file to store video data 
  fd = fopen(filename,"wb");
  if (fd==NULL)
  {
    printf("Unable to open file for storing video data.\n");
    vidcapdsp->StopAll(); 
    delete vidcapdsp;
    exit(0); 
  }

  // Capture frames
  if (vidcapdsp->CaptureFrames(nframes, fd) != 0)
  {
    printf("Error capturing frames.\n");
  }

  // Stop OPI Pipelines 
  vidcapdsp->StopAll(); 

  // Close down OPI
  delete vidcapdsp;

  // Close video data file
  fclose(fd);

  return TRUE;

}


