/* $Id: vidcapdsp.cpp 1.8 1998/07/24 19:38:12 ott Initial_Build $ */
/*******************************************************************
 ** File: Video Capture and Display ($RCSfile: vidcapdsp.cpp $)
 **
 ** ($Revision: 1.8 $)
 **
 ** Abstract:
 **   Demonstrates using a video capture pipeline to feed
 **   a video display pipeline and write data to a file
 **
 ** Revision History:
 **   $Log: vidcapdsp.cpp $
 **   Revision 1.8  1998/07/24 19:38:12  ott
 **   #include <stream.h> changed to #include <iostream.h>
 **   Revision 1.7  1998/07/24 19:01:10  ott
 **   Added parameters to some member functions.
 **   Various pipeline settings are now passed in rather than hard-coded.
 **   Revision 1.6  1998/07/10 20:02:21  brettb
 **   Coding standards update.
 **   Revision 1.5  1998/07/09 21:42:40  brettb
 **   coding standards.
 **   Revision 1.4  1998/07/09 20:59:01  brettb
 **   coding standards.
 **
 ** Copyright (c) 1998, MultiMedia Access Corp
 **                     Osprey Systems Division
 ******************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <iostream.h>
#include <string.h>
#include "opi.h"
#include "vidcapdsp.h"


/******************************************************************
 ** Function: Video Capture And Display Class Ctor (CVidCapDsp)
 **
 ** Abstract:
 **   Initializes OPI.
 **
 ** Error Conditions: 
 **    Unable to create OPISystem object.
 **
 ******************************************************************/
CVidCapDsp::CVidCapDsp()
{
  opiSystem     = NULL;
  capPipeline   = NULL;
  dspPipeline   = NULL;
  capVidStream  = NULL;
  dspVidStream  = NULL;

  // Create a new OPI System
  opiSystem = new OPISystem();
  if (opiSystem==NULL)
  {
    cerr << "Unable to create OPISystem object." << endl;
    exit(0);
  }

  // setDebugLevel will control the amount of debug information 
  // provided by OPI system. Currently, the only effect 
  // this has is disabling detection of streaming timeouts
  // if passed a non-zero value.
  opiSystem->setDebugLevel(1);

}

/*******************************************************************
 ** Function: Video Capture And Display Class Dtor (CVidCapDsp)
 **
 ** Abstract:
 **   Shuts down OPI.
 **
 ** Error Conditions: None
 **
 ******************************************************************/
CVidCapDsp::~CVidCapDsp()
{
  // delete opi system if it exists
  if (opiSystem)
  {
    delete opiSystem; 
    opiSystem = NULL;
  }
}


/*******************************************************************
 ** Function: Start Display (CVidCapDsp::StartDisplay)
 **
 ** Abstract:
 **   Builds a display pipeline.
 **
 ** Error Conditions: 
 **   Could not create display pipeline
 **   Unable to map display pipeline
 **
 ******************************************************************/
int CVidCapDsp::StartDisplay(char* color_format, int color_depth, int w, int h)
{  

  // Create the display pipeline
  if ((dspPipeline = new OPIPipeline(opiSystem)) == NULL) 
  {
    cerr << "Could not create display pipeline" << endl;
    return -1;
  }     
  
  // Initialize Display Input Format
  OPIFormatVideo dspFormat;
  dspFormat.compression   = color_format;
  dspFormat.externalInput = FALSE;
  dspFormat.width         = w;
  dspFormat.height        = h;
  dspFormat.depth         = color_depth;

  // Initialize Display Output (window) Format  
  OPIFormatVideo windowFmt;
  windowFmt.compression    = "WINDOW";
  windowFmt.externalOutput = FALSE;
  windowFmt.width          = w;
  windowFmt.height         = h;
  windowFmt.depth          = color_depth;

  // Add CAP_VID_OUTPUT stage
  dspPipeline->firstStage(CAP_VID_OUTPUT, &dspFormat, &windowFmt);

  // Map the pipeline
  if (dspPipeline->mapToDevices() == FALSE) 
  { 
    cerr << "Unable to map display pipeline" << endl; 
    return -1;
  }
  
  // Create input stream with no buffers since we will be passing
  // buffers directly from the captured video stream to this pipeline
  dspVidStream = new OPIStream(INSTREAM, 0, 0);
  dspPipeline->setInStream(dspVidStream);

  // Finishing the display pipeline construction.
  dspPipeline->finishConstruction();
  
  // Start the dspPipeline.
  dspPipeline->start();   

  return 0;
}

/*******************************************************************
 ** Function: Start Capture (CVidCapDsp::StartCapture)
 **
 ** Abstract:
 **   Builds a capture pipeline.
 **
 ** Error Conditions: 
 **   Could not MAP capture pipeline
 **   Could not create an OPIPipeline
 **
 ******************************************************************/
int CVidCapDsp::StartCapture(char* cap_format, char* color_format, int color_depth, int w, int h)
{
  OPIStage       *cap_stage;
  OPIFormatVideo  outFmt;
  OPIFormatVideo  inFmt;
  
  // Set the format for the input of capture
  inFmt.compression   = cap_format;
  inFmt.externalInput = TRUE;

  // Set the format for the output of capture  
  outFmt.width  = w;
  outFmt.height = h;
  outFmt.depth  = color_depth;
  outFmt.compression = color_format;
  
  // Create the capture pipeline
  if ((capPipeline = new OPIPipeline(opiSystem)) == NULL)
  {
    cerr << "Could not create an OPIPipeline" << endl;
    return -1;
  }

  // Create the first (and only) stage in our capture capPipeline
  cap_stage = capPipeline->firstStage(CAP_VID_CAPTURE, &inFmt, &outFmt);

  // Now MAP the capture pipeline to the device
  if (!capPipeline->mapToDevices()) 
  {  
    cerr << "Could not MAP capture pipeline" << endl;
    return -1;
  }

  // Finish stage the capPipeline construction
  capPipeline->finishConstruction();

  // Get the output stream from the capture pipeline
  capVidStream = capPipeline->getOutStream(NULL);

  // Start the capture pipeline
  capPipeline->start();

  return 0;
}

/*******************************************************************
 ** Function: Stop All (CVidCapDsp::StopAll)
 **
 ** Abstract:
 **    Stops and and destroyes the capture and the display pipeline.  
 **
 ** Error Conditions: None
 **
 ******************************************************************/
int CVidCapDsp::StopAll()
{
  // stop/delete capture pipeline if it exists
  if (capPipeline)
  {
    capPipeline->stop();  
    delete capPipeline;
    capPipeline = NULL;
  }

  // stop/delete display pipeline if it exists
  if (dspPipeline)
  {
    dspPipeline->stop();  
    delete dspPipeline;
    dspPipeline = NULL;
  }


  return 0;
}

/*******************************************************************
 ** Function: Capture Frames (CVidCapDsp::CaptureFrames)
 **
 ** Abstract:
 **   Gets data from the video stream pipeline and gives it
 **   to the display pipeline until all of the frames have
 **   been received.
 **
 ** Error Conditions: None
 **
 ******************************************************************/
int CVidCapDsp::CaptureFrames(int nframes, FILE *fd)
{
  int        frame_count = 0;
  OPIBuffer *buf         = NULL;

  // Sleep until desired number of frames acquired
  while (frame_count < nframes) 
  {
    
    // See if the display pipeline is finished with any buffers.
    // if so, return them to the capture pipeline.
    while ((buf=dspVidStream->getBuffer()) != NULL)
    {
      capVidStream->returnBuffer(buf);
    }

    // Wait for capture buffer to be available
    if ((buf=capVidStream->getData()) == NULL)
    {
      OPISleep(5);
      continue;
    }

    // Write captured video data to file 
    fwrite(buf->data, 1, buf->fsize, fd);
        
    // Add buffer to display stream 
    dspVidStream->addData(buf);

    frame_count++;
  }
  return 0;
}
