/* $Id: play.cpp 1.9 1998/10/21 14:12:18 scl Initial_Build $ */
/*****************************************************************************
 ** File: File Long Name ($RCSfile: play.cpp $)
 **
 ** ($Revision: 1.9 $)
 **
 ** Abstract:
 **   Contains the implementation for the playback operations.
 **
 ** Revision History:
 **   $Log: play.cpp $
 **   Revision 1.9  1998/10/21 14:12:18  scl
 **   fixing problem where window deleted before pipeline stopped
 **   Revision 1.8  1998/10/14 14:11:37  scl
 **   added stdio.h
 **   Revision 1.7  1998/07/27 18:56:11  ott
 **   Minor coding standard change.
 **
 ** Copyright (c) 1998, MultiMedia Access Corp. - Osprey Systems Division
 ****************************************************************************/
#include <iostream.h>
#include <stdio.h>
#include "play.h"

/****************************************************************
 ** Method: PlayH261  (CPlay::CPlay)
 **
 ** Abstract:
 **  Constructor
 **
 ** Error Conditions: None
 **
 ***************************************************************/
CPlay::CPlay()
{
  pipeline   = NULL;
  num_queued = 0;
  videofile  = NULL;
  
  // Open the OPI system.  
  opi = new OPISystem();
  
  // setDebugLevel will control the amount of debug information 
  // provided by OPI system. Currently, the only effect 
  // this has is disabling detection of streaming timeouts
  // if passed a non-zero value.
  opi->setDebugLevel(1);
}

/****************************************************************
 ** Method: PlayH261  (CPlay::~CPlay)
 **
 ** Abstract:
 **  Destructor
 **
 ** Error Conditions: None
 **
 ***************************************************************/
CPlay::~CPlay()
{
  if (opi) {
    delete opi;
  }
}

/****************************************************************
 ** Method: PlayH261  (CPlay::Playback)
 **
 ** Abstract:
 **   Opens the video file for reading.
 **   Creates the pipeline to convert the compressed IP64 image to 
 **   uncompressed RGB565.
 **   Creates another branch of the pipeline to send the uncompressed
 **   RGB565 image to an opi window for display.
 **
 ** Error Conditions: 
 **   Could not open video file.
 **   Video file is not specified.
 **   Could not map to pipeline.
 **   Could not create OPI window.
 **   Unable to map to devices.
 **
 ***************************************************************/
int CPlay::PlayBack(char* colorFormat, int colorDepth, int w, int h)
{
  
  FILE* fd = NULL;
  
  // Open the file for reading
  if (videofile) {
    if ((fd = fopen(videofile, "rb")) == NULL) {
	  cerr<<"Could not open video file"<<endl;
	  return APP_ERROR;
	}
  } else {
    cerr << "Video file is not specified" << endl;
    return APP_ERROR;
  }
  
  // Creating the pipeline
  if ((pipeline = new OPIPipeline(opi)) == NULL) {
    cerr << "Could not create pipeline\n";
    return APP_ERROR;
  }	
  
  OPIFormatVideo recordedFmt;
  recordedFmt.compression   = colorFormat;
  recordedFmt.externalInput = FALSE;
  recordedFmt.width         = w;
  recordedFmt.height        = h;
  recordedFmt.depth         = colorDepth;
  
  OPIFormatVideo windowFmt;
  windowFmt.compression    = "WINDOW";
  windowFmt.externalOutput = FALSE;
  windowFmt.width          = recordedFmt.width;
  windowFmt.height         = recordedFmt.height;
  
  // Creating the OPI window for the image display.
  OPIWindow* window = NULL;
  if ((window = new OPIWindow(0,0,windowFmt.width,windowFmt.height,"Output")) == NULL) {
    cerr << "Could not create OPI window\n" ;
    delete pipeline;
    return APP_ERROR;
  }
  windowFmt.param = (void *)window;
  
  pipeline->firstStage(CAP_VID_OUTPUT, &recordedFmt, &windowFmt);
  
  // Assigning the actual hardware and software to each of 
  // the required actions in the pipeline. If it returns false 
  // then delete the pipeline and window.
  if (pipeline->mapToDevices() == FALSE) { 
    cerr << "Unable to map to devices\n"; 
    delete window;
    delete pipeline;
    return APP_ERROR;
  }
  
  // Finishing the pipeline construction.
  pipeline->finishConstruction();
  
  // Setting up the call back for the input stream.
  OPIStream* inVidStream = NULL;
  inVidStream = pipeline->getInStream();
  inVidStream->setDataDoneCallback(&DoneBuffer, (void*)this);
  
  // Start the pipeline.
  pipeline->start();
  
  // Read data from a file and then send data to decompressor
  OPIBuffer* vidBuf;
  for(;;) {
    //Get an empty buffer that we can use.
    if ((vidBuf = inVidStream->getBuffer()) == NULL) {
      OPISleep(10);
      continue;
    }
    // Read data from file and place it into buffer.
    if((vidBuf->fsize = fread(vidBuf->data, 1, vidBuf->bufferSize(), fd)) == 0)
      break;
    
    // Send data to pipeline.
    inVidStream->addData(vidBuf);
    OPISleep(10);
    num_queued++;
  }
  
  // Wait for buffers to empty.
  while (num_queued > 0) 
    OPISleep(50);

  // stop the pipeline
  pipeline->stop();
  
  // delete the window and then the pipeline
  delete window;
  delete pipeline;

  fclose(fd);
  
  return APP_OK;
}

/****************************************************************
 ** Method: PlayH261  (CPlay::DoneBuffer)
 **
 ** Abstract:
 **   Returns the buffer and decrements the num_queued.
 **
 ** Error Conditions: None
 **
 ***************************************************************/
void CALLBACK CPlay::DoneBuffer(OPIBuffer* buffer, void* userData)
{
	CPlay* playPtr = (CPlay*) userData;
	playPtr->pipeline->getInStream()->returnBuffer(buffer);
	playPtr->num_queued--;
}

